/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_ARBITER_MPI_EVENT_H_
#define DBUG_ARBITER_MPI_EVENT_H_ 1

#include <list>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/mpi.h"
#include "dbug/arbiter/request.h"
#include "dbug/arbiter/thread.h"
#include "dbug/common/clock.h"
#include "dbug/common/macros.h"

namespace dbug {
namespace arbiter {
namespace mpi {

class MatchedEvent;
class Send;
class Recv;
class SendRecv;

class Event {
 public:
  enum State {
    AWAITING,
    POSTED,
    COMPLETED
  };

  struct PointerComparison {
    bool operator() (const Event *e1, const Event *e2);
  };

  typedef std::set<Event *, PointerComparison> Set;

  Event(Request *request, int communicator, int origin);
  virtual ~Event();

  // accessors
  int tid() const;
  Clock clock() const;
  Request *request() const;
  int communicator() const;
  int origin() const;
  State state() const;

  // mutators
  void join_with_clock(const Clock &clock);
  void set_clock(const Clock &clock);
  void set_request(Request *request);
  void set_state(State state);

  virtual MatchedEvent *ExtractMatchedEvent() = 0;
  virtual Recv *ExtractRecv() = 0;
  virtual Send *ExtractSend() = 0;
  virtual void Match(const MPI::EventOriginMap &origin_map,
                     const std::vector<int> &nondet_choices) = 0;
  virtual void TestMatch(const MPI::EventOriginMap &origin_map,
                         std::set<Option> *options) = 0;
  virtual std::string ToString() const;

 protected:
  int tid_;
  Clock clock_;
  Request *request_;
  int communicator_;
  int origin_;
  State state_;

 private:
  DISALLOW_COPY_AND_ASSIGN(Event);
};

class MatchedEvent {
 public:
  MatchedEvent();
  virtual ~MatchedEvent();
  virtual std::string handle() const = 0;

 private:
  DISALLOW_COPY_AND_ASSIGN(MatchedEvent);
};

class Recv : public Event {
 public:
  Recv(Request *request, int communicator, int origin, int source, int tag);

  // accessors
  int source() const;
  int tag() const;
  Event *match() const;
  Event *parent() const;

  // mutators
  void set_match(Event *match);
  void set_parent(Event *parent);

  virtual MatchedEvent *ExtractMatchedEvent();
  virtual Recv *ExtractRecv();
  virtual Send *ExtractSend();
  virtual void Match(const MPI::EventOriginMap &origin_map,
                     const std::vector<int> &nondet_choices);
  Event *ProbeMatch(const MPI::EventOriginMap &origin_map,
                    const std::vector<int> &nondet_choices);
  virtual void TestMatch(const MPI::EventOriginMap &origin_map,
                         std::set<Option> *options);
  virtual std::string ToString() const;

 private:
  friend class SendRecv;
  std::vector<Send *> GenerateMatches(const MPI::EventOriginMap &origin_map);
  void InternalComputeCoscheduleSet(const Event *first,
                                    Request::Set *coschedule_set);
  void InternalMatch(const MPI::EventOriginMap &origin_map,
                     const std::vector<int> &nondet_choices,
                     const Event *first,
                     int *index);
  Event *InternalProbeMatch(const MPI::EventOriginMap &origin_map,
                            const std::vector<int> &nondet_choices,
                            const Event *first,
                            int index,
                            Clock *clock);
  void InternalTestMatch(const MPI::EventOriginMap &origin_map,
                         std::set<Option> *options,
                         Request::Set *coschedule_set,
                         std::vector<int> *nondet_choices,
                         const Event *first);

  int source_;
  int tag_;
  Event *match_;
  Event *parent_;

  DISALLOW_COPY_AND_ASSIGN(Recv);
};

class MatchedRecv : public Recv, public MatchedEvent {
 public:
  MatchedRecv(Request *request,
              int communicator,
              int origin,
              int source,
              int tag,
              std::string handle);

  // accessors
  virtual std::string handle() const;

  virtual MatchedEvent *ExtractMatchedEvent();
  virtual std::string ToString() const;

 private:
  std::string handle_;

  DISALLOW_COPY_AND_ASSIGN(MatchedRecv);
};

class Send : public Event {
 public:
  enum Type {
    BUFFERED,
    SYNCHRONOUS
  };

  Send(Request *request,
       int communicator,
       int origin,
       int nbytes,
       int destination,
       int tag,
       Type type);

  // accessors
  int nbytes() const;
  int destination() const;
  int tag() const;
  Type type() const;
  Event *match() const;
  Event *parent() const;

  // mutators
  void set_type(Type type);
  void set_match(Event *match);
  void set_parent(Event *parent);

  virtual MatchedEvent *ExtractMatchedEvent();
  virtual Recv *ExtractRecv();
  virtual Send *ExtractSend();
  virtual void Match(const MPI::EventOriginMap &origin_map,
                     const std::vector<int> &nondet_choices);
  virtual void TestMatch(const MPI::EventOriginMap &origin_map,
                         std::set<Option> *options);
  virtual std::string ToString() const;

 private:
  friend class SendRecv;
  std::vector<Recv *> GenerateMatches(const MPI::EventOriginMap &origin_map);
  void InternalComputeCoscheduleSet(const Event *first,
                                    Request::Set *coschedule_set);
  void InternalMatch(const MPI::EventOriginMap &origin_map,
                     const std::vector<int> &nondet_choices,
                     const Event *first,
                     int *index);
  void InternalTestMatch(const MPI::EventOriginMap &origin_map,
                         std::set<Option> *options,
                         Request::Set *coschedule_set,
                         std::vector<int> *nondet_choices,
                         const Event *first);

  int nbytes_;
  int destination_;
  int tag_;
  Type type_;
  Event *match_;
  Event *parent_;

  DISALLOW_COPY_AND_ASSIGN(Send);
};

class SendRecv : public Event {
 public:
  SendRecv(Request *request,
           int communicator,
           int origin,
           int nbytes,
           int destination,
           int sendtag,
           int source,
           int recvtag);
  ~SendRecv();

  virtual MatchedEvent *ExtractMatchedEvent();
  virtual Recv *ExtractRecv();
  virtual Send *ExtractSend();
  virtual void Match(const MPI::EventOriginMap &origin_map,
                     const std::vector<int> &nondet_choices);
  virtual void TestMatch(const MPI::EventOriginMap &origin_map,
                         std::set<Option> *options);
  virtual std::string ToString() const;

 private:
  Recv *recv_;
  Send *send_;

  DISALLOW_COPY_AND_ASSIGN(SendRecv);
};

}  // namespace mpi
}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_MPI_EVENT_H_
