/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_ARBITER_POSIX_H_
#define DBUG_ARBITER_POSIX_H_ 1

#include <map>
#include <list>
#include <string>

#include "dbug/arbiter/posix-resource.h"
#include "dbug/common/macros.h"

namespace dbug {
namespace arbiter {
namespace posix {

class POSIX {
 public:
  class Statistics {
   public:
    Statistics();
    ~Statistics();

    std::list<std::string> ToString() const;

   private:
    friend class POSIX;
    int nbarriers_;
    int ncvars_;
    int nepolls_;
    int nfiles_;
    int nmutexes_;
    int npipes_;
    int nrwlocks_;
    int nsemaphores_;
    int nsockets_;
    int nspins_;

    DISALLOW_COPY_AND_ASSIGN(Statistics);
  };

  typedef std::map<std::string, Barrier *> Barriers;
  typedef std::map<std::string, CVar *> CVars;
  typedef std::map<std::string, EPoll *> EPolls;
  typedef std::map<std::string, File *> Files;
  typedef std::map<std::string, Mutex *> Mutexes;
  typedef std::map<std::string, Pipe *> Pipes;
  typedef std::map<std::string, RWLock *> RWLocks;
  typedef std::map<std::string, Semaphore *> Semaphores;
  typedef std::map<std::string, Socket *> Sockets;
  typedef std::map<std::string, Spin *> Spins;

  POSIX();
  ~POSIX();

  // accessors
  Barriers barriers() const;
  Barrier *barrier_lookup(std::string rid) const;
  CVars cvars() const;
  CVar *cvar_lookup(std::string rid) const;
  EPolls epolls() const;
  EPoll *epoll_lookup(std::string rid) const;
  Files files() const;
  File *file_lookup(std::string rid) const;
  Mutexes mutexes() const;
  Mutex *mutex_lookup(std::string rid) const;
  Pipes pipes() const;
  Pipe *pipe_lookup(std::string rid) const;
  Socket *remote_socket_lookup(std::string node_id,
                               std::string address,
                               int port) const;
  RWLocks rwlocks() const;
  RWLock *rwlock_lookup(std::string rid) const;
  Semaphores semaphores() const;
  Semaphore *semaphore_lookup(std::string rid) const;
  Sockets sockets() const;
  Socket *socket_lookup(std::string rid) const;
  Spins spins() const;
  Spin *spin_lookup(std::string rid) const;

  // mutators
  void add_barrier(Barrier *barrier);
  void add_cvar(CVar *cvar);
  void add_epoll(EPoll *epoll);
  void add_file(File *file);
  void add_mutex(Mutex *mutex);
  void add_pipe(Pipe *pipe);
  void add_rwlock(RWLock *rwlock);
  void add_semaphore(Semaphore *semaphore);
  void add_socket(Socket *socket);
  void add_spin(Spin *spin);
  void remove_barrier(Barrier *barrier);
  void remove_cvar(CVar *cvar);
  void remove_epoll(EPoll *epoll);
  void remove_file(File *file);
  void remove_mutex(Mutex *mutex);
  void remove_pipe(Pipe *pipe);
  void remove_rwlock(RWLock *rwlock);
  void remove_semaphore(Semaphore *semaphore);
  void remove_socket(Socket *socket);
  void remove_spin(Spin *spin);

  std::list<std::string> ToString() const;

 private:
  Barriers barriers_;
  CVars cvars_;
  EPolls epolls_;
  Files files_;
  Mutexes mutexes_;
  Pipes pipes_;
  RWLocks rwlocks_;
  Semaphores semaphores_;
  Sockets sockets_;
  Spins spins_;
  Statistics statistics_;

  DISALLOW_COPY_AND_ASSIGN(POSIX);
};

}  // namespace posix
}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_POSIX_H_
