/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_ARBITER_PROCESS_H_
#define DBUG_ARBITER_PROCESS_H_ 1

#include <list>
#include <map>
#include <set>
#include <string>

#include "dbug/common/macros.h"

namespace dbug {
namespace arbiter {

class Thread;

class Process {
 public:
  enum State {
    RUNNING,
    CONTINUED,
    STOPPED,
    TERMINATED,
    SIGNALED,
  };

  static std::string StateToString(Process::State state);

  Process();
  ~Process();

  // accessors
  std::set<Process *> children() const;
  bool collected() const;
  int id() const;
  int nonreentrant_function_count(std::string fname) const;
  Process *parent() const;
  int pid() const;
  int ppid() const;
  int pgid() const;
  int return_status() const;
  int signal() const;
  Process::State state() const;
  std::set<Thread *> threads() const;

  // mutators
  void add_child(Process *child);
  void add_thread(Thread *thread);
  void decrement_nonreentrant_function_count(std::string fname);
  void increment_nonreentrant_function_count(std::string fname);
  void set_parent(Process *parent);
  void set_pid(int pid);
  void set_ppid(int ppid);
  void set_pgid(int pgid);
  void set_return_status(int return_status);
  void set_state(Process::State state);
  void set_signal(int signal);
  void set_collected(bool collected);

  std::list<std::string> ToString() const;

 private:
  static int id_counter_;
  std::set<Process *> children_;
  bool collected_;
  int id_;
  std::map<std::string, int> nonreentrant_map_;
  Process *parent_;
  int pid_;
  int ppid_;
  int pgid_;
  int return_status_;
  int signal_;
  Process::State state_;
  std::set<Thread *> threads_;

  DISALLOW_COPY_AND_ASSIGN(Process);
};

}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_PROCESS_H_
