/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_ARBITER_REQUEST_H_
#define DBUG_ARBITER_REQUEST_H_ 1

#include <list>
#include <set>
#include <string>
#include <utility>
#include <vector>

#include "dbug/arbiter/thread.h"
#include "dbug/common/macros.h"
#include "dbug/common/system.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::thrift::Location;

namespace dbug {
namespace arbiter {

class Server;

class Request {
 public:
  struct PointerComparison {
    bool operator() (const Request *r1, const Request *r2);
  };

  typedef std::set<Request *, PointerComparison> Set;

  struct Resolved {
    bool operator<(const Resolved &rhs) const;
    Request *request;
    std::vector<int> choices;
  };

  typedef std::set<Resolved> ResolvedSet;

  class Group {
   public:
    bool operator<(const Group &other) const;

    Group();

    // accessors
    bool enabled() const;
    std::set<std::string> read_set() const;
    ResolvedSet resolved_set() const;
    bool serialize() const;
    std::set<std::string> write_set() const;

    // mutators
    void add_request(Request *request);
    void add_request(Request *request, std::vector<int> choices);
    void add_to_read_set(std::string id);
    void add_to_write_set(std::string id);
    void set_enabled(bool enabled);
    void set_serialize(bool serialize);

    std::list<std::string> ToString() const;

   private:
    bool enabled_;
    std::set<std::string> read_set_;
    ResolvedSet resolved_set_;
    bool serialize_;
    std::set<std::string> write_set_;
  };

  typedef std::set<Group> GroupSet;

  enum State {
    INITIAL,
    BUSY,
    DEADLOCK,
    INVALID,
    PERMISSION,
    VALID
  };

  static bool SortByThreadIDs(Request *r1, Request *r2);
  static bool SortByTimestamps(Request *r1, Request *r2);
  static std::string StateToString(State state);

  bool operator<(const Request *other) const;

  Request(Server *server,
          Thread *thread,
          const std::string &func,
          const std::vector<Location> &backtrace);
  virtual ~Request();

  // accessors
  std::vector<Location> backtrace() const;
  bool canceled() const;
  std::string function() const;
  int timestamp() const;
  Thread *thread() const;

  // mutators
  void set_nondet_choices(std::vector<int> nondet_choices);
  void set_state(State state);

  void Cancel();
  void Enable(Group *group);
  virtual void PreSchedule();
  void Remove();
  void Schedule();
  virtual void Test(Set requests, GroupSet *groups);
  virtual std::list<std::string> ToString() const;
  void WakeUp();

 protected:
  virtual void Execute() = 0;

  std::vector<Location> backtrace_;
  bool canceled_;
  System::CVar cvar_;
  std::string function_;
  std::vector<int> nondet_choices_;
  Server *server_;
  State state_;
  Thread *thread_;
  int timestamp_;

 private:
  DISALLOW_COPY_AND_ASSIGN(Request);
};

typedef std::pair<Request::Set, std::vector<int> > Option;

}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_REQUEST_H_
