/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_COMMON_BIG_FLOAT_H_
#define DBUG_COMMON_BIG_FLOAT_H_ 1

#include <gmp.h>
#include <mpfr.h>

#include <string>

#include "dbug/common/macros.h"

namespace dbug {

// A simple C++ wrapper for a subset of the arbitrary-precision
// arithmetic of the GMP and MPFR libraries. This class defines the
// common arithmetic and comparison operations, which is used by dbug
// to estimate the execution tree size.
class BigFloat {
 public:
  static void Initialize();
  static void Finalize();

  // Constructors.
  BigFloat();
  BigFloat(const BigFloat &bf);
  explicit BigFloat(double d);
  explicit BigFloat(int i);
  // Destructor.
  ~BigFloat();

  // Assignemnt operator.
  BigFloat &operator=(const BigFloat &rhs);

  // Addition operator.
  BigFloat operator+(const BigFloat &rhs);
  // Subtraction operator.
  BigFloat operator-(const BigFloat &rhs);
  // Multiplication operator.
  BigFloat operator*(const BigFloat &rhs);
  // Division operator.
  BigFloat operator/(const BigFloat &rhs);

  // Comparison operators.
  bool operator<(const BigFloat &rhs);
  bool operator<=(const BigFloat &rhs);
  bool operator==(const BigFloat &rhs);
  bool operator!=(const BigFloat &rhs);
  bool operator>=(const BigFloat &rhs);
  bool operator>(const BigFloat &rhs);

  // Square root operator.
  static BigFloat sqrt(const BigFloat &bf);

  // String representation of the value represented by the instance of
  // this class.
  std::string ToString() const;

 private:
  // Internal representation of the arbitrary-precision value
  // represented by the instance of this class.
  mpfr_t value_;
};

}  // namespace dbug

#endif  // DBUG_COMMON_BIG_FLOAT_H_
