/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_COMMON_LOGGING_H_
#define DBUG_COMMON_LOGGING_H_ 1

#include <map>
#include <sstream>
#include <string>

#include "dbug/common/macros.h"
#include "dbug/common/system.h"

#define ASSERT_ON_CHANNEL(channel, condition, fmt, args...) do {        \
    dbug::Logging::Assert(__FILE__, __FUNCTION__, __LINE__,             \
                          channel, condition, #condition, fmt, ##args); \
  } while (0)

#define ERROR_ON_CHANNEL(channel, fmt, args...) do {            \
    dbug::Logging::Error(__FILE__, __FUNCTION__, __LINE__,      \
                         channel, fmt, ##args);                 \
  } while (0)

#define WARNING_ON_CHANNEL(channel, fmt, args...) do {          \
    dbug::Logging::Warning(__FILE__, __FUNCTION__, __LINE__,    \
                           channel, fmt, ##args);               \
  } while (0)

#define MESSAGE_ON_CHANNEL(channel, fmt, args...) do {          \
    dbug::Logging::Message(__FILE__, __FUNCTION__, __LINE__,    \
                           channel, fmt, ##args);               \
  } while (0)

#define DEBUG_ON_CHANNEL(channel, fmt, args...) do {            \
    dbug::Logging::Debug(__FILE__, __FUNCTION__, __LINE__,      \
                         channel, fmt, ##args);                 \
  } while (0)

#define DEBUG_EXTERNAL_ON_CHANNEL(channel, fmt1, fmt2, args...) do {    \
    dbug::Logging::External(__FILE__, __FUNCTION__, __LINE__,           \
                            channel, fmt1, fmt2, ##args);               \
  } while (0)

#define DEBUG_INTERNAL_ON_CHANNEL(channel, fmt1, fmt2, args...) do {    \
    dbug::Logging::Internal(__FILE__, __FUNCTION__, __LINE__,           \
                            channel, fmt1, fmt2, ##args);               \
  } while (0)

namespace dbug {

class Logging {
 public:
  static const int kMinNumberLength = 5;
  static const int kMaxStringLength = 20;
  static const int kMaxLocationLength = 40;

  enum Channel {
    CODEGEN,
    EXPLORER,
    ARBITER,
    INTERPOSITION,
    SYSTEM,
    UTIL
  };

  enum Verbosity {
    ERROR = 1,
    WARNING = 2,
    MESSAGE = 3,
    DEBUG = 4
  };

  class Config {
   public:
    Config(System::File *log_file, Verbosity verbosity);
    ~Config();

    // accessors
    System::File *log_file() const;
    Verbosity verbosity() const;

   private:
    System::File *log_file_;
    Verbosity verbosity_;

    DISALLOW_COPY_AND_ASSIGN(Config);
  };

  static void Initialize();
  static void Finalize();

  // accessors
  static const Config *get_channel(Channel channel);

  // mutators
  static void add_channel(Channel channel,
                          System::File *log_file,
                          Verbosity verbosity);

  static void Assert(const char *filename,
                     const char *function,
                     int line,
                     Channel channel,
                     bool condition,
                     const char *assert,
                     const char *format, ...);
  static void Debug(const char *filename,
                    const char *function,
                    int line,
                    Channel channel,
                    const char *format, ...);
  static void Error(const char *filename,
                    const char *function,
                    int line,
                    Channel channel,
                    const char *format, ...) DBUG_NORETURN;
  static void External(const char *filename,
                       const char *function,
                       int line,
                       Channel channel,
                       const char *format1,
                       const char *format2, ...);
  static void Internal(const char *filename,
                       const char *function,
                       int line,
                       Channel channel,
                       const char *format1,
                       const char *format2, ...);
  static void Message(const char *filename,
                      const char *function,
                      int line,
                      Channel channel,
                      const char *format, ...);
  static void Warning(const char *filename,
                      const char *function,
                      int line,
                      Channel channel,
                      const char *format, ...);

 private:
  static std::string FormatLocation(const char *filename,
                                    const char *function,
                                    int line);

  static std::map<Channel, Config *> *channels_;

  DISALLOW_ALL_DEFAULT(Logging);
};

}  // namespace dbug

#endif  // DBUG_COMMON_LOGGING_H_
