/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_EXPLORER_SERVER_H_
#define DBUG_EXPLORER_SERVER_H_ 1

#include <thrift/server/TServer.h>
#include <thrift/server/TThreadedServer.h>

#include <boost/scoped_ptr.hpp>
#include <list>
#include <queue>
#include <string>
#include <vector>

#include "dbug/common/command.h"
#include "dbug/common/macros.h"
#include "dbug/common/system.h"
#include "dbug/explorer/frontier.h"
#include "dbug/explorer/node.h"
#include "src/thrift/explorer/explorer_types.h"

using dbug::thrift::Move;
using apache::thrift::server::TThreadedServer;

namespace dbug {
namespace explorer {

class Server {
 public:
  class EventHandler : public ::apache::thrift::server::TServerEventHandler {
   public:
    explicit EventHandler(Server *server);
    virtual void preServe();

   private:
    Server *server_;

    DISALLOW_COPY_AND_ASSIGN(EventHandler);
  };

  class Lock {
   public:
    explicit Lock(Server *server);
    ~Lock();

   private:
    Server *server_;

    DISALLOW_COPY_AND_ASSIGN(Lock);
  };

  explicit Server(const char *config_path);
  ~Server();

  // accessors
  Node *execution_tree() const;
  Frontier *frontier() const;
  int port() const;
  std::vector<Move> prefix() const;
  bool use_dpor() const;

  // mutators
  void set_arbiter_finished(bool flag);
  void set_handler_started(bool flag);
  void increment_num_errors();
  void increment_process_started();
  void set_arbiter_started(bool flag);
  void set_thrift_server(TThreadedServer *server);

  void Configure();
  void Explore();
  void Stop();

 private:
  void CheckForFiles();
  void CleanUpLogs();
  void CollectLogs(int n);
  void ComputeEstimate();
  void WriteResults();

  Command *arbiter_;
  bool arbiter_finished_;
  int arbiter_port_;
  bool arbiter_started_;
  std::list<Command *> commands_;
  std::string config_path_;
  System::CVar cvar_;
  Node *execution_tree_;
  boost::scoped_ptr<Frontier> frontier_;
  bool handler_started_;
  std::string init_;
  std::string interposition_;
  std::string log_dir_;
  int max_iterations_;
  System::Mutex mutex_;
  int num_errors_;
  int num_iterations_;
  int num_processes_started_;
  int port_;
  std::vector<Move> prefix_;
  TThreadedServer *thrift_server_;
  int timeout_;
  bool use_dpor_;

  DISALLOW_COPY_AND_ASSIGN(Server);
};

}  // namespace explorer
}  // namespace dbug

#endif  // DBUG_EXPLORER_SERVER_H_
