/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/epoll.h>

#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/posix-resource.h"
#include "dbug/arbiter/request-prototypes.h"
#include "dbug/common/util.h"

namespace dbug {
namespace arbiter {

using posix::EPoll;
using posix::File;
using posix::Pipe;
using posix::Socket;

void EPollCreate::Execute() {
  server_->posix()->add_epoll(new EPoll(rid_, thread_->clock()));
  MESSAGE("Thread %d creates %s.", thread_->id(), rid_.c_str());
}

void EPollCreate::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  EPoll *epoll = server_->posix()->epoll_lookup(rid_);
  if (!epoll) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> EPollCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: creating " << rid_;
  result.push_back(stream.str());
  return result;
}

void EPollControl::Execute() {
  EPoll *epoll = server_->posix()->epoll_lookup(rid_);
  ASSERT(epoll, "Epoll does not exist.");
  EPoll::Event *event = epoll->get_event(event_.fd_rid);
  switch (operation_) {
    case EPOLL_CTL_ADD:
      ASSERT(!event, "Event to be added already exists.");
      event = new EPoll::Event(event_.events, event_.data);
      epoll->add_event(event_.fd_rid, event);
      MESSAGE("Thread %d adds %s to %s.",
              thread_->id(), event_.fd_rid.c_str(), rid_.c_str());
      break;
    case EPOLL_CTL_DEL:
      ASSERT(event, "Event to be deleted does not exist.");
      epoll->remove_event(event_.fd_rid);
      MESSAGE("Thread %d deletes %s from %s.",
              thread_->id(), event_.fd_rid.c_str(), rid_.c_str());
      break;
    case EPOLL_CTL_MOD:
      ASSERT(event, "Event to be modified does not exist.");
      event->set_events(event_.events);
      event->set_data(event_.data);
      MESSAGE("Thread %d modifies %s in %s.",
              thread_->id(), event_.fd_rid.c_str(), rid_.c_str());
      break;
    default:
      WARNING("Controling an epoll with an unknown operation.");
      break;
  }
  thread_->join_with_clock(epoll->clock());
  epoll->join_with_clock(thread_->clock());
}

void EPollControl::Test(Set requests, GroupSet *groups) {
  EPoll *epoll = server_->posix()->epoll_lookup(rid_);
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  if (epoll) {
    EPoll::Event *event = epoll->get_event(event_.fd_rid);
    switch (operation_) {
      case EPOLL_CTL_ADD:
        if (!event) { Enable(&group); }
        break;
      case EPOLL_CTL_MOD:
      case EPOLL_CTL_DEL:
        if (event) { Enable(&group); }
        break;
      default:
        WARNING("Controling an epoll with an unknown operation.");
        break;
    }
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> EPollControl::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: controling " << rid_;
  result.push_back(stream.str());
  return result;
}

void EPollDelete::Execute() {
  EPoll *epoll = server_->posix()->epoll_lookup(rid_);
  ASSERT(epoll, "Epoll does not exist.");
  server_->posix()->remove_epoll(epoll);
  delete epoll;
  MESSAGE("Thread %d deletes %s.", thread_->id(), rid_.c_str());
}

void EPollDelete::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  EPoll *epoll = server_->posix()->epoll_lookup(rid_);
  if (epoll) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> EPollDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: deleting " << rid_;
  result.push_back(stream.str());
  return result;
}

void EPollWait::Execute() {
  EPoll *epoll = server_->posix()->epoll_lookup(rid_);
  ASSERT(epoll, "Epoll does not exist.");
  std::list<std::string> events;
  std::vector<dbug::thrift::EPollEvent> result;
  for (std::pair<std::string, EPoll::Event *> pair : epoll->events()) {
    // TODO(jsimsa): Account for events that combine EPOLLIN and EPOLLOUT.
    std::string rid = pair.first;
    const EPoll::Event *event = pair.second;
    // File
    File *file = server_->posix()->file_lookup(rid);
    if (file) {
      // Merge clocks of the calling thread and the file.
      thread_->join_with_clock(file->clock());
      file->join_with_clock(thread_->clock());
      if ((event->events() & EPOLLIN) && file->IsReadable()) {
        dbug::thrift::EPollEvent thrift_event;
        thrift_event.fd_rid = rid;
        thrift_event.events = EPOLLIN;
        thrift_event.data = event->data();
        result.push_back(thrift_event);
        events.push_back(rid);
        if (result.size() == (unsigned) maxevents_) { break; }
        continue;
      }
      if ((event->events() & EPOLLOUT) && file->IsWritable()) {
        dbug::thrift::EPollEvent thrift_event;
        thrift_event.fd_rid = rid;
        thrift_event.events = EPOLLIN;
        thrift_event.data = event->data();
        result.push_back(thrift_event);
        events.push_back(rid);
        if (result.size() == (unsigned) maxevents_) { break; }
        continue;
      }
      continue;
    }
    // Pipe
    Pipe *pipe = server_->posix()->pipe_lookup(rid);
    if (pipe) {
      // Merge clocks of the calling thread and the pipe.
      thread_->join_with_clock(pipe->clock());
      pipe->join_with_clock(thread_->clock());
      if ((event->events() & EPOLLIN) && pipe->IsReadable()) {
        dbug::thrift::EPollEvent thrift_event;
        thrift_event.fd_rid = rid;
        thrift_event.events = EPOLLIN;
        thrift_event.data = event->data();
        result.push_back(thrift_event);
        events.push_back(rid);
        if (result.size() == (unsigned) maxevents_) { break; }
        continue;
      }
      if ((event->events() & EPOLLOUT) && pipe->IsWritable()) {
        dbug::thrift::EPollEvent thrift_event;
        thrift_event.fd_rid = rid;
        thrift_event.events = EPOLLIN;
        thrift_event.data = event->data();
        result.push_back(thrift_event);
        events.push_back(rid);
        if (result.size() == (unsigned) maxevents_) { break; }
        continue;
      }
      continue;
    }
    // Socket
    Socket *socket = server_->posix()->socket_lookup(rid);
    if (socket) {
      // Merge clocks of the calling thread and the local socket.
      thread_->join_with_clock(socket->clock());
      socket->join_with_clock(thread_->clock());
      if ((event->events() & EPOLLIN) && socket->IsReadable()) {
        // Join the clock of the calling thread with the clock of the
        // readable remote socket.
        Socket *remote_socket = socket->GetReadableRemote();
        if (remote_socket) { thread_->join_with_clock(remote_socket->clock()); }
        // Add this event to the set of events returned to the caller.
        dbug::thrift::EPollEvent thrift_event;
        thrift_event.fd_rid = rid;
        thrift_event.events = EPOLLIN;
        thrift_event.data = event->data();
        result.push_back(thrift_event);
        events.push_back(rid);
        if (result.size() == (unsigned) maxevents_) { break; }
        continue;
      }
      if (event->events() & EPOLLOUT && socket->IsWritable()) {
        // Add this event to the set of events returned to the caller.
        dbug::thrift::EPollEvent thrift_event;
        thrift_event.fd_rid = rid;
        thrift_event.events = EPOLLOUT;
        thrift_event.data = event->data();
        result.push_back(thrift_event);
        events.push_back(rid);
        if (result.size() == (unsigned) maxevents_) { break; }
        continue;
      }
      continue;
    }
    WARNING("Polling on an unknown resource %s.", rid.c_str());
  }
  return_value_->return_value = result;
  MESSAGE("Thread %d wakes up from epolling returning file descriptors %s.",
          thread_->id(), Util::ListToString<std::string>(events).c_str());
}

void EPollWait::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_read_set(rid_);
  EPoll *epoll = server_->posix()->epoll_lookup(rid_);
  if (epoll) {
    for (std::pair<std::string, EPoll::Event *> pair : epoll->events()) {
      std::string rid = pair.first;
      const EPoll::Event *event = pair.second;
      group.add_to_read_set(rid);
      // File
      File *file = server_->posix()->file_lookup(rid);
      if (file) {
        if ((event->events() & EPOLLIN) && file->IsReadable()) {
          Enable(&group);
          continue;
        }
        if ((event->events() & EPOLLOUT) && file->IsWritable()) {
          Enable(&group);
          continue;
        }
      }
      // Pipe
      Pipe *pipe = server_->posix()->pipe_lookup(rid);
      if (pipe) {
        if ((event->events() & EPOLLIN) && pipe->IsReadable()) {
          Enable(&group);
          continue;
        }
        if ((event->events() & EPOLLOUT) && pipe->IsWritable()) {
          Enable(&group);
          continue;
        }
      }
      // Socket
      Socket *socket = server_->posix()->socket_lookup(rid);
      if (socket) {
        if ((event->events() & EPOLLIN) && socket->IsReadable()) {
          Enable(&group);
          continue;
        }
        if ((event->events() & EPOLLOUT) && socket->IsWritable()) {
          Enable(&group);
          continue;
        }
      }
    }
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> EPollWait::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: epolling on " << rid_;
  result.push_back(stream.str());
  return result;
}

void FileClose::Execute() {
  File *file = server_->posix()->file_lookup(rid_);
  ASSERT(file, "File does not exist.");
  file->decrement_nrefs();
  if (file->nrefs() == 0) {
    // Temporary hack for multiple fork-exec-close of pipes.
    // server_->posix()->remove_file(file);
    // delete file;
  }
  MESSAGE("Thread %d closes %s.", thread_->id(), rid_.c_str());
}

void FileClose::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  File *file = server_->posix()->file_lookup(rid_);
  if (file) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> FileClose::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: closing " << rid_;
  result.push_back(stream.str());
  return result;
}

void FileOpen::Execute() {
  File *file = server_->posix()->file_lookup(rid_);
  if (!file) {
    file = new File(rid_, thread_->clock());
    server_->posix()->add_file(file);
  }
  file->increment_nrefs();
  MESSAGE("Thread %d opens %s.", thread_->id(), rid_.c_str());
}

void FileOpen::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> FileOpen::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: opening " << rid_;
  result.push_back(stream.str());
  return result;
}

void FileRead::Execute() {
  File *file = server_->posix()->file_lookup(rid_);
  ASSERT(file, "File does not exist.");
  // Merge clocks of the calling thread and the file.
  thread_->join_with_clock(file->clock());
  file->join_with_clock(thread_->clock());
  MESSAGE("Thread %d reads up to %d bytes from %s.",
          thread_->id(), nbytes_, rid_.c_str());
}

void FileRead::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  File *file = server_->posix()->file_lookup(rid_);
  if (file && file->IsReadable()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> FileRead::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: reading up to " << nbytes_ << " bytes from " << rid_;
  result.push_back(stream.str());
  return result;
}

void FileWrite::Execute() {
  File *file = server_->posix()->file_lookup(rid_);
  // Merge clocks of the calling thread and the file.
  thread_->join_with_clock(file->clock());
  file->join_with_clock(thread_->clock());
  MESSAGE("Thread %d writes up to %d bytes to %s.",
          thread_->id(), nbytes_, rid_.c_str());
}

void FileWrite::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  File *file = server_->posix()->file_lookup(rid_);
  if (file && file->IsWritable()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> FileWrite::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: writing up to " << nbytes_ << " bytes to " << rid_;
  result.push_back(stream.str());
  return result;
}

void PipeCreate::Execute() {
  std::stringstream stream;
  stream << "pipe-" << Pipe::GenerateID();
  std::string rid = stream.str();
  Pipe *pipe = new Pipe(rid, thread_->clock());
  server_->posix()->add_pipe(pipe);
  return_value_->return_value = rid;
  MESSAGE("Thread %d creates %s.", thread_->id(), rid.c_str());
}

void PipeCreate::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> PipeCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: creating a pipe";
  result.push_back(stream.str());
  return result;
}

void PipeDelete::Execute() {
  // TODO(jsimsa): Garbage collect the pipe resource.
  Pipe *pipe = server_->posix()->pipe_lookup(rid_);
  ASSERT(pipe, "Pipe does not exist.");
  if (pipe->output() == fd_) {
    MESSAGE("Thread %d closes read-end of %s.", thread_->id(), rid_.c_str());
  } else {
    if (pipe->input() == fd_) {
      MESSAGE("Thread %d closes write-end of %s.", thread_->id(), rid_.c_str());
    } else {
      WARNING("Thread %d uses incorrent file descriptor %d to close %s",
              thread_->id(), fd_, rid_.c_str());
    }
  }
}

void PipeDelete::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid_);
  if (pipe) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> PipeDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: deleting " << rid_;
  result.push_back(stream.str());
  return result;
}

void PipeRead::Execute() {
  Pipe *pipe = server_->posix()->pipe_lookup(rid_);
  ASSERT(pipe, "Pipe does not exist.");
  int nbytes = pipe->nbytes() > nbytes_ ? nbytes_ : pipe->nbytes();
  // Merge clocks of the calling thread and the local
  // socket. Further, join the clock of the calling thread with
  // the clock of the remote socket.
  thread_->join_with_clock(pipe->clock());
  pipe->join_with_clock(thread_->clock());
  MESSAGE("Thread %d reads up to %d bytes from %s.",
          thread_->id(), nbytes, rid_.c_str());
}

void PipeRead::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid_);
  if (pipe && pipe->IsReadable()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> PipeRead::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: reading up to " << nbytes_ << " bytes from " << rid_;
  result.push_back(stream.str());
  return result;
}

void PipeWrite::Execute() {
  Pipe *pipe = server_->posix()->pipe_lookup(rid_);
  // Merge clocks of the calling thread and the local socket.
  thread_->join_with_clock(pipe->clock());
  pipe->join_with_clock(thread_->clock());
  MESSAGE("Thread %d writes up to %d bytes to %s.",
          thread_->id(), nbytes_, rid_.c_str());
}

void PipeWrite::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid_);
  if (pipe && pipe->IsWritable()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> PipeWrite::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: writing up to " << nbytes_ << " bytes to " << rid_;
  result.push_back(stream.str());
  return result;
}

void Select::Execute() {
  std::set<std::string> rfds, wfds, efds;
  // Write file descriptors.
  for (const std::string &rid : wfds_) {
    // File
    File *file = server_->posix()->file_lookup(rid);
    if (file) {
      thread_->join_with_clock(file->clock());
      file->join_with_clock(thread_->clock());
      if (file->IsWritable() || (!timeout_ && !file->IsBlocking())) {
        wfds.insert(rid);
      }
      continue;
    }
    // Pipe
    Pipe *pipe = server_->posix()->pipe_lookup(rid);
    if (pipe) {
      thread_->join_with_clock(pipe->clock());
      pipe->join_with_clock(thread_->clock());
      if (pipe->IsWritable() || (!timeout_ && !pipe->IsBlocking())) {
        wfds.insert(rid);
      }
      continue;
    }
    // Socket
    Socket *socket = server_->posix()->socket_lookup(rid);
    if (socket) {
      thread_->join_with_clock(socket->clock());
      socket->join_with_clock(thread_->clock());
      if (socket->IsWritable() || (!timeout_ && !socket->IsBlocking())) {
        wfds.insert(rid);
      }
      continue;
    }
    WARNING("Polling on an unknown resource %s.", rid.c_str());
  }
  // Read file descriptors
  for (const std::string &rid : rfds_) {
    // File
    File *file = server_->posix()->file_lookup(rid);
    if (file) {
      thread_->join_with_clock(file->clock());
      file->join_with_clock(thread_->clock());
      if (file->IsReadable() || (!timeout_ && !file->IsBlocking())) {
        rfds.insert(rid);
      }
      continue;
    }
    // Pipe
    Pipe *pipe = server_->posix()->pipe_lookup(rid);
    if (pipe) {
      thread_->join_with_clock(pipe->clock());
      pipe->join_with_clock(thread_->clock());
      if (pipe->IsReadable() || (!timeout_ && !pipe->IsBlocking())) {
        rfds.insert(rid);
      }
      continue;
    }
    // Socket
    Socket *socket = server_->posix()->socket_lookup(rid);
    if (socket) {
      thread_->join_with_clock(socket->clock());
      socket->join_with_clock(thread_->clock());
      if (socket->IsReadable() || (!timeout_ && !socket->IsBlocking())) {
        // Further, join the clock of the calling thread with the
        // clock of the readable remote socket.
        Socket *remote_socket = socket->GetReadableRemote();
        if (remote_socket) { thread_->join_with_clock(remote_socket->clock()); }
        rfds.insert(rid);
      }
      continue;
    }
    WARNING("Polling on an unknown resource %s.", rid.c_str());
  }
  return_value_->return_value.rfds = rfds;
  return_value_->return_value.wfds = wfds;
  MESSAGE("Thread %d wakes up from polling returning "
          "read file descriptors %s and write file descriptors %s.",
          thread_->id(),
          Util::SetToString<std::string>(rfds).c_str(),
          Util::SetToString<std::string>(wfds).c_str());
}

void Select::Test(Set requests, GroupSet *groups) {
  bool enabled = false;
  Request::Group group;
  group.add_request(this);
  // Read file descriptors
  for (const std::string &rid : rfds_) {
    group.add_to_read_set(rid);
    // File
    File *file = server_->posix()->file_lookup(rid);
    if (file && (file->IsReadable() || !file->IsBlocking())) {
      enabled = true;
      continue;
    }
    // Pipe
    Pipe *pipe = server_->posix()->pipe_lookup(rid);
    if (pipe && (pipe->IsReadable() || !pipe->IsBlocking())) {
      enabled = true;
      continue;
    }
    // Socket
    Socket *socket = server_->posix()->socket_lookup(rid);
    if (socket && (socket->IsReadable() || !socket->IsBlocking())) {
      enabled = true;
      continue;
    }
  }
  // Write file descriptors
  for (const std::string &rid : wfds_) {
    group.add_to_read_set(rid);
    // File
    File *file = server_->posix()->file_lookup(rid);
    if (file && (file->IsWritable() || !file->IsBlocking())) {
      enabled = true;
      continue;
    }
    // Pipe
    Pipe *pipe = server_->posix()->pipe_lookup(rid);
    if (pipe && (pipe->IsWritable() || !pipe->IsBlocking())) {
      enabled = true;
      continue;
    }
    // Socket
    Socket *socket = server_->posix()->socket_lookup(rid);
    if (socket && (socket->IsWritable() || !socket->IsBlocking())) {
      enabled = true;
      continue;
    }
  }
  if (enabled || timeout_) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> Select::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: ";
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "    polling with read file descriptors "
         << Util::SetToString<std::string>(rfds_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "    polling with write file descriptors "
         << Util::SetToString<std::string>(wfds_);
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

void SocketAccept::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  Socket *remote_socket = socket->remove_pending_connection();
  if (remote_socket) {
    Socket *new_socket =
        new Socket(new_rid_,
                   socket->domain(),
                   socket->type(),
                   socket->protocol(),
                   thread_->clock());
    server_->posix()->add_socket(new_socket);
    // Match <new_socket> with a pending connection or <socket>.
    remote_socket->set_connection(new_socket);
    new_socket->set_connection(remote_socket);
    // Merge the clocks of the thread and the pair of sockets.
    thread_->join_with_clock(socket->clock());
    socket->set_clock(thread_->clock());
    thread_->join_with_clock(new_socket->clock());
    new_socket->set_clock(thread_->clock());
    MESSAGE("Thread %d accepts a connection on %s using %s.",
            thread_->id(), rid_.c_str(), new_rid_.c_str());
  } else {
    MESSAGE("Thread %d cannot accept a connection on %s using %s.",
            thread_->id(), rid_.c_str(), new_rid_.c_str());
  }
}

void SocketAccept::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  group.add_to_write_set(new_rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket) {
    std::list<Socket *> pending = socket->pending_connections();
    if (!pending.empty()) {
      Socket *remote_socket = pending.front();
      group.add_to_write_set(remote_socket->id());
      Enable(&group);
    }
    if (!socket->IsBlocking()) { Enable(&group); }
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketAccept::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: accepting connection from "
         << rid_ << " using " << new_rid_;
  result.push_back(stream.str());
  return result;
}

void SocketBind::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  socket->set_address(address_);
  socket->set_port(port_);
  thread_->join_with_clock(socket->clock());
  socket->set_clock(thread_->clock());
  MESSAGE("Thread %d binds %s to %s:%d.",
          thread_->id(), rid_.c_str(), address_.c_str(), port_);
}

void SocketBind::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketBind::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: binding " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketClose::Execute() {
  // TODO(jsimsa): Garbage collect the socket resource.
  MESSAGE("Thread %d closes %s.", thread_->id(), rid_.c_str());
}

void SocketClose::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketClose::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: closing " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketConnect::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  Socket *remote_socket =
      server_->posix()->remote_socket_lookup(socket->node_id(),
                                             address_,
                                             port_);
  if (remote_socket) {
    // Merge the clocks of the thread and the connected sockets.
    thread_->join_with_clock(socket->clock());
    thread_->join_with_clock(remote_socket->clock());
    socket->set_clock(thread_->clock());
    remote_socket->set_clock(thread_->clock());
    if (socket->type() == SOCK_STREAM) {
      if (remote_socket->listening()) {
        remote_socket->add_pending_connection(socket);
        MESSAGE("Thread %d connects %s to %s:%d.",
                thread_->id(), rid_.c_str(), address_.c_str(), port_);
      } else {
        WARNING("Thread %d fails to connect %s to %s:%d.",
                thread_->id(), rid_.c_str(), address_.c_str(), port_);
      }
    }
    if (socket->type() == SOCK_DGRAM) {
      MESSAGE("Thread %d connects %s to %s:%d.",
              thread_->id(), rid_.c_str(), address_.c_str(), port_);
    }
  } else {
    MESSAGE("Thread %d fails to connect %s to %s:%d.",
            thread_->id(), rid_.c_str(), address_.c_str(), port_);
  }
}

void SocketConnect::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket) {
    Socket *remote_socket =
        server_->posix()->remote_socket_lookup(socket->node_id(),
                                               address_,
                                               port_);
    if (remote_socket) { group.add_to_write_set(remote_socket->id()); }
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketConnect::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: connecting to " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketListen::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  socket->set_listening(true);
  thread_->join_with_clock(socket->clock());
  socket->set_clock(thread_->clock());
  MESSAGE("Thread %d listens on %s.", thread_->id(), rid_.c_str());
}

void SocketListen::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketListen::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: listening on " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketOpen::Execute() {
  Socket *socket =
      new Socket(rid_, domain_, type_, protocol_, thread_->clock());
  server_->posix()->add_socket(socket);
  MESSAGE("Thread %d opens %s.", thread_->id(), rid_.c_str());
}

void SocketOpen::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (!socket) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketOpen::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: opening " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketPair::Execute() {
  Socket *socket1 =
      new Socket(rid1_, domain_, type_, protocol_, thread_->clock());
  server_->posix()->add_socket(socket1);
  Socket *socket2 =
      new Socket(rid2_, domain_, type_, protocol_, thread_->clock());
  server_->posix()->add_socket(socket2);
  socket1->set_connection(socket2);
  socket2->set_connection(socket1);
  MESSAGE("Thread %d creates socket pair %s:%s.",
          thread_->id(), rid1_.c_str(), rid2_.c_str());
}

void SocketPair::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid1_);
  group.add_to_write_set(rid2_);
  Socket *socket1 = server_->posix()->socket_lookup(rid1_);
  Socket *socket2 = server_->posix()->socket_lookup(rid2_);
  if (!socket1 && !socket2) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketPair::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: creating a socket pair " << rid1_ << ":" << rid2_;
  result.push_back(stream.str());
  return result;
}

void SocketReceive::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  thread_->join_with_clock(socket->clock());
  socket->join_with_clock(thread_->clock());
  Socket *remote_socket = socket->connection();
  if (remote_socket) {
    int nbytes = (remote_socket->nbytes() > nbytes_ ?
                  nbytes_ : remote_socket->nbytes());
    thread_->join_with_clock(remote_socket->clock());
    MESSAGE("Thread %d reads up to %d bytes from %s.",
            thread_->id(), nbytes, rid_.c_str());
  } else {
    MESSAGE("Thread %d fails to read from %s.",
            thread_->id(), rid_.c_str());
  }
}

void SocketReceive::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket && socket->IsReadable()) {
    Socket *remote_socket = socket->GetReadableRemote();
    if (remote_socket) {
      group.add_to_read_set(remote_socket->id());
    }
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketReceive::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: reading up to " << nbytes_ << " bytes from " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketReceiveFrom::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  thread_->join_with_clock(socket->clock());
  socket->join_with_clock(thread_->clock());
  Socket *remote_socket = socket->remove_pending_message();
  if (remote_socket) {
    int nbytes = (remote_socket->nbytes() > nbytes_ ?
                  nbytes_ : remote_socket->nbytes());
    thread_->join_with_clock(remote_socket->clock());
    MESSAGE("Thread %d reads up to %d bytes from %s.",
            thread_->id(), nbytes, rid_.c_str());
  } else {
    MESSAGE("Thread %d fails to read from %s.",
            thread_->id(), rid_.c_str());
  }
}

void SocketReceiveFrom::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_read_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket && socket->IsReadable()) {
    // TODO(jsimsa): Add the resource ID of the remote end.
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketReceiveFrom::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: reading up to " << nbytes_ << " bytes from " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketSend::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  // Merge clocks of the calling thread and the local socket.
  thread_->join_with_clock(socket->clock());
  socket->join_with_clock(thread_->clock());
  MESSAGE("Thread %d writes up to %d bytes to %s.",
          thread_->id(), nbytes_, rid_.c_str());
}

void SocketSend::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket && socket->IsWritable()) {
    Socket *remote_socket = socket->connection();
    if (remote_socket) { group.add_to_write_set(remote_socket->id()); }
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketSend::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: writing up to " << nbytes_ << " bytes to " << rid_;
  result.push_back(stream.str());
  return result;
}

void SocketSendTo::Execute() {
  Socket *socket = server_->posix()->socket_lookup(rid_);
  ASSERT(socket, "Socket does not exist.");
  Socket *remote_socket =
      server_->posix()->remote_socket_lookup(socket->node_id(),
                                             address_,
                                             port_);
  ASSERT(remote_socket, "Unable to find remote communication end-point.");
  remote_socket->add_pending_message(socket);
  // Merge the clocks of the thread and the connected sockets.
  thread_->join_with_clock(socket->clock());
  thread_->join_with_clock(remote_socket->clock());
  socket->set_clock(thread_->clock());
  remote_socket->set_clock(thread_->clock());
  MESSAGE("Thread %d writes up to %d bytes to %s.",
          thread_->id(), nbytes_, rid_.c_str());
}

void SocketSendTo::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Socket *socket = server_->posix()->socket_lookup(rid_);
  if (socket && socket->IsWritable()) {
    // TODO(jsimsa): Add the resource ID of the remote end.
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SocketSendTo::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: writing up to " << nbytes_ << " bytes to " << rid_;
  result.push_back(stream.str());
  return result;
}

}  // namespace arbiter
}  // namespace dbug
