/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/mpi-event.h"

#include <mpi.h>

#include <list>
#include <map>
#include <set>
#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/mpi.h"
#include "dbug/arbiter/request.h"
#include "dbug/arbiter/thread.h"
#include "dbug/common/util.h"

namespace dbug {
namespace arbiter {
namespace mpi {

// Event class

bool Event::PointerComparison::operator()(const Event *e1, const Event *e2) {
  ASSERT(e1, "Event does not exist.");
  ASSERT(e2, "Event does not exist.");
  return e1->tid() < e2->tid();
}

Event::Event(Request *request, int communicator, int origin)
    : request_(request), communicator_(communicator),
      origin_(origin), state_(AWAITING) {
  tid_ = request->thread()->id();
  clock_.JoinWith(request->thread()->clock());
}

Event::~Event() {}

// accessors

int Event::tid() const { return tid_; }

Clock Event::clock() const { return clock_; }

Request *Event::request() const { return request_; }

int Event::communicator() const { return communicator_; }

int Event::origin() const { return origin_; }

Event::State Event::state() const { return state_; }

// mutators

void Event::join_with_clock(const Clock &clock) { clock_.JoinWith(clock); }

void Event::set_clock(const Clock &clock) { clock_ = clock; }

void Event::set_request(Request *request) { request_ = request; }

void Event::set_state(State state) { state_ = state; }

// public methods

std::string Event::ToString() const {
  std::stringstream stream;
  stream << " request = " << request_
         << " clock = " << Util::VectorToString<int>(clock_)
         << " communicator = " << communicator_
         << " origin = " << origin_
         << " state = " << state_;
  return stream.str();
}

// MatchedEvent class

MatchedEvent::MatchedEvent() {}

MatchedEvent::~MatchedEvent() {}

// Recv class

Recv::Recv(Request *request, int communicator, int origin, int source, int tag)
    : Event(request, communicator, origin),
      source_(source),
      tag_(tag),
      match_(NULL),
      parent_(NULL) {}

// accessors

int Recv::source() const { return source_; }

int Recv::tag() const { return tag_; }

Event *Recv::match() const { return match_; }

Event *Recv::parent() const { return parent_; }

// mutators

void Recv::set_match(Event *match) { match_ = match; }

void Recv::set_parent(Event *parent) { parent_ = parent; }

// pubic methods

MatchedEvent *Recv::ExtractMatchedEvent() { return NULL; }

Recv *Recv::ExtractRecv() { return this; }

Send *Recv::ExtractSend() { return NULL; }

void Recv::Match(const MPI::EventOriginMap &origin_map,
                 const std::vector<int> &nondet_choices) {
  int index = 0;
  InternalMatch(origin_map, nondet_choices, this, &index);
}

mpi::Event *Recv::ProbeMatch(const MPI::EventOriginMap &origin_map,
                             const std::vector<int> &nondet_choices) {
  Clock clock;
  return InternalProbeMatch(origin_map, nondet_choices, this, 0, &clock);
}

void Recv::TestMatch(const MPI::EventOriginMap &origin_map,
                     std::set<Option> *options) {
  if (match_ == NULL) {
    Request::Set empty_set;
    std::vector<int> empty_vector;
    InternalTestMatch(origin_map,
                      options,
                      &empty_set,
                      &empty_vector,
                      this);
  } else {
    Option option;
    InternalComputeCoscheduleSet(this, &option.first);
    options->insert(option);
  }
}

std::string Recv::ToString() const {
  std::stringstream stream;
  stream << "Receive"
         << Event::ToString()
         << " source = " << source_
         << " tag = " << tag_;
  return stream.str();
}

// private methods

std::vector<Send *> Recv::GenerateMatches(
    const MPI::EventOriginMap &origin_map) {
  std::set<Send *, PointerComparison> matches;
  for (const std::pair<int, MPI::Events> &pair : origin_map) {
    MPI::Events events = pair.second;
    for (Event *event : events) {
      Send *send_event = event->ExtractSend();
      if (send_event != NULL &&
          send_event->match() == NULL &&
          send_event->destination() == origin_ &&
          (source_ == MPI_ANY_SOURCE ||
           source_ == send_event->origin()) &&
          (tag_ == MPI_ANY_TAG ||
           tag_ == send_event->tag())) {
        std::set<Send *, PointerComparison>::iterator it = matches.begin();
        bool valid_match = true;
        while (it != matches.end()) {
          Send *handle = *it;
          Clock::Order order = handle->clock().CompareWith(send_event->clock());
          switch (order) {
            case Clock::EQUAL:
              ASSERT(0, "Encountered events with identical clocks.");
              break;
            case Clock::AFTER:
              it = matches.erase(it);
              break;
            case Clock::BEFORE:
              valid_match = false;
              it = matches.end();
              break;
            default:
              ++it;
          }
        }
        if (valid_match) { matches.insert(send_event); }
      }
    }
  }
  std::vector<Send *> result(matches.begin(), matches.end());
  return result;
}

void Recv::InternalComputeCoscheduleSet(const Event *first,
                                        Request::Set *coschedule_set) {
  if (match_) {
    mpi::Send *send_event = match_->ExtractSend();
    ASSERT(send_event, "Receive is not matched with a send.");
    if (send_event->state() == Event::AWAITING) {
      coschedule_set->insert(send_event->request());
    }
    if (send_event->parent() != NULL && send_event->parent() != first) {
      mpi::Recv *recv_event = send_event->parent()->ExtractRecv();
      ASSERT(recv_event, "Send-receive is missing the receive event.");
      recv_event->InternalComputeCoscheduleSet(first, coschedule_set);
    }
  }
}

void Recv::InternalMatch(const MPI::EventOriginMap &origin_map,
                         const std::vector<int> &nondet_choices,
                         const Event *first,
                         int *index) {
  // Check if match has already been found.
  if (match_ != NULL) { return; }
  // Check if the remote end-point exists.
  if (source_ == MPI_PROC_NULL) { return; }
  // Check all events for a possible match.
  ASSERT((unsigned) *index < nondet_choices.size(),
         "Vector of nondeterministic choices is too short.");
  std::vector<Send *> matches = GenerateMatches(origin_map);
  ASSERT((unsigned) nondet_choices[*index] < matches.size(),
         "Vector of matches is too short.");
  Send *send_event = matches[nondet_choices[*index]];
  match_ = send_event;
  send_event->set_match(this);
  *index += 1;
  if (!send_event->parent()) {
    // The matched send event is not part of a send-receive
    // event and the matching has succeeded.
    send_event->join_with_clock(clock_);
    clock_.JoinWith(send_event->clock());
  } else {
    // The matched send event is part of a send-receive event. If
    // this is the same send-receive event that started the
    // matching process, matching has succeeded.
    if (first == send_event->parent()) { return; }
    // Finally, match the receieve portion of the matched send-receive
    // event.
    Recv *recv_event = send_event->parent()->ExtractRecv();
    recv_event->join_with_clock(clock_);
    recv_event->InternalMatch(origin_map, nondet_choices, first, index);
    clock_.JoinWith(recv_event->clock());
    recv_event->parent()->join_with_clock(clock_);
  }
}

mpi::Event *Recv::InternalProbeMatch(const MPI::EventOriginMap &origin_map,
                                     const std::vector<int> &nondet_choices,
                                     const Event *first,
                                     int index,
                                     Clock *clock) {
  ASSERT(match_ == NULL, "Probing a matched event.");
  // Check if the remote end-point exists.
  if (source_ == MPI_PROC_NULL) { return NULL; }
  // Check all events for a possible match.
  ASSERT((unsigned) index < nondet_choices.size(),
         "Vector of nondeterministic choices is too short.");
  std::vector<Send *> matches = GenerateMatches(origin_map);
  ASSERT((unsigned) nondet_choices[index] < matches.size(),
         "Vector of matches is too short.");
  Send *send_event = matches[nondet_choices[index]];
  clock->JoinWith(send_event->clock());
  if (send_event->parent()) {
    if (first != send_event->parent()) {
      match_ = send_event;
      send_event->set_match(this);
      Recv *recv_event = send_event->parent()->ExtractRecv();
      recv_event->InternalProbeMatch(origin_map,
                                     nondet_choices,
                                     first,
                                     index++,
                                     clock);
      send_event->set_match(NULL);
      match_ = NULL;
    }
  }
  if (index == 0) { clock_.JoinWith(*clock); }
  return send_event;
}

void Recv::InternalTestMatch(const MPI::EventOriginMap &origin_map,
                             std::set<Option> *options,
                             Request::Set *coschedule_set,
                             std::vector<int> *nondet_choices,
                             const Event *first) {
  // Check if match has already been found.
  if (match_ != NULL) {
    Option option;
    option.first = *coschedule_set;
    option.second = *nondet_choices;
    options->insert(option);
    return;
  }
  // Check if the remote end-point exists.
  if (source_ == MPI_PROC_NULL) {
    Option option;
    option.first = *coschedule_set;
    option.second = *nondet_choices;
    options->insert(option);
    match_ = NULL;
    return;
  }
  std::vector<Send *> matches = GenerateMatches(origin_map);
  for (unsigned i = 0; i < matches.size(); i++) {
    Send *send_event = matches[i];
    if (!send_event->parent()) {
      // The matched send event is not part of a send-receive
      // event and the matching has succeeded.
      Option option;
      option.first = *coschedule_set;
      if (send_event->state() == Event::AWAITING) {
        option.first.insert(send_event->request());
      }
      option.second = *nondet_choices;
      option.second.push_back(i);
      options->insert(option);
      continue;
    } else {
      Recv *recv_event = send_event->parent()->ExtractRecv();
      // The matched send event is part of a send-receive event. If
      // this is the same send-receive event that started the
      // matching process, matching has succeeded.
      if (first == send_event->parent()) {
        Option option;
        option.first = *coschedule_set;
        if (send_event->state() == Event::AWAITING) {
          option.first.insert(send_event->request());
        }
        option.second = *nondet_choices;
        option.second.push_back(i);
        options->insert(option);
        continue;
      }
      // Finally, try to match the receieve portion of the matched
      // send-receive event.
      if (send_event->state() == Event::AWAITING) {
        coschedule_set->insert(send_event->request());
      }
      nondet_choices->push_back(i);
      match_ = send_event;
      send_event->set_match(this);
      recv_event->InternalTestMatch(origin_map,
                                    options,
                                    coschedule_set,
                                    nondet_choices,
                                    first);
      match_ = NULL;
      send_event->set_match(NULL);
      nondet_choices->pop_back();
      if (send_event->state() == Event::AWAITING) {
        coschedule_set->erase(send_event->request());
      }
    }
  }
}

// MatchedRecv class

MatchedRecv::MatchedRecv(Request *request,
                         int communicator,
                         int origin,
                         int source,
                         int tag,
                         std::string handle)
    : Recv(request, communicator, origin, source, tag),
      handle_(handle) {}

// accessors

std::string MatchedRecv::handle() const { return handle_; }

MatchedEvent *MatchedRecv::ExtractMatchedEvent() { return this; }

// public methods

std::string MatchedRecv::ToString() const {
  std::stringstream stream;
  stream << Recv::ToString()
         << " handle = " << handle_;
  return stream.str();
}

// Send class

Send::Send(Request *request,
           int communicator,
           int origin,
           int nbytes,
           int destination,
           int tag,
           Type type)
    : Event(request, communicator, origin),
      nbytes_(nbytes),
      destination_(destination),
      tag_(tag),
      type_(type),
      match_(NULL),
      parent_(NULL) {}

// accessors

int Send::nbytes() const { return nbytes_; }

int Send::destination() const { return destination_; }

int Send::tag() const { return tag_; }

Send::Type Send::type() const { return type_; }

Event *Send::match() const { return match_; }

Event *Send::parent() const { return parent_; }

// mutators

void Send::set_type(Type type) { type_ = type; }

void Send::set_match(Event *match) { match_ = match; }

void Send::set_parent(Event *parent) { parent_ = parent; }

// public methods

MatchedEvent *Send::ExtractMatchedEvent() { return NULL; }

Recv *Send::ExtractRecv() { return NULL; }

Send *Send::ExtractSend() { return this; }

void Send::Match(const MPI::EventOriginMap &origin_map,
                 const std::vector<int> &nondet_choices) {
  int index = 0;
  return InternalMatch(origin_map, nondet_choices, this, &index);
}

void Send::TestMatch(const MPI::EventOriginMap &origin_map,
                     std::set<Option> *options) {
  if (match_ == NULL) {
    Request::Set empty_set;
    std::vector<int> empty_vector;
    InternalTestMatch(origin_map,
                      options,
                      &empty_set,
                      &empty_vector,
                      this);
  } else {
    Option option;
    InternalComputeCoscheduleSet(this, &option.first);
    options->insert(option);
  }
}

std::string Send::ToString() const {
  std::stringstream stream;
  stream << "Send"
         << Event::ToString()
         << " nbytes = " << nbytes_
         << " destination = " << destination_
         << " tag = " << tag_;
  return stream.str();
}

// private methods

std::vector<Recv *> Send::GenerateMatches(
    const MPI::EventOriginMap &origin_map) {
  std::set<Recv *, PointerComparison> matches;
  for (std::pair<int, std::list<Event *> > pair : origin_map) {
    std::list<Event *> events = pair.second;
    for (Event *event : events) {
      Recv *recv_event = event->ExtractRecv();
      if (recv_event != NULL &&
          recv_event->match() == NULL &&
          recv_event->origin() == destination_ &&
          (recv_event->source() == MPI_ANY_SOURCE ||
           recv_event->source() == origin_) &&
          (recv_event->tag() == MPI_ANY_TAG ||
           recv_event->tag() == tag_)) {
        std::set<Recv *, PointerComparison>::iterator it = matches.begin();
        bool valid_match = true;
        while (it != matches.end()) {
          Recv *handle = *it;
          Clock::Order order = handle->clock().CompareWith(recv_event->clock());
          switch (order) {
            case Clock::EQUAL:
              ASSERT(0, "Encountered events with identical clocks.");
              break;
            case Clock::AFTER:
              it = matches.erase(it);
              break;
            case Clock::BEFORE:
              valid_match = false;
              it = matches.end();
              break;
            default:
              ++it;
          }
        }
        if (valid_match) { matches.insert(recv_event); }
      }
    }
  }
  std::vector<Recv *> result(matches.begin(), matches.end());
  return result;
}

void Send::InternalComputeCoscheduleSet(const Event *first,
                                        Request::Set *coschedule_set) {
  if (match_ != NULL) {
    mpi::Recv *recv_event = match_->ExtractRecv();
    ASSERT(recv_event, "Send is not matched with a receive.");
    if (recv_event->state() == Event::AWAITING) {
      coschedule_set->insert(recv_event->request());
    }
    if (recv_event->parent() != NULL && recv_event->parent() != first) {
      mpi::Send *send_event = recv_event->parent()->ExtractSend();
      ASSERT(send_event, "Send-receive is missing the send event.");
      send_event->InternalComputeCoscheduleSet(first, coschedule_set);
    }
  }
}

void Send::InternalMatch(const MPI::EventOriginMap &origin_map,
                         const std::vector<int> &nondet_choices,
                         const Event *first,
                         int *index) {
  // Check if match has already been found.
  if (match_ != NULL) { return; }
  // Check if the remote end-point exists.
  if (destination_ == MPI_PROC_NULL) { return ;}
  // Check all events for a possible match.
  ASSERT((unsigned) *index < nondet_choices.size(),
         "Vector of nondeterministic choices is too short.");
  std::vector<Recv *> matches = GenerateMatches(origin_map);
  ASSERT((unsigned) nondet_choices[*index] < matches.size(),
         "Vector of matches is too short.");
  Recv *recv_event = matches[nondet_choices[*index]];
  match_ = recv_event;
  recv_event->set_match(this);
  if (!recv_event->parent()) {
    // The matched send event is not part of a send-receive
    // event and the matching has succeeded.
    recv_event->join_with_clock(clock_);
    clock_.JoinWith(recv_event->clock());
  } else {
    // The matched send event is part of a send-receive event. If
    // this is the same send-receive event that started the
    // matching process, matching has succeeded.
    if (first == recv_event->parent()) { return; }
    // Finally, try to match the receieve portion of the matched
    // send-receive event.
    Send *send_event = recv_event->parent()->ExtractSend();
    send_event->join_with_clock(clock_);
    *index += 1;
    send_event->InternalMatch(origin_map, nondet_choices, first, index);
    clock_.JoinWith(send_event->clock());
    send_event->parent()->join_with_clock(clock_);
  }
}

void Send::InternalTestMatch(const MPI::EventOriginMap &origin_map,
                             std::set<Option> *options,
                             Request::Set *coschedule_set,
                             std::vector<int> *nondet_choices,
                             const Event *first) {
  // Check if match has already been found.
  if (match_ != NULL) {
    Option option;
    option.first = *coschedule_set;
    option.second = *nondet_choices;
    options->insert(option);
    return;
  }
  // Check if the remote end-point exists.
  if (destination_ == MPI_PROC_NULL) {
    Option option;
    option.first = *coschedule_set;
    option.second = *nondet_choices;
    options->insert(option);
    match_ = NULL;
    return;
  }
  std::vector<Recv *> matches = GenerateMatches(origin_map);
  for (unsigned i = 0; i < matches.size(); i++) {
    Recv *recv_event = matches[i];
    if (!recv_event->parent()) {
      // The matched send event is not part of a send-receive
      // event and the matching has succeeded.
      Option option;
      option.first = *coschedule_set;
      if (recv_event->state() == Event::AWAITING) {
        option.first.insert(recv_event->request());
      }
      option.second = *nondet_choices;
      option.second.push_back(i);
      options->insert(option);
      continue;
    } else {
      // The matched send event is part of a send-receive event. If
      // this is the same send-receive event that started the
      // matching process, matching has succeeded.
      if (first == recv_event->parent()) {
        Option option;
        option.first = *coschedule_set;
        if (recv_event->state() == Event::AWAITING) {
          option.first.insert(recv_event->request());
        }
        option.second = *nondet_choices;
        option.second.push_back(i);
        options->insert(option);
        continue;
      }
      // Finally, try to match the receieve portion of the matched
      // send-receive event.
      if (recv_event->state() == Event::AWAITING) {
        coschedule_set->insert(recv_event->request());
      }
      nondet_choices->push_back(i);
      match_ = recv_event;
      recv_event->set_match(this);
      Send *send_event = recv_event->parent()->ExtractSend();
      send_event->InternalTestMatch(origin_map,
                                    options,
                                    coschedule_set,
                                    nondet_choices,
                                    first);
      match_ = NULL;
      recv_event->set_match(NULL);
      nondet_choices->pop_back();
      if (recv_event->state() == Event::AWAITING) {
        coschedule_set->erase(recv_event->request());
      }
    }
  }
}

// SendRecv class

SendRecv::SendRecv(Request *request,
                   int communicator,
                   int origin,
                   int nbytes,
                   int destination,
                   int sendtag,
                   int source,
                   int recvtag)
    : Event(request, communicator, origin),
      recv_(new Recv(request,
                     communicator,
                     origin,
                     source,
                     recvtag)),
      send_(new Send(request,
                     communicator,
                     origin,
                     nbytes,
                     destination,
                     sendtag,
                     Send::SYNCHRONOUS)) {
  recv_->set_parent(this);
  send_->set_parent(this);
}

SendRecv::~SendRecv() {
  delete send_;
  delete recv_;
}

// public methods

MatchedEvent *SendRecv::ExtractMatchedEvent() { return NULL; }

Recv *SendRecv::ExtractRecv() { return recv_; }

Send *SendRecv::ExtractSend() { return send_; }

void SendRecv::Match(const MPI::EventOriginMap &origin_map,
                     const std::vector<int> &nondet_choices) {
  int index = 0;
  recv_->InternalMatch(origin_map, nondet_choices, this, &index);
  send_->InternalMatch(origin_map, nondet_choices, this, &index);
}

void SendRecv::TestMatch(const MPI::EventOriginMap &origin_map,
                         std::set<Option> * options) {
  if (!recv_->match() || !send_->match()) {
    ASSERT(!recv_->match(), "Partially matched send-receive.");
    ASSERT(!send_->match(), "Partially matched send-receive.");
    std::set<Option> recv_options, send_options;
    Request::Set empty_set;
    std::vector<int> empty_vector;
    recv_->InternalTestMatch(origin_map,
                             &recv_options,
                             &empty_set,
                             &empty_vector,
                             this);
    send_->InternalTestMatch(origin_map,
                             &send_options,
                             &empty_set,
                             &empty_vector,
                             this);
    for (const Option &recv_option : recv_options) {
      for (const Option &send_option : send_options) {
        Option option;
        option.first.insert(recv_option.first.begin(),
                            recv_option.first.end());
        option.first.insert(send_option.first.begin(),
                            send_option.first.end());
        option.first.erase(request_);
        option.second.insert(option.second.end(),
                             recv_option.second.begin(),
                             recv_option.second.end());
        option.second.insert(option.second.end(),
                             send_option.second.begin(),
                             send_option.second.end());
        options->insert(option);
      }
    }
  } else {
    Option option;
    recv_->InternalComputeCoscheduleSet(this, &option.first);
    send_->InternalComputeCoscheduleSet(this, &option.first);
    option.first.erase(request_);
    options->insert(option);
  }
}

std::string SendRecv::ToString() const {
  std::stringstream stream;
  stream << Event::ToString()
         << " destination = " << send_->destination()
         << " sendtag = " << send_->tag()
         << " source = " << recv_->source()
         << " recvtag = " << recv_->tag()
         << " request = " << request_ << ")";
  return stream.str();
}

}  // namespace mpi
}  // namespace arbiter
}  // namespace dbug
