/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <mpi.h>

#include <list>
#include <set>
#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/mpi.h"
#include "dbug/arbiter/mpi-collective.h"
#include "dbug/arbiter/mpi-event.h"
#include "dbug/arbiter/request-prototypes.h"
#include "dbug/common/util.h"

namespace dbug {
namespace arbiter {

void MPIAbort::Execute() {
  server_->set_terminate(true);
  MESSAGE("Thread %d executes MPI abort.", thread_->id());
}

void MPIAbort::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIAbort::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI abort"
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPIAllToAll::Execute() {
  mpi::AllToAll *alltoall = server_->mpi()->alltoall_lookup(this);
  ASSERT(alltoall, "Request is missing its all-to-all event.");
  alltoall->add_to_departed(origin_);
  alltoall->remove_from_waiting(this);
  thread_->join_with_clock(alltoall->clock());
  if (alltoall->num_departed() == size_) {
    server_->mpi()->remove_alltoall(alltoall);
    delete alltoall;
  }
  server_->mpi()->remove_alltoall_lookup(this);
  MESSAGE("Thread %d executes MPI all-to-all.", thread_->id());
}

void MPIAllToAll::PreSchedule() {
  mpi::MPI::AllToAlls alltoalls =
      server_->mpi()->alltoalls(communicator_, type_);
  mpi::AllToAll *alltoall = NULL;
  for (mpi::AllToAll *handle : alltoalls) {
    std::set<int> arrived = handle->arrived();
    if (arrived.find(origin_) == arrived.end()) {
      alltoall = handle;
      break;
    }
  }
  if (!alltoall) {
    alltoall = new mpi::AllToAll(communicator_, size_, type_);
    server_->mpi()->add_alltoall(alltoall);
  }
  alltoall->join_with_clock(thread_->clock());
  alltoall->add_to_arrived(origin_);
  alltoall->add_to_waiting(this);
  alltoall->add_to_ranks(world_rank_);
  server_->mpi()->set_alltoall_lookup(this, alltoall);
}

void MPIAllToAll::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.set_serialize(false);
  mpi::AllToAll *alltoall = server_->mpi()->alltoall_lookup(this);
  ASSERT(alltoall, "Request is missing its all-to-all event.");
  for (int rank : alltoall->ranks()) {
    std::stringstream stream;
    stream << "mpi-rank-" << rank;
    group.add_to_write_set(stream.str());
  }
  if (requests.find(this) != requests.end() &&
      alltoall->num_arrived() == size_) {
    for (Request *request : alltoall->waiting()) {
      ASSERT(requests.find(request) != requests.end(),
             "Request waiting for an all-to-all event does not exist.");
      group.add_request(request);
      requests.erase(request);
    }
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIAllToAll::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI all-to-all"
         << " communicator=" << communicator_
         << " origin=" << origin_
         << " size=" << size_
         << " type=" << type_;
  result.push_back(stream.str());
  return result;
}

void MPIAllToOne::Execute() {
  mpi::AllToOne *alltoone = server_->mpi()->alltoone_lookup(this);
  ASSERT(alltoone, "Request is missing its all-to-one event.");
  alltoone->add_to_departed(origin_);
  alltoone->remove_from_waiting(this);
  if (origin_ == sink_) { thread_->join_with_clock(alltoone->clock()); }
  if (alltoone->num_departed() == size_) {
    server_->mpi()->remove_alltoone(alltoone);
    delete alltoone;
  }
  server_->mpi()->remove_alltoone_lookup(this);
  MESSAGE("Thread %d executes MPI all-to-one.", thread_->id());
}

void MPIAllToOne::PreSchedule() {
  mpi::MPI::AllToOnes alltoones =
      server_->mpi()->alltoones(communicator_, type_);
  mpi::AllToOne *alltoone = NULL;
  for (mpi::AllToOne *handle : alltoones) {
    std::set<int> arrived = handle->arrived();
    if (arrived.find(origin_) == arrived.end()) {
      alltoone = handle;
      break;
    }
  }
  if (!alltoone) {
    alltoone = new mpi::AllToOne(communicator_, sink_, size_, type_);
    server_->mpi()->add_alltoone(alltoone);
  }
  alltoone->join_with_clock(thread_->clock());
  alltoone->add_to_arrived(origin_);
  alltoone->add_to_waiting(this);
  server_->mpi()->set_alltoone_lookup(this, alltoone);
}

void MPIAllToOne::Test(Set requests, GroupSet *groups) {
  mpi::AllToOne *alltoone = server_->mpi()->alltoone_lookup(this);
  ASSERT(alltoone, "Request is missing its all-to-one event.");
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  if (origin_ != sink_ || alltoone->num_departed() == size_ - 1) {
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIAllToOne::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI all-to-one"
         << " communicator=" << communicator_
         << " origin=" << origin_
         << " size=" << size_
         << " sink=" << sink_
         << " type=" << type_;
  result.push_back(stream.str());
  return result;
}

void MPIBufferedSend::Execute() {
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Send *send_event = event->ExtractSend();
  ASSERT(send_event, "Event is not a send.");
  if (!nondet_choices_.empty() && nondet_choices_.back() == 0) {
    send_event->set_state(mpi::Event::POSTED);
    send_event->set_request(NULL);
  } else {
    if (!nondet_choices_.empty()) { nondet_choices_.pop_back(); }
    send_event->Match(server_->mpi()->events(communicator_),
                      nondet_choices_);
    thread_->join_with_clock(event->clock());
    server_->mpi()->remove_event(event);
  }
  MESSAGE("Thread %d executes buffered MPI send.", thread_->id());
}

void MPIBufferedSend::PreSchedule() {
  mpi::Send *send_event =
      new mpi::Send(this, communicator_, origin_, nbytes_,
                    destination_, tag_, mpi::Send::BUFFERED);
  server_->mpi()->add_event(send_event);
}

void MPIBufferedSend::Test(Set requests, GroupSet *groups) {
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Send *send_event = event->ExtractSend();
  ASSERT(send_event, "Event is not a send event.");
  // 1) Ignore this event.
  // TODO(jsimsa): Make sure that we don't need this.
  // Request::Test(requests, groups);
  // 2) Buffer the send event.
  if (!send_event->match()) {
    Request::Group group;
    std::vector<int> nondet_choices;
    nondet_choices.push_back(0);
    group.add_request(this, nondet_choices);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);
    groups->insert(group);
    Request::Test(requests, groups);
  }
  // 3) Try to match the send event.
  std::set<Option> options;
  send_event->TestMatch(server_->mpi()->events(communicator_), &options);
  for (Option option : options) {
    Set requests_copy = requests;
    Request::Group group;
    group.set_serialize(false);
    option.second.push_back(1);
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    bool valid_group = true;
    for (Request *request : option.first) {
      if (requests_copy.erase(request) == 1) {
        group.add_request(request);
      } else {
        valid_group = false;
      }
    }
    Enable(&group);
    if (valid_group) { groups->insert(group); }
    Request::Test(requests_copy, groups);
  }
}

std::list<std::string> MPIBufferedSend::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: buffered MPI send"
         << " origin=" << origin_
         << " destination=" << destination_
         << " tag=" << tag_
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPIBufferDetach::Execute() {
  MESSAGE("Thread %d executes MPI buffer detach request.", thread_->id());
}

void MPIBufferDetach::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  mpi::MPI::EventCommMap communicator_map = server_->mpi()->events();
  for (std::pair<int, mpi::MPI::EventOriginMap> pair : communicator_map) {
    mpi::MPI::EventOriginMap origin_map = pair.second;
    for (std::pair<int, mpi::MPI::Events> pair : origin_map) {
      mpi::MPI::Events events = pair.second;
      for (mpi::Event *event : events) {
        mpi::Send *send_event = event->ExtractSend();
        if (send_event &&
            send_event->type() == mpi::Send::BUFFERED) {
          groups->insert(group);
          Request::Test(requests, groups);
          return;
        }
      }
    }
  }
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIBufferDetach::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI buffer detach";
  result.push_back(stream.str());
  return result;
}

void MPICreateCommunicator::Execute() {
  mpi::CommCreate *create = server_->mpi()->comm_create_lookup(this);
  ASSERT(create, "Request is missing its create event.");
  create->add_to_departed(origin_);
  create->remove_from_waiting(this);
  thread_->join_with_clock(create->clock());
  return_value_->return_value =
      server_->mpi()->comm_id(create, tag_, subgroup_);
  if (create->num_departed() == create->group_size()) {
    server_->mpi()->remove_comm_create(create);
    server_->mpi()->remove_comm_id(create, tag_);
    delete create;
  }
  server_->mpi()->remove_comm_create_lookup(this);
  MESSAGE("Thread %d creates MPI communicator.", thread_->id());
}

void MPICreateCommunicator::PreSchedule() {
  mpi::MPI::CommCreates creates =
      server_->mpi()->comm_creates(communicator_);
  mpi::CommCreate *create = NULL;
  for (mpi::CommCreate *handle : creates) {
    std::set<int> group = handle->group();
    std::set<int> arrived = handle->arrived();
    if (handle->tag() == tag_ &&
        group.find(origin_) != group.end() &&
        arrived.find(origin_) == arrived.end()) {
      create = handle;
      break;
    }
  }
  if (!create) {
    create = new mpi::CommCreate(communicator_, group_, tag_);
    server_->mpi()->add_comm_create(create);
  }
  create->join_with_clock(thread_->clock());
  create->add_to_arrived(origin_);
  create->add_to_waiting(this);
  create->add_to_ranks(world_rank_);
  server_->mpi()->set_comm_create_lookup(this, create);
  server_->mpi()->create_comm_id(create, tag_, subgroup_);
}

void MPICreateCommunicator::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.set_serialize(false);
  mpi::CommCreate *create = server_->mpi()->comm_create_lookup(this);
  ASSERT(create, "Request is missing its create event.");
  for (int rank : create->ranks()) {
    std::stringstream stream;
    stream << "mpi-rank-" << rank;
    group.add_to_write_set(stream.str());
  }
  if (requests.find(this) != requests.end()
      && create->num_arrived() == create->group_size()) {
    for (Request *request : create->waiting()) {
      ASSERT(requests.find(request) != requests.end(),
             "Request waiting for a create event does not exist.");
      group.add_request(request);
      requests.erase(request);
    }
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPICreateCommunicator::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI communicator create"
         << " communicator=" << communicator_
         << " origin=" << origin_
         << " group=" << Util::SetToString<int>(group_)
         << " tag=" << tag_
         << " subgroup=" << subgroup_;
  result.push_back(stream.str());
  return result;
}

void MPIMatchingProbe::Execute() {
  mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
  recv_event.Match(server_->mpi()->events(communicator_),
                   nondet_choices_);
  thread_->join_with_clock(recv_event.clock());
  mpi::MatchedRecv *matched_event;
  if (recv_event.match()) {
    mpi::Send *send_event = recv_event.match()->ExtractSend();
    ASSERT(send_event, "Receive is not matched with a send.");
    matched_event = new mpi::MatchedRecv(this,
                                         communicator_,
                                         origin_,
                                         send_event->origin(),
                                         send_event->tag(),
                                         message_id_);
    matched_event->set_match(send_event);
    matched_event->join_with_clock(send_event->clock());
    send_event->set_match(matched_event);
    return_value_->return_value.nbytes = send_event->nbytes();
  } else {
    matched_event = new mpi::MatchedRecv(this,
                                         communicator_,
                                         origin_,
                                         MPI_PROC_NULL,
                                         MPI_ANY_TAG,
                                         message_id_);
    return_value_->return_value.nbytes = 0;
  }
  return_value_->return_value.source = matched_event->source();
  return_value_->return_value.tag = matched_event->tag();
  server_->mpi()->add_event(matched_event);
  matched_event->set_request(NULL);
  MESSAGE("Thread %d executes MPI matching probe.", thread_->id());
}

void MPIMatchingProbe::Test(Set requests, GroupSet *groups) {
  mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
  std::set<Option> options;
  recv_event.TestMatch(server_->mpi()->events(communicator_), &options);
  for (const Option &option : options) {
    Request::Group group;
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);
    groups->insert(group);
  }
  if (options.empty()) {
    Request::Group group;
    group.add_request(this);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    groups->insert(group);
  }
  Request::Test(requests, groups);
}

std::list<std::string> MPIMatchingProbe::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI matching probe"
         << " origin=" << origin_
         << " source=" << source_
         << " tag=" << tag_
         << " communicator=" << communicator_
         << " message=" << message_id_;
  result.push_back(stream.str());
  return result;
}

void MPIImmediateMatchingProbe::Execute() {
  if (nondet_choices_.back() != 0) {
    nondet_choices_.pop_back();
    mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
    recv_event.Match(server_->mpi()->events(communicator_), nondet_choices_);
    thread_->join_with_clock(recv_event.clock());
    mpi::MatchedRecv *matched_event;
    return_value_->return_value.flag = true;
    if (recv_event.match()) {
      mpi::Send *send_event = recv_event.match()->ExtractSend();
      ASSERT(send_event, "Probe is not matched with a send.");
      matched_event = new mpi::MatchedRecv(this,
                                           communicator_,
                                           origin_,
                                           send_event->origin(),
                                           send_event->tag(),
                                           message_id_);
      matched_event->set_match(send_event);
      matched_event->join_with_clock(send_event->clock());
      send_event->set_match(matched_event);
      return_value_->return_value.nbytes = send_event->nbytes();
    } else {
      matched_event = new mpi::MatchedRecv(this,
                                           communicator_,
                                           origin_,
                                           MPI_PROC_NULL,
                                           MPI_ANY_TAG,
                                           message_id_);
      return_value_->return_value.nbytes = 0;
    }
    return_value_->return_value.source = matched_event->source();
    return_value_->return_value.tag = matched_event->tag();
    server_->mpi()->add_event(matched_event);
    matched_event->set_request(NULL);
  } else {
    return_value_->return_value.flag = false;
  }
  MESSAGE("Thread %d executes non-blocking MPI matching probe.", thread_->id());
}

void MPIImmediateMatchingProbe::Test(Set requests, GroupSet *groups) {
  mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
  std::set<Option> options;
  recv_event.TestMatch(server_->mpi()->events(communicator_), &options);
  for (Option option : options) {
    Request::Group group;
    option.second.push_back(1);
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);
    groups->insert(group);
  }
  if (options.empty()) {
    Request::Group group;
    std::vector<int> nondet_choices;
    nondet_choices.push_back(0);
    group.add_request(this, nondet_choices);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);
    groups->insert(group);
  }
  Request::Test(requests, groups);
}

std::list<std::string> MPIImmediateMatchingProbe::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: non-blocking MPI matching probe"
         << " origin=" << origin_
         << " source=" << source_
         << " tag=" << tag_
         << " communicator=" << communicator_
         << " message=" << message_id_;
  result.push_back(stream.str());
  return result;
}

void MPIMatchingReceive::Execute() {
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Recv *recv_event = event->ExtractRecv();
  ASSERT(recv_event, "Event is not a receive.");
  if (recv_event->match()) {
    mpi::Send *send_event = recv_event->match()->ExtractSend();
    ASSERT(send_event, "Receive is not matched with a send.");
    if (send_event->type() == mpi::Send::BUFFERED &&
        send_event->state() == mpi::Event::POSTED) {
      server_->mpi()->remove_event(send_event);
    }
  }
  thread_->join_with_clock(event->clock());
  server_->mpi()->remove_event(event);
  MESSAGE("Thread %d executes MPI matching receive.", thread_->id());
}

void MPIMatchingReceive::PreSchedule() {
  mpi::Event *event = server_->mpi()->matched_event_lookup(message_id_);
  ASSERT(event, "Unable to look up a previously matched event.");
  mpi::Recv *recv_event = event->ExtractRecv();
  ASSERT(recv_event, "Event is not a receive.");
  mpi::Recv *new_recv_event =
      new mpi::Recv(this,
                    recv_event->communicator(),
                    recv_event->origin(),
                    recv_event->source(),
                    recv_event->tag());
  new_recv_event->set_match(recv_event->match());
  recv_event->set_match(NULL);
  if (new_recv_event->match()) {
    mpi::Send *send_event = new_recv_event->match()->ExtractSend();
    ASSERT(send_event, "Receive is not matched with a send.");
    new_recv_event->join_with_clock(send_event->clock());
    send_event->set_match(new_recv_event);
    // Propagate new_event->clock() through the matched chain.
    while (1) {
      send_event->join_with_clock(new_recv_event->clock());
      if (send_event->parent()) {
        send_event->parent()->join_with_clock(new_recv_event->clock());
        mpi::Recv *recv_event = send_event->parent()->ExtractRecv();
        ASSERT(recv_event, "Send-receive is missing the receive event.");
        ASSERT(recv_event->match(), "Partially matched send-receive.");
        send_event = recv_event->match()->ExtractSend();
      } else {
        break;
      }
    }
  }
  server_->mpi()->add_event(new_recv_event);
  server_->mpi()->remove_event(event);
}

void MPIMatchingReceive::Test(Set requests, GroupSet *groups) {
  // 1) Ignore this event.
  // TODO(jsimsa): Make sure that we don't need this.
  // Request::Test(requests, groups);
  // 2) Try to match the receive event.
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Recv *recv_event = event->ExtractRecv();
  ASSERT(recv_event, "Event is not a receive.");
  std::set<Option> options;
  recv_event->TestMatch(server_->mpi()->events(recv_event->communicator()),
                        &options);
  ASSERT(options.size() == 1, "Nondeterministic matching receive.");
  for (const Option &option : options) {
    Request::Group group;
    group.set_serialize(false);
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    bool valid_group = true;
    for (Request *request : option.first) {
      if (requests.erase(request) == 1) {
        group.add_request(request);
      } else {
        valid_group = false;
      }
    }
    Enable(&group);
    if (valid_group) { groups->insert(group); }
    Request::Test(requests, groups);
  }
}

std::list<std::string> MPIMatchingReceive::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI matching receive"
         << " message=" << message_id_;
  result.push_back(stream.str());
  return result;
}

void MPIOneToAll::Execute() {
  mpi::OneToAll *onetoall = server_->mpi()->onetoall_lookup(this);
  ASSERT(onetoall, "Request is missing its one-to-all event.");
  onetoall->add_to_departed(origin_);
  onetoall->remove_from_waiting(this);
  if (origin_ != root_) { thread_->join_with_clock(onetoall->clock()); }
  if (onetoall->num_departed() == size_) {
    server_->mpi()->remove_onetoall(onetoall);
    delete onetoall;
  }
  server_->mpi()->remove_onetoall_lookup(this);
  MESSAGE("Thread %d executes MPI one-to-all.", thread_->id());
}

void MPIOneToAll::PreSchedule() {
  mpi::MPI::OneToAlls onetoalls =
      server_->mpi()->onetoalls(communicator_, type_);
  mpi::OneToAll *onetoall = NULL;
  for (mpi::OneToAll *handle : onetoalls) {
    std::set<int> arrived = handle->arrived();
    if (arrived.find(origin_) == arrived.end()) {
      onetoall = handle;
      break;
    }
  }
  if (!onetoall) {
    onetoall = new mpi::OneToAll(communicator_, size_, root_, type_);
    server_->mpi()->add_onetoall(onetoall);
  }
  onetoall->add_to_arrived(origin_);
  onetoall->add_to_waiting(this);
  server_->mpi()->set_onetoall_lookup(this, onetoall);
  if (origin_ == root_) {
    onetoall->join_with_clock(thread_->clock());
  }
}

void MPIOneToAll::Test(Set requests, GroupSet *groups) {
  mpi::OneToAll *onetoall = server_->mpi()->onetoall_lookup(this);
  ASSERT(onetoall, "Request is missing its one-to-all event.");
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  if (origin_ == root_ || onetoall->num_departed() > 0) {
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIOneToAll::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI one-to-all"
         << " communicator=" << communicator_
         << " origin=" << origin_
         << " size=" << size_
         << " root=" << root_
         << " type=" << type_;
  result.push_back(stream.str());
  return result;
}

void MPIProbe::Execute() {
  mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
  mpi::Event *match =
      recv_event.ProbeMatch(server_->mpi()->events(communicator_),
                            nondet_choices_);
  thread_->join_with_clock(recv_event.clock());
  int nbytes, source, tag;
  if (match) {
    mpi::Send *send_event = match->ExtractSend();
    ASSERT(send_event, "Probe is not matched with a send.");
    nbytes = send_event->nbytes();
    source = send_event->origin();
    tag = send_event->tag();
  } else {
    nbytes = 0;
    source = MPI_PROC_NULL;
    tag = MPI_ANY_TAG;
  }
  return_value_->return_value.nbytes = nbytes;
  return_value_->return_value.source = source;
  return_value_->return_value.tag = tag;
  MESSAGE("Thread %d executes MPI probe.", thread_->id());
}

void MPIProbe::Test(Set requests, GroupSet *groups) {
  mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
  std::set<Option> options;
  recv_event.TestMatch(server_->mpi()->events(communicator_), &options);
  for (const Option &option : options) {
    Request::Group group;
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);
    groups->insert(group);
  }
  if (options.empty()) {
    Request::Group group;
    group.add_request(this);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    groups->insert(group);
  }
  Request::Test(requests, groups);
}

std::list<std::string> MPIProbe::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI probe"
         << " origin=" << origin_
         << " source=" << source_
         << " tag=" << tag_
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPIImmediateProbe::Execute() {
  if (nondet_choices_.back() != 0) {
    nondet_choices_.pop_back();
    mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
    mpi::Event *match =
        recv_event.ProbeMatch(server_->mpi()->events(communicator_),
                              nondet_choices_);
    thread_->join_with_clock(recv_event.clock());
    int nbytes, source, tag;
    if (match) {
      mpi::Send *send_event = match->ExtractSend();
      ASSERT(send_event, "Probe is not matched with a send.");
      nbytes = send_event->nbytes();
      source = send_event->origin();
      tag = send_event->tag();
    } else {
      nbytes = 0;
      source = MPI_PROC_NULL;
      tag = MPI_ANY_TAG;
    }
    return_value_->return_value.nbytes = nbytes;
    return_value_->return_value.source = source;
    return_value_->return_value.tag = tag;
    return_value_->return_value.flag = true;
  } else {
    return_value_->return_value.flag = false;
  }
  MESSAGE("Thread %d executes non-blocking MPI probe.", thread_->id());
}

void MPIImmediateProbe::Test(Set requests, GroupSet *groups) {
  mpi::Recv recv_event(this, communicator_, origin_, source_, tag_);
  std::set<Option> options;
  recv_event.TestMatch(server_->mpi()->events(communicator_), &options);
  for (Option option : options) {
    Request::Group group;
    option.second.push_back(1);
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);
    groups->insert(group);
  }
  if (options.empty()) {
    Request::Group group;
    std::vector<int> nondet_choices;
    nondet_choices.push_back(0);
    group.add_request(this, nondet_choices);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);
    groups->insert(group);
  }
  Request::Test(requests, groups);
}

std::list<std::string> MPIImmediateProbe::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: non-blocking MPI probe"
         << " origin=" << origin_
         << " source=" << source_
         << " tag=" << tag_
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPIReadySend::Execute() {
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Send *send_event = event->ExtractSend();
  ASSERT(send_event, "Event is not a send.");
  if (!nondet_choices_.empty()) {
    send_event->Match(server_->mpi()->events(communicator_), nondet_choices_);
  }
  thread_->join_with_clock(event->clock());
  server_->mpi()->remove_event(event);
  MESSAGE("Thread %d executes ready MPI send.", thread_->id());
}

void MPIReadySend::PreSchedule() {
  mpi::Send *send_event =
      new mpi::Send(this, communicator_, origin_, nbytes_,
                    destination_, tag_, mpi::Send::SYNCHRONOUS);
  server_->mpi()->add_event(send_event);
}

void MPIReadySend::Test(Set requests, GroupSet *groups) {
  // 1) Ignore this event.
  // TODO(jsimsa): Make sure that we don't need this.
  // Request::Test(requests, groups);
  // 2) Try to match the send event.
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Send *send_event = event->ExtractSend();
  ASSERT(send_event, "Event is not a send.");
  std::set<Option> options;
  send_event->TestMatch(server_->mpi()->events(communicator_), &options);
  for (const Option &option : options) {
    Set requests_copy = requests;
    Request::Group group;
    group.set_serialize(false);
    bool valid_group = true;
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    for (Request *request : option.first) {
      if (requests_copy.erase(request) == 1) {
        group.add_request(request);
      } else {
        valid_group = false;
      }
    }
    Enable(&group);
    if (valid_group) { groups->insert(group); }
    Request::Test(requests_copy, groups);
  }
  if (options.empty()) {
    Request::Group group;
    group.add_request(this);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    Enable(&group);  // Execute ready send unmatched.
    groups->insert(group);
    Request::Test(requests, groups);
  }
}

std::list<std::string> MPIReadySend::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: ready MPI send"
         << " origin=" << origin_
         << " destination=" << destination_
         << " tag=" << tag_
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPIReceive::Execute() {
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Recv *recv_event = event->ExtractRecv();
  ASSERT(recv_event, "Event is not a receive.");
  recv_event->Match(server_->mpi()->events(communicator_), nondet_choices_);
  if (recv_event->match()) {
    mpi::Send *send_event = recv_event->match()->ExtractSend();
    ASSERT(send_event, "Receive is not matched with a send.");
    return_value_->return_value.source = send_event->origin();
    return_value_->return_value.tag = send_event->tag();
    if (send_event->type() == mpi::Send::BUFFERED &&
        send_event->state() == mpi::Event::POSTED) {
      server_->mpi()->remove_event(send_event);
    }
  } else {
    return_value_->return_value.source = MPI_PROC_NULL;
    return_value_->return_value.tag = MPI_ANY_TAG;
  }
  thread_->join_with_clock(event->clock());
  server_->mpi()->remove_event(event);
  MESSAGE("Thread %d executes MPI receive.", thread_->id());
}

void MPIReceive::PreSchedule() {
  mpi::Recv *recv_event =
      new mpi::Recv(this, communicator_, origin_, source_, tag_);
  server_->mpi()->add_event(recv_event);
}

void MPIReceive::Test(Set requests, GroupSet *groups) {
  // 1) Ignore the event
  // TODO(jsimsa): Make sure that we don't need this.
  // Request::Test(requests, groups);
  // 2) Try to match the receive event.
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Recv *recv_event = event->ExtractRecv();
  ASSERT(recv_event, "Event is not a receive.");
  std::set<Option> options;
  recv_event->TestMatch(server_->mpi()->events(communicator_), &options);
  for (const Option &option : options) {
    Set requests_copy = requests;
    Request::Group group;
    group.set_serialize(false);
    bool valid_group = true;
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    for (Request *request : option.first) {
      if (requests_copy.erase(request) == 1) {
        group.add_request(request);
      } else {
        valid_group = false;
      }
    }
    Enable(&group);
    if (valid_group) { groups->insert(group); }
    Request::Test(requests_copy, groups);
  }
  if (options.empty()) {
    Request::Group group;
    group.add_request(this);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    groups->insert(group);
    Request::Test(requests, groups);
  }
}

std::list<std::string> MPIReceive::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI receive"
         << " origin=" << origin_
         << " source=" << source_
         << " tag=" << tag_
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPIRequestFree::Execute() {
  server_->mpi()->remove_request(request_id_);
  MESSAGE("Thread %d frees MPI request.", thread_->id());
}

void MPIRequestFree::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIRequestFree::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI request free"
         << " request=" << request_id_;
  result.push_back(stream.str());
  return result;
}

void MPIScan::Execute() {
  mpi::Scan *scan = server_->mpi()->scan_lookup(this);
  ASSERT(scan, "Request is missing its scan event.");
  scan->add_to_departed(origin_);
  scan->remove_from_waiting(this);
  thread_->join_with_clock(scan->clock());
  scan->join_with_clock(thread_->clock());
  if (scan->num_departed() == size_) {
    server_->mpi()->remove_scan(scan);
    delete scan;
  }
  server_->mpi()->remove_scan_lookup(this);
  MESSAGE("Thread %d executes MPI scan.", thread_->id());
}

void MPIScan::PreSchedule() {
  mpi::MPI::Scans scans = server_->mpi()->scans(communicator_, type_);
  for (mpi::Scan *scan : scans) {
    std::set<int> arrived = scan->arrived();
    if (arrived.find(origin_) == arrived.end()) {
      scan->add_to_arrived(origin_);
      scan->add_to_waiting(this);
      server_->mpi()->set_scan_lookup(this, scan);
      return;
    }
  }
  mpi::Scan *scan = new mpi::Scan(communicator_, size_, type_);
  scan->add_to_arrived(origin_);
  scan->add_to_waiting(this);
  server_->mpi()->add_scan(scan);
  server_->mpi()->set_scan_lookup(this, scan);
}

void MPIScan::Test(Set requests, GroupSet *groups) {
  mpi::Scan *scan = server_->mpi()->scan_lookup(this);
  ASSERT(scan, "Request is missing its scan event.");
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  if (origin_ == scan->num_departed()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIScan::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI scan"
         << " communicator=" << communicator_
         << " origin=" << origin_
         << " size=" << size_
         << " type=" << type_;
  result.push_back(stream.str());
  return result;
}

void MPISendReceive::Execute() {
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  event->Match(server_->mpi()->events(communicator_), nondet_choices_);
  mpi::Recv *recv_event = event->ExtractRecv();
  ASSERT(recv_event, "Send-receive is missing the receive event.");
  if (recv_event->match()) {
    mpi::Send *send_event = recv_event->match()->ExtractSend();
    ASSERT(send_event, "Receive is not matched with a send.");
    return_value_->return_value.source = send_event->origin();
    return_value_->return_value.tag = send_event->tag();
    if (send_event->type() == mpi::Send::BUFFERED &&
        send_event->state() == mpi::Event::POSTED) {
      server_->mpi()->remove_event(send_event);
    }
  } else {
    return_value_->return_value.source = MPI_PROC_NULL;
    return_value_->return_value.tag = MPI_ANY_TAG;
  }
  thread_->join_with_clock(event->clock());
  server_->mpi()->remove_event(event);
  MESSAGE("Thread %d executes MPI send-receive.", thread_->id());
}

void MPISendReceive::PreSchedule() {
  mpi::SendRecv *sendrecv_event =
      new mpi::SendRecv(this, communicator_, origin_, nbytes_,
                        destination_, sendtag_, source_, recvtag_);
  server_->mpi()->add_event(sendrecv_event);
}

void MPISendReceive::Test(Set requests, GroupSet *groups) {
  // 1) Ignore this event.
  // TODO(jsimsa): Make sure that we don't need this.
  // Request::Test(requests, groups);
  // 2) Try to match the send-receive event.
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  std::set<Option> options;
  event->TestMatch(server_->mpi()->events(communicator_), &options);
  for (const Option &option : options) {
    Set requests_copy = requests;
    Request::Group group;
    group.set_serialize(false);
    bool valid_group = true;
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    for (Request *request : option.first) {
      if (requests_copy.erase(request) == 1) {
        group.add_request(request);
      } else {
        valid_group = false;
      }
    }
    Enable(&group);
    if (valid_group) { groups->insert(group); }
    Request::Test(requests_copy, groups);
  }
  if (options.empty()) {
    Request::Group group;
    group.add_request(this);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    groups->insert(group);
    Request::Test(requests, groups);
  }
}

std::list<std::string> MPISendReceive::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI send-receive"
         << " origin=" << origin_
         << " destination=" << destination_
         << " sendtag=" << sendtag_
         << " source=" << source_
         << " recvtag=" << recvtag_
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPISynchronousSend::Execute() {
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Send *send_event = event->ExtractSend();
  ASSERT(send_event, "Event is not a send.");
  send_event->Match(server_->mpi()->events(communicator_), nondet_choices_);
  thread_->join_with_clock(event->clock());
  server_->mpi()->remove_event(event);
  MESSAGE("Thread %d executes synchronous MPI send.", thread_->id());
}

void MPISynchronousSend::PreSchedule() {
  mpi::Send *send_event =
      new mpi::Send(this, communicator_, origin_, nbytes_,
                    destination_, tag_, mpi::Send::SYNCHRONOUS);
  server_->mpi()->add_event(send_event);
}

void MPISynchronousSend::Test(Set requests, GroupSet *groups) {
  // 1) Ignore this event.
  // TODO(jsimsa): Make sure that we don't need this.
  // Request::Test(requests, groups);
  // 2) Try to match the send event.
  mpi::Event *event = server_->mpi()->event_lookup(this);
  ASSERT(event, "Request is missing its event.");
  mpi::Send *send_event = event->ExtractSend();
  ASSERT(send_event, "Event is not a send.");
  std::set<Option> options;
  send_event->TestMatch(server_->mpi()->events(communicator_), &options);
  for (const Option &option : options) {
    Set requests_copy = requests;
    Request::Group group;
    group.set_serialize(false);
    bool valid_group = true;
    group.add_request(this, option.second);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    for (Request *request : option.first) {
      if (requests_copy.erase(request) == 1) {
        group.add_request(request);
      } else {
        valid_group = false;
      }
    }
    Enable(&group);
    if (valid_group) { groups->insert(group); }
    Request::Test(requests_copy, groups);
  }
  if (options.empty()) {
    Request::Group group;
    group.add_request(this);
    std::stringstream stream;
    stream << "mpi-rank-" << world_rank_;
    group.add_to_write_set(stream.str());
    groups->insert(group);
    Request::Test(requests, groups);
  }
}

std::list<std::string> MPISynchronousSend::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: synchronous MPI send"
         << " origin=" << origin_
         << " destination=" << destination_
         << " tag=" << tag_
         << " communicator=" << communicator_;
  result.push_back(stream.str());
  return result;
}

void MPITest::Execute() {
  return_value_->return_value = server_->mpi()->request_exists(request_id_);
  if (return_value_->return_value) {
    Clock clock = server_->mpi()->request_clock(request_id_);
    thread_->join_with_clock(clock);
    server_->mpi()->remove_request(request_id_);
  }
  MESSAGE("Thread %d executes MPI test.", thread_->id());
}

void MPITest::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPITest::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI test"
         << " request=" << request_id_;
  result.push_back(stream.str());
  return result;
}

void MPITestAll::Execute() {
  bool all_enabled = true;
  for (std::string request_id : request_ids_) {
    all_enabled &= server_->mpi()->request_exists(request_id);
  }
  return_value_->return_value = all_enabled;
  if (all_enabled) {
    for (std::string request_id : request_ids_) {
      Clock clock = server_->mpi()->request_clock(request_id);
      thread_->join_with_clock(clock);
      server_->mpi()->remove_request(request_id);
    }
  }
  MESSAGE("Thread %d executes MPI test all.", thread_->id());
}

void MPITestAll::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPITestAll::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI test all"
         << " requests=" << Util::VectorToString<std::string>(request_ids_);
  result.push_back(stream.str());
  return result;
}

void MPITestAny::Execute() {
  return_value_->return_value = MPI_UNDEFINED;
  for (unsigned i = 0; i < request_ids_.size(); i++) {
    std::string request_id = request_ids_[i];
    if (server_->mpi()->request_exists(request_id)) {
      return_value_->return_value = i;
      break;
    }
  }
  if (return_value_->return_value != MPI_UNDEFINED) {
    std::string request_id = request_ids_[return_value_->return_value];
    Clock clock = server_->mpi()->request_clock(request_id);
    thread_->join_with_clock(clock);
    server_->mpi()->remove_request(request_id);
  }
  MESSAGE("Thread %d executes MPI test any.", thread_->id());
}

void MPITestAny::Test(Set requests, GroupSet *groups) {
  // TODO(jsimsa): Enumerate possible matches.
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPITestAny::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI test any"
         << " requests=" << Util::VectorToString<std::string>(request_ids_);
  result.push_back(stream.str());
  return result;
}

void MPITestSome::Execute() {
  for (unsigned i = 0; i < request_ids_.size(); i++) {
    std::string request_id = request_ids_[i];
    if (server_->mpi()->request_exists(request_id)) {
      return_value_->return_value.push_back(i);
    }
  }
  for (int index : return_value_->return_value) {
    Clock clock = server_->mpi()->request_clock(request_ids_[index]);
    thread_->join_with_clock(clock);
    server_->mpi()->remove_request(request_ids_[index]);
  }
  MESSAGE("Thread %d executes MPI test some.", thread_->id());
}

void MPITestSome::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPITestSome::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI test some"
         << " requests=" << Util::VectorToString<std::string>(request_ids_);
  result.push_back(stream.str());
  return result;
}

void MPIWait::Execute() {
  Clock clock = server_->mpi()->request_clock(request_id_);
  thread_->join_with_clock(clock);
  server_->mpi()->remove_request(request_id_);
  MESSAGE("Thread %d executes MPI wait.", thread_->id());
}

void MPIWait::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  if (server_->mpi()->request_exists(request_id_)) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIWait::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI wait"
         << " request=" << request_id_;
  result.push_back(stream.str());
  return result;
}

void MPIWaitAll::Execute() {
  for (const std::string &request_id : request_ids_) {
    Clock clock = server_->mpi()->request_clock(request_id);
    thread_->join_with_clock(clock);
    server_->mpi()->remove_request(request_id);
  }
  MESSAGE("Thread %d executes MPI wait all.", thread_->id());
}

void MPIWaitAll::Test(Set requests, GroupSet *groups) {
  bool all_enabled = true;
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  for (unsigned i = 0; i < request_ids_.size(); i++) {
    all_enabled &= server_->mpi()->request_exists(request_ids_[i]);
  }
  if (all_enabled) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIWaitAll::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI wait all"
         << " requests=" << Util::VectorToString<std::string>(request_ids_);
  result.push_back(stream.str());
  return result;
}

void MPIWaitAny::Execute() {
  unsigned index;
  for (index = 0; index < request_ids_.size(); index++) {
    if (server_->mpi()->request_exists(request_ids_[index])) { break; }
  }
  ASSERT(index < request_ids_.size(),
         "Vector of requests is too short.");
  Clock clock = server_->mpi()->request_clock(request_ids_[index]);
  thread_->join_with_clock(clock);
  server_->mpi()->remove_request(request_ids_[index]);
  MESSAGE("Thread %d executes MPI wait any.", thread_->id());
}

void MPIWaitAny::Test(Set requests, GroupSet *groups) {
  // TODO(jsimsa): Enumerate possible matches.
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  for (unsigned i = 0; i < request_ids_.size(); i++) {
    if (server_->mpi()->request_exists(request_ids_[i])) {
      Enable(&group);
      break;
    }
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIWaitAny::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI wait any"
         << " requests=" << Util::VectorToString<std::string>(request_ids_);
  result.push_back(stream.str());
  return result;
}

void MPIWaitSome::Execute() {
  for (unsigned i = 0; i < request_ids_.size(); i++) {
    if (server_->mpi()->request_exists(request_ids_[i])) {
      return_value_->return_value.push_back(i);
    }
  }
  for (int index : return_value_->return_value) {
    Clock clock = server_->mpi()->request_clock(request_ids_[index]);
    thread_->join_with_clock(clock);
    server_->mpi()->remove_request(request_ids_[index]);
  }
  MESSAGE("Thread %d executes MPI wait some.", thread_->id());
}

void MPIWaitSome::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "mpi-rank-" << world_rank_;
  group.add_to_write_set(stream.str());
  for (unsigned i = 0; i < request_ids_.size(); i++) {
    if (server_->mpi()->request_exists(request_ids_[i])) {
      Enable(&group);
      break;
    }
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MPIWaitSome::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MPI wait some"
         << " requests=" << Util::VectorToString<std::string>(request_ids_);
  result.push_back(stream.str());
  return result;
}

}  // namespace arbiter
}  // namespace dbug
