/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/mpi.h"

#include "dbug/arbiter/mpi-collective.h"
#include "dbug/arbiter/mpi-event.h"

namespace dbug {
namespace arbiter {
namespace mpi {

bool MPI::CommIDKey::operator<(const CommIDKey &rhs) const {
  if (create != rhs.create) { return create < rhs.create; }
  if (tag != rhs.tag) { return tag < rhs.tag; }
  if (subgroup != rhs.subgroup) { return subgroup < rhs.subgroup; }
  return false;
}

MPI::MPI() : comm_id_counter_(0) {}

MPI::AllToAllCommMap MPI::alltoalls() const { return alltoalls_; }

MPI::AllToAllTypeMap MPI::alltoalls(int comm) const {
  if (alltoalls_.find(comm) != alltoalls_.end()) {
    return alltoalls_.at(comm);
  }
  AllToAllTypeMap empty;
  return empty;
}

MPI::AllToAlls MPI::alltoalls(int comm, int type) const {
  if (alltoalls_.find(comm) != alltoalls_.end()) {
    AllToAllTypeMap type_map = alltoalls_.at(comm);
    if (type_map.find(type) != type_map.end()) {
      return type_map.at(type);
    }
  }
  AllToAlls empty;
  return empty;
}

mpi::AllToAll *MPI::alltoall_lookup(Request *key) const {
  if (alltoalls_lookup_.find(key) != alltoalls_lookup_.end()) {
    return alltoalls_lookup_.at(key);
  }
  return NULL;
}

MPI::AllToOneCommMap MPI::alltoones() const { return alltoones_; }

MPI::AllToOneTypeMap MPI::alltoones(int comm) const {
  if (alltoones_.find(comm) != alltoones_.end()) {
    return alltoones_.at(comm);
  }
  AllToOneTypeMap empty;
  return empty;
}

MPI::AllToOnes MPI::alltoones(int comm, int type) const {
  if (alltoones_.find(comm) != alltoones_.end()) {
    AllToOneTypeMap type_map = alltoones_.at(comm);
    if (type_map.find(type) != type_map.end()) {
      return type_map.at(type);
    }
  }
  AllToOnes empty;
  return empty;
}

mpi::AllToOne *MPI::alltoone_lookup(Request *key) const {
  if (alltoones_lookup_.find(key) != alltoones_lookup_.end()) {
    return alltoones_lookup_.at(key);
  }
  return NULL;
}

int MPI::any_source() const { return any_source_; }

int MPI::any_tag() const { return any_tag_; }

MPI::CommCreateCommMap MPI::comm_creates() const {
  return comm_creates_;
}

MPI::CommCreates MPI::comm_creates(int comm) const {
  if (comm_creates_.find(comm) != comm_creates_.end()) {
    return comm_creates_.at(comm);
  }
  CommCreates empty;
  return empty;
}

mpi::CommCreate *MPI::comm_create_lookup(Request *key) const {
  if (comm_creates_lookup_.find(key) != comm_creates_lookup_.end()) {
    return comm_creates_lookup_.at(key);
  }
  return NULL;
}

int MPI::comm_id(CommCreate *create, int tag, std::string subgroup) const {
  CommIDKey key;
  key.create = create;
  key.tag = tag;
  key.subgroup = subgroup;
  if (comm_ids_.find(key) != comm_ids_.end()) {
    return comm_ids_.at(key);
  }
  return -1;
}

MPI::EventCommMap MPI::events() const { return events_; }

MPI::EventOriginMap MPI::events(int communicator) const {
  if (events_.find(communicator) != events_.end()) {
    return events_.at(communicator);
  }
  EventOriginMap empty;
  return empty;
}

MPI::Events MPI::events(int communicator, int origin) const {
  if (events_.find(communicator) != events_.end()) {
    EventOriginMap events = events_.at(communicator);
    if (events.find(origin) != events.end()) { return events.at(origin); }
  }
  Events empty;
  return empty;
}

mpi::Event *MPI::event_lookup(Request *key) const {
  if (events_lookup_.find(key) != events_lookup_.end()) {
    return events_lookup_.at(key);
  }
  return NULL;
}

mpi::Event *MPI::matched_event_lookup(std::string key) const {
  if (matched_events_.find(key) != matched_events_.end()) {
    return matched_events_.at(key);
  }
  return NULL;
}

MPI::OneToAllCommMap MPI::onetoalls() const { return onetoalls_; }

MPI::OneToAllTypeMap MPI::onetoalls(int comm) const {
  if (onetoalls_.find(comm) != onetoalls_.end()) {
    return onetoalls_.at(comm);
  }
  OneToAllTypeMap empty;
  return empty;
}

MPI::OneToAlls MPI::onetoalls(int comm, int type) const {
  if (onetoalls_.find(comm) != onetoalls_.end()) {
    OneToAllTypeMap type_map = onetoalls_.at(comm);
    if (type_map.find(type) != type_map.end()) {
      return type_map.at(type);
    }
  }
  OneToAlls empty;
  return empty;
}

mpi::OneToAll *MPI::onetoall_lookup(Request *key) const {
  if (onetoalls_lookup_.find(key) != onetoalls_lookup_.end()) {
    return onetoalls_lookup_.at(key);
  }
  return NULL;
}

int MPI::rank(int rank) const { return rank_.at(rank); }

bool MPI::request_exists(std::string request_id) const {
  return (requests_.find(request_id) != requests_.end());
}

Clock MPI::request_clock(std::string request_id) const {
  if (requests_.find(request_id) != requests_.end()) {
    return requests_.at(request_id);
  }
  Clock empty;
  return empty;
}

MPI::ScanCommMap MPI::scans() const { return scans_; }

MPI::ScanTypeMap MPI::scans(int comm) const {
  if (scans_.find(comm) != scans_.end()) {
    return scans_.at(comm);
  }
  ScanTypeMap empty;
  return empty;
}

MPI::Scans MPI::scans(int comm, int type) const {
  if (scans_.find(comm) != scans_.end()) {
    ScanTypeMap type_map = scans_.at(comm);
    if (type_map.find(type) != type_map.end()) {
      return type_map.at(type);
    }
  }
  Scans empty;
  return empty;
}

mpi::Scan *MPI::scan_lookup(Request *key) const {
  if (scans_lookup_.find(key) != scans_lookup_.end()) {
    return scans_lookup_.at(key);
  }
  return NULL;
}

// mutators

void MPI::add_alltoall(AllToAll *alltoall) {
  AllToAllTypeMap type_map = alltoalls(alltoall->communicator());
  AllToAlls alltoalls = type_map[alltoall->type()];
  alltoalls.push_back(alltoall);
  type_map[alltoall->type()] = alltoalls;
  alltoalls_[alltoall->communicator()] = type_map;
}

void MPI::remove_alltoall(AllToAll *alltoall) {
  AllToAllTypeMap type_map = alltoalls(alltoall->communicator());
  AllToAlls alltoalls = type_map[alltoall->type()];
  alltoalls.remove(alltoall);
  type_map[alltoall->type()] = alltoalls;
  alltoalls_[alltoall->communicator()] = type_map;
}

void MPI::set_alltoall_lookup(Request *key, AllToAll *value) {
  alltoalls_lookup_[key] = value;
}

void MPI::remove_alltoall_lookup(Request *key) {
  alltoalls_lookup_.erase(key);
}

void MPI::add_alltoone(AllToOne *alltoone) {
  AllToOneTypeMap type_map = alltoones(alltoone->communicator());
  AllToOnes alltoones = type_map[alltoone->type()];
  alltoones.push_back(alltoone);
  type_map[alltoone->type()] = alltoones;
  alltoones_[alltoone->communicator()] = type_map;
}

void MPI::remove_alltoone(AllToOne *alltoone) {
  AllToOneTypeMap type_map = alltoones(alltoone->communicator());
  AllToOnes alltoones = type_map[alltoone->type()];
  alltoones.remove(alltoone);
  type_map[alltoone->type()] = alltoones;
  alltoones_[alltoone->communicator()] = type_map;
}

void MPI::set_alltoone_lookup(Request *key, AllToOne *value) {
  alltoones_lookup_[key] = value;
}

void MPI::remove_alltoone_lookup(Request *key) {
  alltoones_lookup_.erase(key);
}

void MPI::set_any_source(int any_source) { any_source_ = any_source; }

void MPI::set_any_tag(int any_tag) { any_tag_ = any_tag; }

void MPI::add_comm_create(CommCreate *create) {
  CommCreates creates = comm_creates(create->communicator());
  creates.push_back(create);
  comm_creates_[create->communicator()] = creates;
}

void MPI::remove_comm_create(CommCreate *create) {
  CommCreates creates = comm_creates(create->communicator());
  creates.remove(create);
  comm_creates_[create->communicator()] = creates;
}

void MPI::set_comm_create_lookup(Request *key, CommCreate *value) {
  comm_creates_lookup_[key] = value;
}

void MPI::remove_comm_create_lookup(Request *key) {
  comm_creates_lookup_.erase(key);
}

void MPI::create_comm_id(CommCreate *create,
                         int tag,
                         std::string subgroup) {
  CommIDKey key;
  key.create = create;
  key.tag = tag;
  key.subgroup = subgroup;
  if (comm_ids_.find(key) == comm_ids_.end()) {
    comm_ids_[key] = ++comm_id_counter_;
  }
}

void MPI::remove_comm_id(CommCreate *create, int tag) {
  std::map<CommIDKey, int>::iterator it = comm_ids_.begin();
  while (it != comm_ids_.end()) {
    if (it->first.create == create && it->first.tag == tag) {
      it = comm_ids_.erase(it);
    } else {
      ++it;
    }
  }
}

void MPI::add_event(Event *event) {
  MatchedEvent *matched_event = event->ExtractMatchedEvent();
  if (matched_event) {
    matched_events_[matched_event->handle()] = event;
  } else {
    EventOriginMap origin_map = events(event->communicator());
    Events e = events(event->communicator(), event->origin());
    e.push_back(event);
    origin_map[event->origin()] = e;
    events_[event->communicator()] = origin_map;
    events_lookup_[event->request()] = event;
  }
}

void MPI::remove_event(Event *event) {
  event->set_state(Event::COMPLETED);
  MatchedEvent *matched_event = event->ExtractMatchedEvent();
  if (matched_event) {
    matched_events_.erase(matched_event->handle());
    // TODO(jsimsa): Garbage collect events. If all events reachable
    // by following the match pointers are COMPLETED, then all such
    // events can be deallocated.
    //
    // delete event;
    return;
  } else {
    EventOriginMap origin_map = events(event->communicator());
    Events e = events(event->communicator(), event->origin());
    e.remove(event);
    origin_map[event->origin()] = e;
    events_[event->communicator()] = origin_map;
    // Only erase the lookup mapping, if it points to the event being removed.
    if (events_lookup_[event->request()] == event) {
      events_lookup_.erase(event->request());
    }
    // TODO(jsimsa): Garbage collect events. If all events reachable
    // by following the match pointers are COMPLETED, then all such
    // events can be deallocated.
    //
    // delete event;
  }
}

void MPI::add_onetoall(OneToAll *onetoall) {
  OneToAllTypeMap type_map = onetoalls(onetoall->communicator());
  OneToAlls onetoalls = type_map[onetoall->type()];
  onetoalls.push_back(onetoall);
  type_map[onetoall->type()] = onetoalls;
  onetoalls_[onetoall->communicator()] = type_map;
}

void MPI::remove_onetoall(OneToAll *onetoall) {
  OneToAllTypeMap type_map = onetoalls(onetoall->communicator());
  OneToAlls onetoalls = type_map[onetoall->type()];
  onetoalls.remove(onetoall);
  type_map[onetoall->type()] = onetoalls;
  onetoalls_[onetoall->communicator()] = type_map;
}

void MPI::set_onetoall_lookup(Request *key, OneToAll *value) {
  onetoalls_lookup_[key] = value;
}

void MPI::remove_onetoall_lookup(Request *key) {
  onetoalls_lookup_.erase(key);
}

void MPI::set_rank(int rank, int dbug_rank) { rank_[rank] = dbug_rank; }

void MPI::add_request(std::string request_id, Clock clock) {
  requests_[request_id] = clock;
}

void MPI::remove_request(std::string request_id) {
  requests_.erase(request_id);
}

void MPI::add_scan(Scan *scan) {
  ScanTypeMap type_map = scans(scan->communicator());
  Scans scans = type_map[scan->type()];
  scans.push_back(scan);
  type_map[scan->type()] = scans;
  scans_[scan->communicator()] = type_map;
}

void MPI::remove_scan(Scan *scan) {
  ScanTypeMap type_map = scans(scan->communicator());
  Scans scans = type_map[scan->type()];
  scans.remove(scan);
  type_map[scan->type()] = scans;
  scans_[scan->communicator()] = type_map;
}

void MPI::set_scan_lookup(Request *key, Scan *value) {
  scans_lookup_[key] = value;
}

void MPI::remove_scan_lookup(Request *key) {
  scans_lookup_.erase(key);
}

}  // namespace mpi
}  // namespace arbiter
}  // namespace dbug
