/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/posix-resource.h"

#include <boost/algorithm/string/predicate.hpp>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/common/util.h"
#include "src/thrift/arbiter/arbiter_types.h"

namespace dbug {
namespace arbiter {
namespace posix {

// Resource class

Resource::Resource(std::string id, Clock clock) : clock_(clock), id_(id) {}

Resource::~Resource() {}

// accessors

Clock Resource::clock() const { return clock_; }

std::string Resource::id() const { return id_; }

// mutators

void Resource::join_with_clock(const Clock &clock) { clock_.JoinWith(clock); }

void Resource::set_clock(const Clock &clock) { clock_ = clock; }

// public methods

std::list<std::string> Resource::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  result.push_back(" Resource:");
  stream << "  id: " << id_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  clock: " << Util::VectorToString<int>(clock_);
  result.push_back(stream.str());
  return result;
}

// Barrier class

Barrier::Barrier(std::string id, int count, Clock clock)
    : Resource(id, clock), count_(count) {}

Barrier::~Barrier() {}

// accessors

int Barrier::count() const { return count_; }

int Barrier::num_waiters() const { return waiters_.size(); }

Request::Set Barrier::waiters() const { return waiters_; }

// mutators

void Barrier::add_waiter(Request *request) { waiters_.insert(request); }

void Barrier::remove_waiter(Request *request) { waiters_.erase(request); }

void Barrier::set_count(int count) { count_ = count; }

// public method

std::list<std::string> Barrier::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  count: " << count_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  waiters: {";
  for (Request *request : waiters_) {
    stream << " " << request->thread()->id();
  }
  stream << " }";
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// CVar class

CVar::CVar(std::string id, Clock clock) : Resource(id, clock) {}

CVar::~CVar() {}

// accessors

bool CVar::signaled(Request *request) {
  return (signaled_.find(request) != signaled_.end());
}

bool CVar::signaled_empty() const { return signaled_.empty(); }

std::list<Request *> CVar::waiters() const { return waiters_; }

bool CVar::waiters_empty() const { return waiters_.empty(); }

Request *CVar::waiters_front() const { return waiters_.front(); }

// mutators

void CVar::add_signaled(Request *request) { signaled_.insert(request); }

void CVar::remove_signaled(Request *request) { signaled_.erase(request); }

void CVar::waiters_clear() { waiters_.clear(); }

void CVar::waiters_pop_front() { waiters_.pop_front(); }

void CVar::waiters_push_back(Request * element) { waiters_.push_back(element); }

// public methods

std::list<std::string> CVar::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  waiters: " << Util::ListToString<Request *>(waiters_);
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// EPoll::Event class

EPoll::Event::Event(int events, System::Int64 data)
    : data_(data), events_(events) {}

// accessors

System::Int64 EPoll::Event::data() const { return data_; }

int EPoll::Event::events() const { return events_; }

// mutators

void EPoll::Event::set_data(System::Int64 data) { data_ = data; }

void EPoll::Event::set_events(int events) { events_ = events; }

// public methods

std::string EPoll::Event::ToString() const {
  std::stringstream stream;
  stream << " data = " << data_ << ", events = " << events_;
  return stream.str();
}

// EPoll class

EPoll::EPoll(std::string id, Clock clock) : Resource(id, clock) {}

EPoll::~EPoll() {
  for (std::pair<std::string, Event *> pair : events_) {
    const Event *event = pair.second;
    delete event;
  }
  events_.clear();
}

// accessors

std::map<std::string, EPoll::Event *> EPoll::events() const { return events_; }

EPoll::Event *EPoll::get_event(std::string rid) const {
  if (events_.find(rid) != events_.end()) { return events_.at(rid); }
  return NULL;
}

// mutators

void EPoll::add_event(std::string rid, Event *event) {
  events_[rid] = event;
}

void EPoll::remove_event(std::string rid) {
  if (events_.find(rid) != events_.end()) {
    delete events_[rid];
    events_.erase(rid);
  }
}

// public methods

std::list<std::string> EPoll::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  for (std::pair<std::string, Event *> pair : events_) {
    stream << "  event: fd = " << pair.first << "," << pair.second->ToString();
    result.push_back(stream.str());
    stream.str(std::string());
  }
  return result;
}

// File class

File::File(std::string id, Clock clock)
    : Resource(id, clock), flags_(0), nrefs_(0) {}

File::~File() {}

// accessors

int File::flags() const { return flags_; }

int File::nrefs() const { return nrefs_; }

// mutators

void File::decrement_nrefs() { nrefs_--; }

void File::increment_nrefs() { nrefs_++; }

void File::set_flags(int flags) { flags_ = flags; }

// public methods

bool File::IsBlocking() const { return false; }

bool File::IsReadable() const { return true; }

bool File::IsWritable() const { return true; }

std::list<std::string> File::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  flags: " << flags_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of references: " << nrefs_;
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// Mutex class

std::string Mutex::StateToString(State state) {
  switch (state) {
    case LOCKED:
      return "locked";
    case UNLOCKED:
      return "unlocked";
    default:
      return "unknown";
  }
}

Mutex::Mutex(std::string id, bool recursive, Clock clock)
    : Resource(id, clock),
      counter_(0), owner_(0), recursive_(recursive), state_(UNLOCKED) {}

Mutex::~Mutex() {}

// accessors

int Mutex::counter() const { return counter_; }

int Mutex::owner() const { return owner_; }

bool Mutex::recursive() const { return recursive_; }

Mutex::State Mutex::state() const { return state_; }

// mutators

void Mutex::decrement_counter() { counter_--; }

void Mutex::increment_counter() { counter_++; }

void Mutex::set_owner(int owner) { owner_ = owner; }

void Mutex::set_state(State state) { state_ = state; }

// public methods

std::list<std::string> Mutex::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  counter: " << counter_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  recursive: " << recursive_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  owner: " << owner_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  state: " << StateToString(state_);
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// Pipe class

int Pipe::GenerateID() { return ++id_counter_; }

Pipe::Pipe(std::string id, Clock clock)
    : Resource(id, clock),
      flags_(0),
      input_(-1),
      nbytes_(0),
      output_(-1),
      size_(0) {}

Pipe::~Pipe() {}

// accessors

int Pipe::flags() const { return flags_; }

int Pipe::input() const { return input_; }

int Pipe::nbytes() const { return nbytes_; }

int Pipe::output() const { return output_; }

int Pipe::size() const { return size_; }

// mutators

void Pipe::decrement_nbytes(int delta) {
  nbytes_ = nbytes_ > delta ? nbytes_ - delta : 0;
}

void Pipe::increment_nbytes(int delta) { nbytes_ += delta; }

void Pipe::reset_nbytes() { nbytes_ = 0; }

void Pipe::set_flags(int flags) { flags_ = flags; }

void Pipe::set_input(int fd) { input_ = fd; }

void Pipe::set_output(int fd) { output_ = fd; }

void Pipe::set_size(int size) { size_ = size; }

// public methods

bool Pipe::IsBlocking() const { return !(flags_ & O_NONBLOCK); }

bool Pipe::IsReadable() const { return nbytes_ > 0; }

bool Pipe::IsWritable() const { return nbytes_ < size_; }

std::list<std::string> Pipe::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  flags: " << flags_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  input file descriptor: " << input_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of bytes: " << nbytes_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  output file descriptor: " << output_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  size: " << size_;
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// private static fields

int Pipe::id_counter_ = 0;

// RWLock class

std::string RWLock::StateToString(State state) {
  switch (state) {
    case FREE:
      return "free";
    case SHARED:
      return "shared";
    case EXCLUSIVE:
      return "exclusive";
    default:
      return "unknown";
  }
}

RWLock::RWLock(std::string id, Clock clock)
    : Resource(id, clock), state_(FREE), writer_(0) {}
RWLock::~RWLock() {}

// accessors

bool RWLock::readers_empty() const { return readers_.empty(); }

RWLock::State RWLock::state() const { return state_; }

int RWLock::writer() const { return writer_; }

// mutators

void RWLock::readers_add(int element) { readers_.insert(element); }

void RWLock::readers_clear() { readers_.clear(); }

bool RWLock::readers_contains(int element) {
  return (readers_.find(element) != readers_.end());
}

void RWLock::readers_remove(int element) {
  std::multiset<int>::iterator it = readers_.find(element);
  if (it != readers_.end()) { readers_.erase(it); }
}

void RWLock::set_state(State state) { state_ = state; }

void RWLock::set_writer(int writer) { writer_ = writer; }

// public methods

std::list<std::string> RWLock::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  readers: {";
  for (int id : readers_) { stream << " " << id; }
  stream << " }";
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  state: " << StateToString(state_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  writer: " << writer_;
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// Semaphore class

Semaphore::Semaphore(std::string id, bool named, int value, Clock clock)
    : Resource(id, clock),
      marked_(false),
      named_(named),
      nrefs_(1),
      value_(value) {}

Semaphore::~Semaphore() {}

// accessors

bool Semaphore::marked() const { return marked_; }

bool Semaphore::named() const { return named_; }

int Semaphore::nrefs() const { return nrefs_; }

int Semaphore::value() const { return value_; }

// mutators

void Semaphore::decrement_nrefs() { nrefs_--; }

void Semaphore::decrement_value() { value_--; }

void Semaphore::increment_nrefs() { nrefs_++; }

void Semaphore::increment_value() { value_++; }

void Semaphore::set_marked(bool marked) { marked_ = marked; }

// public methods

std::list<std::string> Semaphore::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  marked: " << marked_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  named: " << named_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  nrefs: " << nrefs_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  value: " << value_;
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// Socket class

Socket::Socket(std::string id,
               int domain,
               int type,
               int protocol,
               Clock clock)
    : Resource(id, clock),
      connection_(NULL),
      domain_(domain),
      flags_(0),
      listening_(false),
      nbytes_(0),
      port_(0),
      protocol_(protocol),
      size_(0),
      type_(type) {
  if (type_ != SOCK_STREAM && type_ != SOCK_DGRAM) {
    WARNING("Unsupported socket type %d.", type_);
  }
}

// accessors

std::string Socket::address() const { return address_; }

Socket *Socket::connection() const { return connection_; }

int Socket::domain() const { return domain_; }

int Socket::flags() const { return flags_; }

bool Socket::listening() const { return listening_; }

int Socket::nbytes() const { return nbytes_; }

std::string Socket::node_id() const { return node_id_; }

std::list<Socket *> Socket::pending_connections() const {
  return pending_connections_;
}

std::list<Socket *> Socket::pending_messages() const {
  return pending_messages_;
}

int Socket::port() const { return port_; }

int Socket::protocol() const { return protocol_; }

int Socket::size() const { return size_; }

int Socket::type() const { return type_; }

// mutators

void Socket::add_pending_connection(Socket *socket) {
  pending_connections_.push_back(socket);
}

void Socket::add_pending_message(Socket *socket) {
  pending_messages_.push_back(socket);
}

void Socket::decrement_nbytes(int delta) {
  nbytes_ = nbytes_ > delta ? nbytes_ - delta : 0;
}

void Socket::increment_nbytes(int delta) { nbytes_ += delta; }

Socket *Socket::remove_pending_connection() {
  if (pending_connections_.empty()) { return NULL; }
  Socket *result = pending_connections_.front();
  pending_connections_.pop_front();
  return result;
}

Socket *Socket::remove_pending_message() {
  if (pending_messages_.empty()) { return NULL; }
  Socket *result = pending_messages_.front();
  pending_messages_.pop_front();
  return result;
}

void Socket::reset_nbytes() { nbytes_ = 0; }

void Socket::set_address(std::string address) { address_ = address; }

void Socket::set_connection(Socket *connection) { connection_ = connection; }

void Socket::set_flags(int flags) { flags_ = flags; }

void Socket::set_listening(bool listening) { listening_ = listening; }

void Socket::set_node_id(std::string node_id) { node_id_ = node_id; }

void Socket::set_port(int port) { port_ = port; }

void Socket::set_size(int size) { size_ = size; }

// public methods

std::list<std::string> Socket::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  address: " << address_;
  result.push_back(stream.str());
  stream.str(std::string());
  if (connection_) {
    stream << "  connection: " << connection_->id();
    result.push_back(stream.str());
    stream.str(std::string());
  }
  stream << "  domain: " << domain_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  flags: " << flags_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  listening: " << listening_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  nbytes: " << nbytes_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  node id: " << node_id_;
  result.push_back(stream.str());
  stream.str(std::string());
  for (const Socket *socket : pending_connections_) {
    stream << "  pending connection: " << socket->id();
    result.push_back(stream.str());
    stream.str(std::string());
  }
  for (const Socket *socket : pending_messages_) {
    stream << "  pending message: " << socket->id();
    result.push_back(stream.str());
    stream.str(std::string());
  }
  stream << "  port: " << port_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  protocol: " << protocol_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  size: " << size_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  type: " << type_;
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

Socket *Socket::GetReadableRemote() const {
  switch (type_) {
    case SOCK_STREAM:
      if (connection_ && connection_->nbytes() > 0) { return connection_; }
      if (!connection_) {
        for (Socket *pending : pending_connections_) {
          if (pending->nbytes() > 0 ||
              pending->flags() & O_NONBLOCK) { return pending; }
        }
      }
      break;
    case SOCK_DGRAM:
      if (!pending_messages_.empty() &&
          pending_messages_.front()->nbytes() > 0) {
        return pending_messages_.front();
      }
      break;
    default:
      break;
  }
  return NULL;
}

bool Socket::IsBlocking() const { return !(flags_ & O_NONBLOCK); }

bool Socket::IsReadable() const { return (GetReadableRemote() != NULL); }

bool Socket::IsWritable() const { return nbytes_ < size_; }

// Spin class

std::string Spin::StateToString(State state) {
  switch (state) {
    case LOCKED:
      return "locked";
    case UNLOCKED:
      return "unlocked";
    default:
      return "unknown";
  }
}

Spin::Spin(std::string id, Clock clock)
    : Resource(id, clock), owner_(0), state_(UNLOCKED) {}
Spin::~Spin() {}

// accessors

int Spin::owner() const { return owner_; }

Spin::State Spin::state() const { return state_; }

// mutators

void Spin::set_owner(int owner) { owner_ = owner; }

void Spin::set_state(State state) { state_ = state; }

// public methods

std::list<std::string> Spin::ToString() const {
  std::list<std::string> result = Resource::ToString();
  std::stringstream stream;
  stream << "  owner: " << owner_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  state: " << StateToString(state_);
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

}  // namespace posix
}  // namespace arbiter
}  // namespace dbug

