/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/posix.h"

#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/posix-resource.h"

namespace dbug {
namespace arbiter {
namespace posix {

// POSIX::Statistics class

POSIX::Statistics::Statistics() {
  nbarriers_ = 0;
  ncvars_ = 0;
  nepolls_ = 0;
  nfiles_ = 0;
  nmutexes_ = 0;
  npipes_ = 0;
  nrwlocks_ = 0;
  nsemaphores_ = 0;
  nsockets_ = 0;
  nspins_ = 0;
}

POSIX::Statistics::~Statistics() {
  for (const std::string &line : this->ToString()) {
    MESSAGE("%s", line.c_str());
  }
}

// public methods

std::list<std::string> POSIX::Statistics::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  stream << "  number of barriers: " << nbarriers_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of condition variables: " << ncvars_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of epolls: " << nepolls_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of files: " << nfiles_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of mutexes: " << nmutexes_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of pipes: " << npipes_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of read-write locks: " << nrwlocks_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of semaphores: " << nsemaphores_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of sockets: " << nsockets_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of spin locks: " << nspins_;
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// POSIX class

POSIX::POSIX() {}

POSIX::~POSIX() {
  if (!barriers_.empty()) {
    WARNING("Destroying arbiter with %d barrier(s).", barriers_.size());
    for (std::pair<std::string, Barrier *> pair : barriers_) {
      Barrier *barrier = pair.second;
      delete barrier;
    }
    barriers_.clear();
  }
  if (!cvars_.empty()) {
    WARNING("Destroying arbiter with %d condition variable(s).", cvars_.size());
    for (std::pair<std::string, CVar *> pair : cvars_) {
      CVar *cvar = pair.second;
      delete cvar;
    }
    cvars_.clear();
  }
  if (!epolls_.empty()) {
    WARNING("Destroying arbiter with %d epoll(s).", epolls_.size());
    for (std::pair<std::string, EPoll *> pair : epolls_) {
      EPoll *epoll = pair.second;
      delete epoll;
    }
    epolls_.clear();
  }
  if (!files_.empty()) {
    WARNING("Destroying arbiter with %d file(s).", files_.size());
    for (std::pair<std::string, File *> pair : files_) {
      File *file = pair.second;
      delete file;
    }
    files_.clear();
  }
  if (!mutexes_.empty()) {
    WARNING("Destroying arbiter with %d mutex(es).", mutexes_.size());
    for (std::pair<std::string, Mutex *> pair : mutexes_) {
      Mutex *mutex = pair.second;
      delete mutex;
    }
    mutexes_.clear();
  }
  if (!pipes_.empty()) {
    WARNING("Destroying arbiter with %d pipe(s).", pipes_.size());
    for (std::pair<std::string, Pipe *> pair : pipes_) {
      Pipe *pipe = pair.second;
      delete pipe;
    }
    pipes_.clear();
  }
  if (!rwlocks_.empty()) {
    WARNING("Destroying arbiter with %d read-write lock(s).", rwlocks_.size());
    for (std::pair<std::string, RWLock *> pair : rwlocks_) {
      RWLock *rwlock = pair.second;
      delete rwlock;
    }
    rwlocks_.clear();
  }
  if (!semaphores_.empty()) {
    WARNING("Destroying arbiter with %d semaphore(s).", semaphores_.size());
    for (std::pair<std::string, Semaphore *> pair : semaphores_) {
      Semaphore *semaphore = pair.second;
      delete semaphore;
    }
    semaphores_.clear();
  }
  if (!sockets_.empty()) {
    WARNING("Destroying arbiter with %d socket(s).", sockets_.size());
    for (std::pair<std::string, Socket *> pair : sockets_) {
      Socket *socket = pair.second;
      delete socket;
    }
    sockets_.clear();
  }
  if (!spins_.empty()) {
    WARNING("Destroying arbiter with %d spin lock(s).", spins_.size());
    for (std::pair<std::string, Spin *> pair : spins_) {
      Spin *spin = pair.second;
      delete spin;
    }
    spins_.clear();
  }
}

// accessors

POSIX::Barriers POSIX::barriers() const { return barriers_; }

Barrier *POSIX::barrier_lookup(std::string rid) const {
  if (barriers_.find(rid) != barriers_.end()) { return barriers_.at(rid); }
  return NULL;
}

POSIX::CVars POSIX::cvars() const { return cvars_; }

CVar *POSIX::cvar_lookup(std::string rid) const {
  if (cvars_.find(rid) != cvars_.end()) { return cvars_.at(rid); }
  return NULL;
}

POSIX::EPolls POSIX::epolls() const { return epolls_; }

EPoll *POSIX::epoll_lookup(std::string rid) const {
  if (epolls_.find(rid) != epolls_.end()) { return epolls_.at(rid); }
  return NULL;
}

POSIX::Files POSIX::files() const { return files_; }

File *POSIX::file_lookup(std::string rid) const {
  if (files_.find(rid) != files_.end()) { return files_.at(rid); }
  return NULL;
}

POSIX::Mutexes POSIX::mutexes() const { return mutexes_; }

Mutex *POSIX::mutex_lookup(std::string rid) const {
  if (mutexes_.find(rid) != mutexes_.end()) { return mutexes_.at(rid); }
  return NULL;
}

POSIX::Pipes POSIX::pipes() const { return pipes_; }

Pipe *POSIX::pipe_lookup(std::string rid) const {
  if (pipes_.find(rid) != pipes_.end()) { return pipes_.at(rid); }
  return NULL;
}

Socket *POSIX::remote_socket_lookup(std::string node_id,
                                    std::string address,
                                    int port) const {
  for (std::pair<std::string, Socket *> pair : sockets_) {
    Socket *socket = pair.second;
    if (port == socket->port() &&
        (address == socket->address() ||
         socket->address() == "0.0.0.0" ||
         (address == "127.0.0.1" && node_id == socket->node_id()))) {
      return socket;
    }
  }
  return NULL;
}

POSIX::RWLocks POSIX::rwlocks() const { return rwlocks_; }

RWLock *POSIX::rwlock_lookup(std::string rid) const {
  if (rwlocks_.find(rid) != rwlocks_.end()) { return rwlocks_.at(rid); }
  return NULL;
}

POSIX::Semaphores POSIX::semaphores() const { return semaphores_; }

Semaphore *POSIX::semaphore_lookup(std::string rid) const {
  if (semaphores_.find(rid) != semaphores_.end()) {
    return semaphores_.at(rid);
  }
  return NULL;
}

POSIX::Sockets POSIX::sockets() const { return sockets_; }

Socket *POSIX::socket_lookup(std::string rid) const {
  if (sockets_.find(rid) != sockets_.end()) { return sockets_.at(rid); }
  return NULL;
}

POSIX::Spins POSIX::spins() const { return spins_; }

Spin *POSIX::spin_lookup(std::string rid) const {
  if (spins_.find(rid) != spins_.end()) { return spins_.at(rid); }
  return NULL;
}

// mutators

void POSIX::add_barrier(Barrier *barrier) {
  statistics_.nbarriers_++;
  barriers_[barrier->id()] = barrier;
}

void POSIX::add_cvar(CVar *cvar) {
  statistics_.ncvars_++;
  cvars_[cvar->id()] = cvar;
}

void POSIX::add_epoll(EPoll *epoll) {
  statistics_.nepolls_++;
  epolls_[epoll->id()] = epoll;
}

void POSIX::add_file(File *file) {
  statistics_.nfiles_++;
  files_[file->id()] = file;
}

void POSIX::add_mutex(Mutex *mutex) {
  statistics_.nmutexes_++;
  mutexes_[mutex->id()] = mutex;
}

void POSIX::add_pipe(Pipe *pipe) {
  statistics_.npipes_++;
  pipes_[pipe->id()] = pipe;
}

void POSIX::add_rwlock(RWLock *rwlock) {
  statistics_.nrwlocks_++;
  rwlocks_[rwlock->id()] = rwlock;
}

void POSIX::add_semaphore(Semaphore *semaphore) {
  statistics_.nsemaphores_++;
  semaphores_[semaphore->id()] = semaphore;
}

void POSIX::add_socket(Socket *socket) {
  statistics_.nsockets_++;
  sockets_[socket->id()] = socket;
}

void POSIX::add_spin(Spin *spin) {
  statistics_.nspins_++;
  spins_[spin->id()] = spin;
}

void POSIX::remove_barrier(Barrier *barrier) { barriers_.erase(barrier->id()); }

void POSIX::remove_cvar(CVar *cvar) { cvars_.erase(cvar->id()); }

void POSIX::remove_epoll(EPoll *epoll) { epolls_.erase(epoll->id()); }

void POSIX::remove_file(File *file) { files_.erase(file->id()); }

void POSIX::remove_mutex(Mutex *mutex) { mutexes_.erase(mutex->id()); }

void POSIX::remove_pipe(Pipe *pipe) { pipes_.erase(pipe->id()); }

void POSIX::remove_rwlock(RWLock *rwlock) { rwlocks_.erase(rwlock->id()); }

void POSIX::remove_semaphore(Semaphore *semaphore) {
  semaphores_.erase(semaphore->id());
}

void POSIX::remove_socket(Socket *socket) { sockets_.erase(socket->id()); }

void POSIX::remove_spin(Spin *spin) { spins_.erase(spin->id()); }

// public methods

std::list<std::string> POSIX::ToString() const {
  std::list<std::string> result;
  for (std::pair<std::string, const Barrier *> pair : barriers_) {
    const Barrier *barrier = pair.second;
    result.splice(result.end(), barrier->ToString());
  }
  for (std::pair<std::string, const CVar *> pair : cvars_) {
    const CVar *cvar = pair.second;
    result.splice(result.end(), cvar->ToString());
  }
  for (std::pair<std::string, const EPoll *> pair : epolls_) {
    const EPoll *epoll = pair.second;
    result.splice(result.end(), epoll->ToString());
  }
  for (std::pair<std::string, const File *> pair : files_) {
    const File *file = pair.second;
    result.splice(result.end(), file->ToString());
  }
  for (std::pair<std::string, const Mutex *> pair : mutexes_) {
    const Mutex *mutex = pair.second;
    result.splice(result.end(), mutex->ToString());
  }
  for (std::pair<std::string, const Pipe *> pair : pipes_) {
    const Pipe *pipe = pair.second;
    result.splice(result.end(), pipe->ToString());
  }
  for (std::pair<std::string, const RWLock *> pair : rwlocks_) {
    const RWLock *rwlock = pair.second;
    result.splice(result.end(), rwlock->ToString());
  }
  for (std::pair<std::string, const Semaphore *> pair : semaphores_) {
    const Semaphore *semaphore = pair.second;
    result.splice(result.end(), semaphore->ToString());
  }
  for (std::pair<std::string, const Socket *> pair : sockets_) {
    const Socket *socket = pair.second;
    result.splice(result.end(), socket->ToString());
  }
  for (std::pair<std::string, const Spin *> pair : spins_) {
    const Spin *spin = pair.second;
    result.splice(result.end(), spin->ToString());
  }
  return result;
}

}  // namespace posix
}  // namespace arbiter
}  // namespace dbug
