/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/posix-resource.h"
#include "dbug/arbiter/request-prototypes.h"

namespace dbug {
namespace arbiter {

using posix::Semaphore;

void SemaphoreClose::Execute() {
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  ASSERT(semaphore, "Request is missing its semaphore.");
  thread_->join_with_clock(semaphore->clock());
  semaphore->join_with_clock(thread_->clock());
  semaphore->decrement_nrefs();
  if (semaphore->nrefs() == 0 && semaphore->marked()) {
    server_->posix()->remove_semaphore(semaphore);
    delete semaphore;
  }
  MESSAGE("Thread %d closed %s.", thread_->id(), rid_.c_str());
}

void SemaphoreClose::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (semaphore) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SemaphoreClose::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: closing " << rid_;
  result.push_back(stream.str());
  return result;
}

void SemaphoreCreate::Execute() {
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (!semaphore) {
    Semaphore *sem = new Semaphore(rid_, named_, value_, thread_->clock());
    server_->posix()->add_semaphore(sem);
  }
  MESSAGE("Thread %d created %s.", thread_->id(), rid_.c_str());
}

void SemaphoreCreate::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (!semaphore || semaphore->named()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SemaphoreCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: creating " << rid_;
  result.push_back(stream.str());
  return result;
}

void SemaphoreDelete::Execute() {
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  ASSERT(semaphore, "Request is missing its semaphore.");
  thread_->join_with_clock(semaphore->clock());
  semaphore->join_with_clock(thread_->clock());
  server_->posix()->remove_semaphore(semaphore);
  delete semaphore;
  MESSAGE("Thread %d deleted %s.", thread_->id(), rid_.c_str());
}

void SemaphoreDelete::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (semaphore) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SemaphoreDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: deleting " << rid_;
  result.push_back(stream.str());
  return result;
}

void SemaphorePost::Execute() {
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  ASSERT(semaphore, "Request is missing its semaphore.");
  thread_->join_with_clock(semaphore->clock());
  semaphore->join_with_clock(thread_->clock());
  semaphore->increment_value();
  MESSAGE("Thread %d released %s.", thread_->id(), rid_.c_str());
}

void SemaphorePost::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (semaphore) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SemaphorePost::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: posting on " << rid_;
  result.push_back(stream.str());
  return result;
}

void SemaphoreTryWait::Execute() {
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  ASSERT(semaphore, "Request is missing its semaphore.");
  thread_->join_with_clock(semaphore->clock());
  semaphore->join_with_clock(thread_->clock());
  if (semaphore->value() > 0) { semaphore->decrement_value(); }
  MESSAGE("Thread %d woken up from waiting for %s.",
          thread_->id(), rid_.c_str());
}

void SemaphoreTryWait::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (semaphore) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SemaphoreTryWait::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: trying to wait for " << rid_;
  result.push_back(stream.str());
  return result;
}

void SemaphoreUnlink::Execute() {
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  ASSERT(semaphore, "Request is missing its semaphore.");
  ASSERT(semaphore->named(), "Unlinked semaphore is not named.");
  thread_->join_with_clock(semaphore->clock());
  semaphore->join_with_clock(thread_->clock());
  semaphore->set_marked(true);
  if (semaphore->nrefs() == 0) {
    server_->posix()->remove_semaphore(semaphore);
    delete semaphore;
  }
  MESSAGE("Thread %d unlinked %d.", thread_->id(), rid_.c_str());
}

void SemaphoreUnlink::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (semaphore && semaphore->named()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SemaphoreUnlink::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: unlinking " << rid_;
  result.push_back(stream.str());
  return result;
}

void SemaphoreWait::Execute() {
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  ASSERT(semaphore, "Request is missing its semaphore.");
  ASSERT(semaphore->value() > 0, "Semaphore cannot be decremented.");
  thread_->join_with_clock(semaphore->clock());
  semaphore->join_with_clock(thread_->clock());
  semaphore->decrement_value();
  MESSAGE("Thread %d waited for %s.", thread_->id(), rid_.c_str());
}

void SemaphoreWait::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(rid_);
  Semaphore *semaphore = server_->posix()->semaphore_lookup(rid_);
  if (semaphore && semaphore->value() > 0) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SemaphoreWait::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: waiting for " << rid_;
  result.push_back(stream.str());
  return result;
}

}  // namespace arbiter
}  // namespace dbug
