/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/server.h"

#include <thrift/Thrift.h>

#include <boost/scoped_ptr.hpp>
#include <list>
#include <map>
#include <set>
#include <string>

#include "dbug/arbiter/explorer-connection.h"
#include "dbug/arbiter/handler.h"
#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/mpi.h"
#include "dbug/arbiter/posix.h"
#include "dbug/arbiter/process.h"
#include "dbug/arbiter/request.h"
#include "dbug/arbiter/thread.h"
#include "dbug/common/configuration.h"
#include "dbug/common/lock.h"
#include "dbug/common/system.h"
#include "dbug/common/time.h"
#include "dbug/common/util.h"
#include "dbug/explorer/server.h"

namespace dbug {
namespace arbiter {

void *arbiter_thread(void *args) {
  Server *server = reinterpret_cast<Server *>(args);
  server->Arbitrate();
  return NULL;
}

void *input_thread(void *args) {
  Server *server = reinterpret_cast<Server *>(args);
  server->InputLoop();
  return NULL;
}

// Server::EventHandler class

Server::EventHandler::EventHandler(Server *server) : server_(server) {}

void Server::EventHandler::preServe(void) { server_->RegisterArbiter(); }

// public methods

// Server::Lock class
Server::Lock::Lock(Server *server) : server_(server) {
  System::MutexLock(&server_->mutex_);
}

Server::Lock::~Lock() {
  System::CVarSignal(&server_->cvar_);
  System::MutexUnlock(&server_->mutex_);
}

// Server::Statistics class

Server::Statistics::Statistics()
    : nprocesses_(0), nrequests_(0), nthreads_(0) {}

Server::Statistics::~Statistics() {
  for (const std::string &line : this->ToString()) {
    MESSAGE("%s", line.c_str());
  }
}

// public methods

std::list<std::string> Server::Statistics::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  for (std::pair<std::string, int> pair : function_calls_) {
    stream << "  " << pair.first << ": " << pair.second;
    result.push_back(stream.str());
    stream.str(std::string());
  }
  stream << "  number of processes: " << nprocesses_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of requests: " << nrequests_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of threads: " << nthreads_;
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// Server class

Server::Server(std::string explorer_hostname, int explorer_port, int port)
    : choice_(-1),
      end_nthreads_(0),
      explorer_hostname_(explorer_hostname),
      explorer_port_(explorer_port),
      last_choice_(-1),
      mode_(ROUND_ROBIN),
#ifdef ENABLE_MPI
      mpi_(new mpi::MPI()),
#else
      mpi_(NULL),
#endif
      nenabled_(0),
      nnondet_events_(0),
      nscheduled_(0),
      nthreads_(0),
      port_(port),
      posix_(new posix::POSIX()),
      random_buffer_(new System::RandomData()),
      random_seed_(1),
      random_state_(new char[Configuration::random_state_length()]()),
      start_nthreads_(1),
      terminate_(false) {
  System::CVarInit(&cvar_, NULL);
  System::CVarInit(&cancelation_cvar_, NULL);
  System::CVarInit(&execution_cvar_, NULL);
  System::MutexInit(&mutex_, NULL);
  System::ReentrantInitState(random_seed_,
                             random_state_.get(),
                             Configuration::random_state_length(),
                             random_buffer_.get());
  strategy_file_ = System::FileOpen("dbug-arbiter.strategy", "w+");
}

Server::~Server() {
  ASSERT(requests_.empty(), "Shutting down arbiter with pending requests.");
  UnregisterArbiter();
  System::FileClose(strategy_file_);
  for (Process *process : process_forest_) { delete process; }
  System::MutexDestroy(&mutex_);
  System::CVarDestroy(&execution_cvar_);
  System::CVarDestroy(&cancelation_cvar_);
  System::CVarDestroy(&cvar_);
}

// accessors

bool Server::has_nondet_events() const {
  return ((unsigned) nscheduled_ < strategy_.size() &&
          (unsigned) nnondet_events_ <
          strategy_[nscheduled_].nondet_events.size());
}

Server::Mode Server::mode() const { return mode_; }

mpi::MPI *Server::mpi() const { return mpi_.get(); }

int Server::nthreads() const { return nthreads_; }

int Server::nenabled() const { return nenabled_; }

int Server::port() const { return port_; }

posix::POSIX *Server::posix() const { return posix_.get(); }

Process *Server::process_lookup(int pid) const {
  for (Process *tree : process_forest_) {
    Process *process = InternalProcessLookup(pid, tree);
    if (process != NULL) { return process; }
  }
  return NULL;
}

std::list<Request *> Server::requests() const { return requests_; }

std::set<int> Server::sleep_set() const { return sleep_set_; }

std::vector<dbug::thrift::Move> Server::strategy() const { return strategy_; }

bool Server::terminate() const { return terminate_; }

Thread *Server::thread_lookup(int id) const {
  for (const Process *tree : process_forest_) {
    Thread *thread = InternalThreadLookup(id, tree);
    if (thread != NULL) { return thread; }
  }
  return NULL;
}

// mutators

void Server::add_move(dbug::thrift::Move move) { strategy_.push_back(move); }

void Server::add_process(Process *process) {
  statistics_.nprocesses_++;
  process_forest_.push_back(process);
}

void Server::add_request(Request *request) {
  if (statistics_.function_calls_.find(request->function()) ==
      statistics_.function_calls_.end()) {
    statistics_.function_calls_[request->function()] = 1;
  } else {
    statistics_.function_calls_[request->function()]++;
  }
  statistics_.nrequests_++;
  requests_.push_back(request);
}

void Server::add_to_sleep_set(int tid) { sleep_set_.insert(tid); }

void Server::clear_sleep_set() { sleep_set_.clear(); }

int Server::consume_nondet_event() {
  assert((unsigned) nscheduled_ < strategy_.size());
  assert((unsigned) nnondet_events_ <
         strategy_[nscheduled_].nondet_events.size());
  return strategy_[nscheduled_].nondet_events[nnondet_events_++];
}

void Server::decrement_nenabled() { nenabled_--; }

void Server::decrement_nthreads() { nthreads_--; }

void Server::increment_nenabled() { nenabled_++; }

void Server::increment_nthreads() {
  statistics_.nthreads_++;
  nthreads_++;
}

void Server::record_nondet_event(int event) { nondet_events_.push_back(event); }

void Server::remove_process(Process *process) {
  process_forest_.remove(process);
}

void Server::remove_request(Request *request) { requests_.remove(request); }

void Server::set_end_nthreads(int nthreads) { end_nthreads_ = nthreads; }

void Server::set_explorer_hostname(std::string explorer_hostname) {
  explorer_hostname_ = explorer_hostname;
}

void Server::set_explorer_port(int explorer_port) {
  explorer_port_ = explorer_port;
}

void Server::set_mode(Mode mode) { mode_ = mode; }

void Server::set_port(int port) { port_ = port; }

void Server::set_random_seed(unsigned int random_seed) {
  random_seed_ = random_seed;
  System::ReentrantSeedRandom(random_seed_, random_buffer_.get());
}

void Server::set_start_nthreads(int nthreads) { start_nthreads_ = nthreads; }

void Server::set_terminate(bool terminate) { terminate_ = terminate; }

void Server::set_thrift_server(TThreadedServer *thrift_server) {
  thrift_server_ = thrift_server;
}

// public methods

void Server::Arbitrate() {
  dbug::MutexLock lock(&mutex_);

  // Wait for <start_nthreads_> threads to appear in the system.
  while (requests_.size() < (unsigned) start_nthreads_) {
    // Wait to be signaled or timeout.
    (mode_ == INTERACTIVE ? Wait() : TimedWait());
    // Terminate if a shutdown has been requested.
    if (terminate_) {
      GenerateRequestGroups();
      Log();
      TerminateSystem();
      return;
    }
  }

  // Main scheduling loop.
  while (true) {
    PrintProcesses(Logging::DEBUG);
    PrintRequestGroups(Logging::DEBUG);
    PrintRequests(Logging::DEBUG);
    PrintResources(Logging::DEBUG);
    // If the terminate flag has been set, cleanly terminate the execution.
    if (terminate_) {
      GenerateRequestGroups();
      Log();
      TerminateSystem();
      break;
    }
    // Check if a request group can be executed. If so, log the
    // decision and the current context and execute the request group.
    if (requests_.size() > 0 && CanExecuteRequest()) {
      if (nenabled_ > 0) {
        // Select a request to execute and record this choice.
        Request::GroupSet::iterator it = SelectRequestGroup();
        if (choice_ == -1) { continue; }
        System::FilePrint(strategy_file_, "%d\n", choice_);
        Log();

        // Schedule all requests from the request group.
        Request::Group group = *it;
        for (const Request::Resolved &resolved : group.resolved_set()) {
          Request *request = resolved.request;
          request->set_nondet_choices(resolved.choices);
          request->WakeUp();
          WaitForExecution();
          // Schedule requests one by one.
          while (group.serialize() && !IsSystemQuiescent()) { Wait(); }
        }
      } else {
        if (nthreads_ != end_nthreads_) {
          WARNING("Encountered a concurrency issue.");
          PrintRequests(Logging::WARNING);
        }
        terminate_ = true;
      }
    } else {
      (mode_ == INTERACTIVE ? Wait() : TimedWait());
    }
  }
}

bool Server::CreateThread(const int parent_tid,
                          const bool create_process,
                          int *pid,
                          int *tid) {
  Thread *new_thread = new Thread();
  if (parent_tid != 0) {
    Thread *parent_thread = thread_lookup(parent_tid);
    if (parent_thread == NULL) {
      WARNING("Non-existing parent thread.");
      *pid = 0;
      *tid = 0;
      delete new_thread;
      return false;
    } else {
      // If registered by a parent thread, copy the vector clock.
      new_thread->set_clock(parent_thread->clock());
      if (create_process) {
        // New process in existing process tree is created.
        Process *parent_process = parent_thread->pgrp();
        Process *new_process = new Process();
        new_process->set_parent(parent_process);
        parent_process->add_child(new_process);
        new_process->add_thread(new_thread);
        new_thread->set_pgrp(new_process);
      } else {
        // New thread of existing process is created.
        parent_thread->pgrp()->add_thread(new_thread);
        new_thread->set_pgrp(parent_thread->pgrp());
      }
    }
  } else {
    NewProcessNotification();
    // New root of a process tree is created.
    Process *new_process = new Process();
    new_process->set_parent(NULL);
    new_process->add_thread(new_thread);
    new_thread->set_pgrp(new_process);
    add_process(new_process);
  }
  increment_nthreads();
  *pid = new_thread->pgrp()->id();
  *tid = new_thread->id();
  return true;
}

void Server::InputLoop() {
  while (1) {
    int ret = System::Input("%d", &choice_);
    if (ret == 1) {
      if (nenabled_ > 0 && (choice_ < 1 || choice_ > nenabled_)) {
        WARNING("The value needs to be in the interval [1,%d].", nenabled_);
      } else {
        System::CVarSignal(&cvar_);
      }
    } else {
      char junk;
      System::Input("%c", &junk);
    }
  }
}

void Server::LoadStrategyFromFile(const char *filename) {
  System::File *strategy = System::FileOpen(filename, "r");
  int nmoves;
  System::FileScan(strategy, "%d\n", &nmoves);
  strategy_.clear();
  for (int i = 0; i < nmoves; i++) {
    dbug::thrift::Move move;
    System::FileScan(strategy, "%d\n", &move.choice_id);
    strategy_.push_back(move);
  }
  System::FileClose(strategy);
}

void Server::PrintRequestGroups(Logging::Verbosity verbosity) {
  const Logging::Config *config = Logging::get_channel(Logging::ARBITER);
  if (config->verbosity() < verbosity) { return; }
  int i = 1;
  LOG(verbosity, "Request groups:");
  for (const Request::Group group : groups_) {
    LOG(verbosity, "#%d:", i++);
    for (std::string line : group.ToString()) {
      LOG(verbosity, "%s", line.c_str());
    }
  }
}

void Server::PrintRequests(Logging::Verbosity verbosity) {
  const Logging::Config *config = Logging::get_channel(Logging::ARBITER);
  if (config->verbosity() < verbosity) { return; }
  LOG(verbosity, "Requests:");
  for (const Request *request : requests_) {
    for (std::string line : request->ToString()) {
      LOG(verbosity, "%s", line.c_str());
    }
  }
}

void Server::PrintResources(Logging::Verbosity verbosity) {
  const Logging::Config *config = Logging::get_channel(Logging::ARBITER);
  if (config->verbosity() < verbosity) { return; }
  LOG(verbosity, "Resources:");
  for (const std::string &line : posix_->ToString()) {
    LOG(verbosity, "%s", line.c_str());
  }
}

void Server::PrintProcessTree(Logging::Verbosity verbosity,
                              const Process *process) {
  const Logging::Config *config = Logging::get_channel(Logging::ARBITER);
  if (config->verbosity() < verbosity) { return; }
  for (std::string line : process->ToString()) {
    LOG(verbosity, "%s", line.c_str());
  }
  for (const Process *child : process->children()) {
    PrintProcessTree(verbosity, child);
  }
}

void Server::PrintProcesses(Logging::Verbosity verbosity) {
  const Logging::Config *config = Logging::get_channel(Logging::ARBITER);
  if (config->verbosity() < verbosity) { return; }
  LOG(verbosity, "Processes:");
  for (const Process *tree : process_forest_) {
    PrintProcessTree(verbosity, tree);
  }
}

void Server::RequestCanceled() { System::CVarSignal(&cancelation_cvar_); }

void Server::RequestExecuted() { System::CVarSignal(&execution_cvar_); }

void Server::Stop() { if (thrift_server_) { thrift_server_->stop(); } }

void Server::TerminateThread(Thread *thread, bool cleanup) {
  switch (thread->state()) {
    case Thread::RUNNING:
      thread->set_state(Thread::JOINABLE);
      break;
    case Thread::DETACHED:
      thread->set_state(Thread::TERMINATED);
      break;
    default:
      WARNING("Terminating a thread with an unexpected state: %u.",
              thread->state());
  }

  // Update the number of threads in the system and if there are no
  // threads left in the system, set the termination flag.
  if (--nthreads_ == 0) { terminate_ = true; }

  if (cleanup) {
    // If this thread has an outstanding request we need to remove it.
    //
    // TODO(jsimsa): This could be made O(1) by keeping a pointer to
    // the request in the thread.
    for (Request *request : requests_) {
      if (request->thread() == thread) {
        request->Cancel();
        WaitForCancelation();
        WARNING("Terminating thread %d with a pending request.", thread->id());
        break;
      }
    }

    // If the parent process is not already terminated and this was
    // the last running thread of the process, terminate the parent
    // process.
    Process *process = thread->pgrp();
    if (process->state() == Process::RUNNING ||
        process->state() == Process::CONTINUED) {
      bool still_running = false;
      for (const Thread *sibling : thread->pgrp()->threads()) {
        if (sibling->state() == Thread::RUNNING ||
            sibling->state() == Thread::DETACHED) {
          still_running = true;
          break;
        }
      }
      if (!still_running) { process->set_state(Process::TERMINATED); }
    }
  }
}

void Server::WaitForCancelation() {
  System::CVarWait(&cancelation_cvar_, &mutex_);
}

void Server::WaitForExecution() {
  System::CVarWait(&execution_cvar_, &mutex_);
}

void Server::WaitingRequest(System::CVar *cond) {
  System::CVarSignal(&cvar_);
  System::CVarWait(cond, &mutex_);
}

// private methods

bool Server::CanExecuteRequest() {
  GenerateRequestGroups();
  // If we are still waiting to execute the previous request, do not
  // schedule another one.
  switch (mode_) {
    case INTERACTIVE:
      // If strategy exists, follow it.
      if ((unsigned) nscheduled_ < strategy_.size() && IsSystemQuiescent()) {
        return true;
      }
      // If user inputed a valid choice, a request can be executed.
      if (choice_ == -1 && IsSystemQuiescent()) {
        PrintRequestGroups(Logging::MESSAGE);
        System::Output("Identify the group to advance: \n");
      }
      return (choice_ != -1);
    case ROUND_ROBIN:
    case RANDOM:
      // If the system is quiescent, a request can be executed.
      return IsSystemQuiescent();
    default:
      WARNING("Unknown server mode: %u", mode_);
      break;
  }
  return false;
}

void Server::CheckProcessTree(const Process *process) {
  if (process->state() != Process::TERMINATED &&
      process->state() != Process::SIGNALED) {
    std::stringstream stream;
    stream << "/proc/" << process->pid() << "/stat";
    if (System::FileExists(stream.str().c_str())) {
      const int kBufferSize = 256;
      char buffer[kBufferSize];
      System::File *file = System::FileOpen(stream.str().c_str(), "r");
      System::FileGetString(buffer, kBufferSize, file);
      std::stringstream tokens(buffer);
      int pid;
      tokens >> pid;
      std::string name;
      tokens >> name;
      std::string status;
      tokens >> status;
      if (status == "Z") {
        WARNING("Process %d has crashed.", process->pid());
        for (Thread *thread : process->threads()) {
          if (thread->state() == Thread::RUNNING ||
              thread->state() == Thread::DETACHED) {
            TerminateThread(thread, true);
          }
        }
      }
      System::FileClose(file);
    } else {
      // TODO(jsimsa): This is a rare race that might occur when a
      // thread registers but its process state is not updated before
      // TimedWait() wakes up.
      WARNING("The file %s does not exist.", stream.str().c_str());
    }
  }
  for (const Process *child : process->children()) {
    CheckProcessTree(child);
  }
}

void Server::CheckSystem() {
  for (const Process *tree : process_forest_) { CheckProcessTree(tree); }
}

void Server::GenerateRequestGroups() {
  Request::Set requests;
  for (Request *request : requests_) {
    if (request->thread()->pgrp()->state() == Process::RUNNING ||
        request->thread()->pgrp()->state() == Process::CONTINUED) {
      requests.insert(request);
    }
  }
  nenabled_ = 0;
  groups_.clear();
  if (!requests.empty()) {
    Request *request = *(requests.begin());
    request->Test(requests, &groups_);
  }
}

Process *Server::InternalProcessLookup(int pid, Process *process) const {
  if (process->pid() == pid) { return process; }
  for (Process *child : process->children()) {
    Process *process = InternalProcessLookup(pid, child);
    if (process != NULL) { return process; }
  }
  return NULL;
}

Thread *Server::InternalThreadLookup(int tid, const Process *process) const {
  for (Thread *thread : process->threads()) {
    if (thread->id() == tid) { return thread; }
  }
  for (const Process *child : process->children()) {
    Thread *thread = InternalThreadLookup(tid, child);
    if (thread != NULL) { return thread; }
  }
  return NULL;
}

bool Server::IsProcessQuiescent(const Process *process) {
  if (process->state() != Process::RUNNING &&
      process->state() != Process::CONTINUED)  { return true; }
  for (const Thread *thread : process->threads()) {
    bool result = IsThreadQuiescent(thread);
    if (!result) return false;
  }
  return true;
}

bool Server::IsProcessTreeQuiescent(const Process *process) {
  bool result = IsProcessQuiescent(process);
  if (!result) { return false; }
  for (const Process *child : process->children()) {
    result = IsProcessTreeQuiescent(child);
    if (!result) return false;
  }
  return true;
}

bool Server::IsSystemQuiescent() {
  for (const Process *tree : process_forest_) {
    bool result = IsProcessTreeQuiescent(tree);
    if (!result) return false;
  }
  return true;
}

bool Server::IsThreadQuiescent(const Thread *thread) {
  if (thread->state() != Thread::RUNNING &&
      thread->state() != Thread::DETACHED) { return true; }
  for (const Request *request : requests_) {
    if (request->thread() == thread) return true;
  }
  return false;
}

void Server::Log() {
  dbug::thrift::Node node;
  node.choice_id = choice_;
  int i = 1;
  for (const Request::Group group : groups_) {
    dbug::thrift::Choice choice;
    choice.id = i++;
    for (const Request::Resolved resolved : group.resolved_set()) {
      dbug::thrift::Thread thread;
      thread.backtrace = resolved.request->backtrace();
      thread.clock = resolved.request->thread()->clock();
      thread.fname = resolved.request->function();
      thread.id = resolved.request->thread()->id();
      choice.threads.push_back(thread);
    }
    choice.enabled = group.enabled();
    choice.read_set = group.read_set();
    choice.write_set = group.write_set();
    node.choices.push_back(choice);
  }
  node.nondet_events = nondet_events_;
  nondet_events_.clear();
  branch_.push_back(node);
  if (choice_ != -1) {
    switch (mode_) {
      case INTERACTIVE:
        DEBUG("You selected request group %d.", choice_);
        break;
      default:
        if ((unsigned) nscheduled_ < strategy_.size()) {
          DEBUG("Strategy selected request group %d.", choice_);
        } else {
          DEBUG("Arbiter selected request group %d.", choice_);
        }
        break;
    }
  }
  ResetSchedulingState();
}

void Server::NewProcessNotification() {
  try {
    boost::scoped_ptr<ExplorerConnection>
        connection(new ExplorerConnection(explorer_hostname_, explorer_port_));
    connection->client()->NewProcessNotification();
  } catch (apache::thrift::TException & tx) {
    WARNING("Connection with the explorer failed: %s", tx.what());
  }
}

void Server::RegisterArbiter() {
  try {
    std::vector<dbug::thrift::Move> register_return;
    boost::scoped_ptr<ExplorerConnection>
        connection(new ExplorerConnection(explorer_hostname_, explorer_port_));
    connection->client()->RegisterArbiter(register_return);
    strategy_.clear();
    for (unsigned i = 0; i < register_return.size(); i++) {
      strategy_.push_back(register_return[i]);
      DEBUG("Adding strategy move: %d", register_return[i].choice_id);
    }
  } catch (apache::thrift::TException & tx) {
    WARNING("Connection with the explorer failed: %s", tx.what());
  }
}

void Server::ResetSchedulingState() {
  nscheduled_++;
  nnondet_events_ = 0;
  choice_ = -1;
}

Request::GroupSet::iterator Server::SelectRequestGroup() {
  Request::GroupSet::iterator it = groups_.begin();
  // If a valid strategy exists, select a request group according to
  // the strategy if a valid strategy exists.
  if ((unsigned) nscheduled_ < strategy_.size()) {
    int choice = strategy_[nscheduled_].choice_id;
    for (int i = 1; i < choice && it != groups_.end(); ++i, ++it) ;
    if (it == groups_.end() || !it->enabled()) {
      WARNING("Invalid request group choice: %d", choice);
      PrintRequestGroups(Logging::WARNING);
      terminate_ = true;
      choice_ = -1;
    } else {
      choice_ = choice;
    }
  } else {
    switch (mode_) {
      case INTERACTIVE:
        // Use user input to select a request group.
        for (int i = 1; i < choice_ && it != groups_.end(); ++i, ++it) {}
        if (it == groups_.end() || !it->enabled()) {
          WARNING("Invalid request group choice: %d", choice_);
          choice_ = -1;
        }
        break;
      case ROUND_ROBIN:
        // Select the next request group to be scheduled using a
        // deterministic round-robin scheduler.
        for (choice_ = 1;
             (choice_ <= last_choice_ && it != groups_.end());
             ++it, ++choice_) {}
        if (it == groups_.end() || !it->enabled()) {
          it = groups_.begin();
          choice_ = 1;
        }
        break;
      case RANDOM:
        // Select the next request group to be scheduled using a
        // deterministic pseudo-random scheduler.
        int choice, i;
        System::ReentrantRandom(random_buffer_.get(), &choice, nenabled_);
        for (i = 0, choice_ = 1;
             (i < choice && it != groups_.end());
             ++i, ++it, ++choice_) {}
        ASSERT(it != groups_.end() && it->enabled(),
               "Incorrect count of enabled request groups.");
        break;
      default:
        WARNING("Unknown server mode: %u", mode_);
        break;
    }
  }
  return it;
}

void Server::TerminateProcessTree(const Process *process) {
  for (Thread *thread : process->threads()) {
    if (thread->state() == Thread::RUNNING ||
        thread->state() == Thread::DETACHED) {
      TerminateThread(thread, true);
    }
  }
  for (const Process *child : process->children()) {
    TerminateProcessTree(child);
  }
}

void Server::TerminateSystem() {
  for (const Process *tree : process_forest_) { TerminateProcessTree(tree); }
}

void Server::TimedWait() {
  System::TimeValue now;
  System::TimeSpec timeout = Configuration::arbiter_timeout();
  System::TimeSpec abs_time;
  System::GetTimeOfDay(&now, NULL);
  int nsecs = (now.tv_usec * Time::kU2N + timeout.tv_nsec);
  abs_time.tv_sec = (now.tv_sec + timeout.tv_sec) + nsecs / Time::kS2N;
  abs_time.tv_nsec = nsecs % Time::kS2N;
  int ret = System::CVarTimedWait(&cvar_, &mutex_, &abs_time);
  if (ret == ETIMEDOUT) { CheckSystem(); }
}

void Server::UnregisterArbiter() {
  try {
    boost::scoped_ptr<ExplorerConnection>
        connection(new ExplorerConnection(explorer_hostname_, explorer_port_));
    connection->client()->UnregisterArbiter(branch_);
  } catch (apache::thrift::TException & tx) {
    WARNING("Connection with the explorer failed: %s", tx.what());
  }
}

void Server::Wait() { System::CVarWait(&cvar_, &mutex_); }

}  // namespace arbiter
}  // namespace dbug
