/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/common/configuration.h"

#include <libconfig.h++>
#include <string>

#include "dbug/common/logging.h"
#include "dbug/common/system.h"

#define DEBUG(args...) DEBUG_ON_CHANNEL(Logging::SYSTEM, ##args)
#define ERROR(args...) ERROR_ON_CHANNEL(Logging::SYSTEM, ##args)
#define MESSAGE(args...) MESSAGE_ON_CHANNEL(Logging::SYSTEM, ##args)
#define WARNING(args...) WARNING_ON_CHANNEL(Logging::SYSTEM, ##args)

using libconfig::Config;
using libconfig::FileIOException;
using libconfig::ParseException;

namespace dbug {

void Configuration::Initialize() {
  std::string config_file = PREFIX "dbug.conf";
  if (System::FileExists("./dbug.conf")) { config_file = "./dbug.conf"; }
  config_ = new Config();
  try {
    config_->readFile(config_file.c_str());
  } catch (ParseException e) {
    Config *tmp = config_;
    config_ = NULL;
    WARNING("Configuration file %s cannot be parsed: line:%d error:%s.",
            config_file.c_str(), e.getLine(), e.getError());
    delete tmp;
    config_ = NULL;
  } catch (FileIOException e) {
    Config *tmp = config_;
    config_ = NULL;
    WARNING("Configuration file %s cannot be read.", config_file.c_str());
    delete tmp;
  }
}

void Configuration::Finalize() {
  if (config_ != NULL) { delete config_; }
}

std::string Configuration::arbiter_hostname() {
  if (config_ && config_->exists("arbiter.hostname")) {
    std::string value;
    config_->lookupValue("arbiter.hostname", value);
    return value;
  }
  static std::string hostname = "localhost";
  return hostname;
}

int Configuration::arbiter_port() {
  if (config_ && config_->exists("arbiter.port")) {
    int value;
    config_->lookupValue("arbiter.port", value);
    return value;
  }
  static int port = 21008;
  return port;
}

System::TimeSpec Configuration::arbiter_timeout() {
  if (config_ &&
      config_->exists("arbiter.timeout.seconds") &&
      config_->exists("arbiter.timeout.nanoseconds")) {
    static System::TimeSpec value;
    config_->lookupValue("arbiter.timeout.seconds", value.tv_sec);
    config_->lookupValue("arbiter.timeout.seconds", value.tv_nsec);
    return value;
  }
  static System::TimeSpec timeout;
  timeout.tv_sec = 1;
  timeout.tv_nsec = 0;
  return timeout;
}

Logging::Verbosity Configuration::arbiter_verbosity() {
  if (config_ && config_->exists("arbiter.verbosity")) {
    int value;
    config_->lookupValue("arbiter.verbosity", value);
    return (Logging::Verbosity) value;
  }
  static Logging::Verbosity verbosity = Logging::WARNING;
  return verbosity;
}

std::string Configuration::explorer_hostname() {
  if (config_ && config_->exists("explorer.hostname")) {
    std::string value;
    config_->lookupValue("explorer.hostname", value);
    return value;
  }
  static std::string hostname = "localhost";
  return hostname;
}

int Configuration::explorer_port() {
  if (config_ && config_->exists("explorer.port")) {
    int value;
    config_->lookupValue("explorer.port", value);
    return value;
  }
  static int port = 21005;
  return port;
}

Logging::Verbosity Configuration::explorer_verbosity() {
  if (config_ && config_->exists("explorer.verbosity")) {
    int value;
    config_->lookupValue("explorer.verbosity", value);
    return (Logging::Verbosity) value;
  }
  static Logging::Verbosity verbosity = Logging::MESSAGE;
  return verbosity;
}

Logging::Verbosity Configuration::interposition_verbosity() {
  if (config_ && config_->exists("interposition.verbosity")) {
    int value;
    config_->lookupValue("interposition.verbosity", value);
    return (Logging::Verbosity) value;
  }
  static Logging::Verbosity verbosity = Logging::WARNING;
  return verbosity;
}

int Configuration::max_backtrace_length() {
  if (config_ && config_->exists("common.max_backtrace_length")) {
    int value;
    config_->lookupValue("common.max_backtrace_length", value);
    return value;
  }
  static int n = 16;
  return n;
}

int Configuration::max_interfaces() {
  if (config_ && config_->exists("common.max_interfaces")) {
    int value;
    config_->lookupValue("common.max_interfaces", value);
    return value;
  }
  static int n = 32;
  return n;
}

int Configuration::random_state_length() {
  if (config_ && config_->exists("common.random_state_length")) {
    int value;
    config_->lookupValue("common.random_state_length", value);
    return value;
  }
  static int n = 32;
  return n;
}

Logging::Verbosity Configuration::verbosity() {
  if (config_ && config_->exists("common.verbosity")) {
    int value;
    config_->lookupValue("common.verbosity", value);
    return (Logging::Verbosity) value;
  }
  static Logging::Verbosity verbosity = Logging::WARNING;
  return verbosity;
}

Config *Configuration::config_ = NULL;

}  // namespace dbug
