/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): This class is not thread-safe. However, using
// pthreads to coordinate its accessors and mutators causes infinite
// recursion when dbug interposes on pthreads.

#include "dbug/common/logging.h"

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>

#include <map>

#include "dbug/common/system.h"

namespace dbug {

// Logging::Config class

Logging::Config::Config(System::File *log_file, Verbosity verbosity)
    : log_file_(log_file), verbosity_(verbosity) {}

Logging::Config::~Config() { if (log_file_) { System::FileClose(log_file_); } }

// accessors

System::File *Logging::Config::log_file() const { return log_file_; }

Logging::Verbosity Logging::Config::verbosity() const { return verbosity_; }

// Logging class

void Logging::Initialize() {
  assert(!channels_);
  channels_ = new std::map<Channel, Config *>();
}

void Logging::Finalize() {
  for (std::pair<Channel, Config *> pair : (*channels_)) {
    Config *config = pair.second;
    delete config;
  }
  delete channels_;
  channels_ = NULL;
}

// accessors

// Returns a configuration for a logging <channel> if it exists and
// NULL otherwise. Notably, this function can be called before the
// <channels_> map is allocated or after it is deallocated. If that
// happens, the function returns NULL.
const Logging::Config *Logging::get_channel(Channel channel) {
  if (!channels_) { return NULL; }
  std::map<Channel, Config *>::iterator it = channels_->find(channel);
  if (it != channels_->end()) { return it->second; }
  return NULL;
}

// mutators

void Logging::add_channel(Channel channel,
                          System::File *log_file,
                          Verbosity verbosity) {
  assert(channels_);
  (*channels_)[channel] = new Config(log_file, verbosity);
}

// public methods

void Logging::Assert(const char *filename,
                     const char *function,
                     int line,
                     Channel channel,
                     bool condition,
                     const char *assert,
                     const char *format, ...) {
  if (!condition) {
    const Config *config = get_channel(channel);
    if (config) {
      std::string prefix = "\e[1;31m";
      std::string suffix = "\e[0m";
      System::File *log_file = stderr;
      std::stringstream fmt_stream;
      fmt_stream << prefix << "[DBUG ASSERT] " << format
                 << " assert = " << assert << ""
                 << " " << FormatLocation(filename, function, line)
                 << "\n" << suffix;
      va_list ap;
      va_start(ap, format);
      vfprintf(log_file, fmt_stream.str().c_str(), ap);
      va_end(ap);
      System::FileFlush(log_file);
    }
    System::Exit(-1);
  }
}

void Logging::Error(const char *filename,
                    const char *function,
                    int line,
                    Channel channel,
                    const char *format, ...) {
  const Config *config = get_channel(channel);
  if (config && config->verbosity() >= ERROR) {
    std::string prefix = (config->log_file() ? "" : "\e[1;31m");
    std::string suffix = (config->log_file() ? "" : "\e[0m");
    System::File *log_file = (config->log_file() ? config->log_file() : stderr);
    std::stringstream fmt_stream;
    fmt_stream << prefix << "[DBUG ERROR] " << format;
    if (config->verbosity() >= DEBUG) {
      fmt_stream << " " << FormatLocation(filename, function, line);
    }
    fmt_stream << "\n" << suffix;
    va_list ap;
    va_start(ap, format);
    vfprintf(log_file, fmt_stream.str().c_str(), ap);
    va_end(ap);
    System::FileFlush(log_file);
  }
  System::Exit(-1);
}

void Logging::Warning(const char *filename,
                      const char *function,
                      int line,
                      Channel channel,
                      const char *format, ...) {
  const Config *config = get_channel(channel);
  if (config && config->verbosity() >= WARNING) {
    std::string prefix = (config->log_file() ? "" : "\e[1;33m");
    std::string suffix = (config->log_file() ? "" : "\e[0m");
    System::File *log_file = (config->log_file() ? config->log_file() : stderr);
    std::stringstream fmt_stream;
    fmt_stream << prefix << "[DBUG WARNING] " << format;
    if (config->verbosity() >= DEBUG) {
      fmt_stream << " " << FormatLocation(filename, function, line);
    }
    fmt_stream << "\n" << suffix;
    va_list ap;
    va_start(ap, format);
    vfprintf(log_file, fmt_stream.str().c_str(), ap);
    va_end(ap);
    System::FileFlush(log_file);
  }
}

void Logging::Message(const char *filename,
                      const char *function,
                      int line,
                      Channel channel,
                      const char *format, ...) {
  const Config *config = get_channel(channel);
  if (config && config->verbosity() >= MESSAGE) {
    std::string prefix = (config->log_file() ? "" : "\e[1;32m");
    std::string suffix = (config->log_file() ? "" : "\e[0m");
    System::File *log_file = (config->log_file() ? config->log_file() : stdout);
    std::stringstream fmt_stream;
    fmt_stream << prefix << "[DBUG MESSAGE] " << format;
    if (config->verbosity() >= DEBUG) {
      fmt_stream << " " << FormatLocation(filename, function, line);
    }
    fmt_stream << "\n" << suffix;
    va_list ap;
    va_start(ap, format);
    vfprintf(log_file, fmt_stream.str().c_str(), ap);
    va_end(ap);
    System::FileFlush(log_file);
  }
}

void Logging::Debug(const char *filename,
                    const char *function,
                    int line,
                    Channel channel,
                    const char *format, ...) {
  const Config *config = get_channel(channel);
  if (config && config->verbosity() >= DEBUG) {
    std::string prefix = (config->log_file() ? "" : "\e[1;36m");
    std::string suffix = (config->log_file() ? "" : "\e[0m");
    System::File *log_file = (config->log_file() ? config->log_file() : stdout);
    std::stringstream fmt_stream;
    fmt_stream << prefix
               << "[DBUG DEBUG] " << format << " "
               << FormatLocation(filename, function, line) << "\n"
               << suffix;
    va_list ap;
    va_start(ap, format);
    vfprintf(log_file, fmt_stream.str().c_str(), ap);
    va_end(ap);
    System::FileFlush(log_file);
  }
}

void Logging::External(const char *filename,
                       const char *function,
                       int line,
                       Channel channel,
                       const char *format1,
                       const char *format2, ...) {
  const Config *config = get_channel(channel);
  if (config && config->verbosity() >= DEBUG) {
    std::string prefix = (config->log_file() ? "" : "\e[1;36m");
    std::string suffix = (config->log_file() ? "" : "\e[0m");
    System::File *log_file = (config->log_file() ? config->log_file() : stdout);
    std::stringstream fmt_stream;
    fmt_stream << prefix
               << "[DBUG DEBUG] External call to " << function
               << "(" << format1 << ") returned " << format2 << " "
               << FormatLocation(filename, function, line) << ".\n"
               << suffix;
    va_list ap;
    va_start(ap, format2);
    vfprintf(log_file, fmt_stream.str().c_str(), ap);
    va_end(ap);
    System::FileFlush(log_file);
  }
}

void Logging::Internal(const char *filename,
                       const char *function,
                       int line,
                       Channel channel,
                       const char *format1,
                       const char *format2, ...) {
  const Config *config = get_channel(channel);
  if (config && config->verbosity() >= DEBUG) {
    std::string prefix = (config->log_file() ? "" : "\e[1;36m");
    std::string suffix = (config->log_file() ? "" : "\e[0m");
    System::File *log_file = (config->log_file() ? config->log_file() : stdout);
    std::stringstream fmt_stream;
    fmt_stream << prefix
               << "[DBUG DEBUG] Internal call to " << function
               << "(" << format1 << ") returned " << format2 << " "
               << FormatLocation(filename, function, line) << ".\n"
               << suffix;
    va_list ap;
    va_start(ap, format2);
    vfprintf(log_file, fmt_stream.str().c_str(), ap);
    va_end(ap);
    System::FileFlush(log_file);
  }
}

std::string Logging::FormatLocation(const char *filename,
                                    const char *function,
                                    int line) {
  std::string file_str = filename;
  std::string func_str = function;
  std::stringstream stream;
  stream << line;
  std::string line_str = stream.str();
  std::string result =
      "[" + func_str + "() at " + file_str + ":" + line_str + "]";
  stream.str(std::string());
  stream.width(kMaxLocationLength);
  stream.fill(' ');
  stream << result;
  return result;
}

// private static fields

std::map<Logging::Channel, Logging::Config *> *Logging::channels_ = NULL;

}  // namespace dbug
