/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/common/system.h"

#include <assert.h>
#include <bfd.h>
#include <dlfcn.h>
#include <errno.h>
#include <execinfo.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <string>
#include <vector>

#include "dbug/common/configuration.h"
#include "dbug/common/logging.h"

#define DEBUG(args...) DEBUG_ON_CHANNEL(Logging::SYSTEM, ##args)
#define ERROR(args...) ERROR_ON_CHANNEL(Logging::SYSTEM, ##args)
#define MESSAGE(args...) MESSAGE_ON_CHANNEL(Logging::SYSTEM, ##args)
#define WARNING(args...) WARNING_ON_CHANNEL(Logging::SYSTEM, ##args)

namespace dbug {

void System::Initialize() {
  Logging::add_channel(Logging::SYSTEM, NULL, Configuration::verbosity());
}

void System::Finalize() {}

// Process

System::Process System::Fork() {
  Process ret = fork();
  if (ret == -1) { ERROR("Fork() failed with: %s", strerror(errno)); }
  return ret;
}

System::Process System::Wait(Process process, int *status, int options) {
  Process ret = waitpid(process, status, options);
  if (ret == -1) { ERROR("Wait() failed with: %s", strerror(errno)); }
  return ret;
}

int System::Kill(Process process, int signal) {
  int ret = kill(process, signal);
  if (ret == -1) { ERROR("Kill() failed with: %s", strerror(errno)); }
  return ret;
}

int System::AtExit(void (*function)(void)) {
  int ret = atexit(function);
  if (ret != 0) { ERROR("AtExit() failed."); }
  return ret;
}

void System::Exit(int status) { exit(status); }

System::Process System::GetProcessID() { return getpid(); }

System::Process System::GetParentProcessID() { return getppid(); }

System::Process System::GetProcessGroupID() { return getpgrp(); }

int System::Sleep(int seconds) { return sleep(seconds); }

int System::Execute(const char *path, char *const argv[], char *const envp[]) {
  int ret;
  if (envp == NULL) {
    ret = execv(path, argv);
  } else {
    ret = execve(path, argv, envp);
  }
  if (ret == -1) { ERROR("Execute() failed with: %s", strerror(errno)); }
  return ret;
}

int System::ExecuteShellCommand(const char *command) {
  int ret = system(command);
  if (ret == -1) {
    ERROR("ExecuteShellCommand() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::Yield() {
  int ret = sched_yield();
  if (ret == -1) { ERROR("Yield() failed with: %s", strerror(errno)); }
  return ret;
}

// Signal

int System::SignalSetEmpty(SignalSet *set) {
  int ret = sigemptyset(set);
  if (ret == -1) {
    ERROR("EmptySignalSet() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SignalSetFill(SignalSet *set) {
  int ret = sigfillset(set);
  if (ret == -1) {
    ERROR("FillSignalSet() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SignalSetAdd(SignalSet *set, int signum) {
  int ret = sigaddset(set, signum);
  if (ret == -1) {
    ERROR("AddToSignalSet() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SignalSetRemove(SignalSet *set, int signum) {
  int ret = sigdelset(set, signum);
  if (ret == -1) {
    ERROR("DeleteFromSignalSet() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SignalSetContains(const SignalSet *set, int signum) {
  int ret = sigismember(set, signum);
  if (ret == -1) {
    ERROR("IsMemberOfSignalSet() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::ControlSignalAction(int signum,
                                const SignalAction *act,
                                SignalAction *oldact) {
  int ret = sigaction(signum, act, oldact);
  if (ret == -1) { ERROR("SignalAction() failed with: %s", strerror(errno)); }
  return ret;
}

// String

char *System::Basename(std::string path) {
  return basename(const_cast<char *>(path.c_str()));
}

int System::CompareStrings(const char *s1, const char *s2) {
  int x = strlen(s1);
  int y = strlen(s2);
  int n = (x < y ? x : y);
  return strncmp(s1, s2, n);
}

char *System::FindString(const char *haystack, const char *needle) {
  return const_cast<char *>(strstr(haystack, needle));
}

char *System::StringCopy(char *dest, const char *src, SizeType n) {
  return strncpy(dest, src, n);
}

System::SizeType System::StringLength(const char *string) {
  return strlen(string);
}

int System::StringToInt(const char *string) { return atoi(string); }

// I/O

int System::Select(int nfds, FDSet *readfds, FDSet *writefds,
                   FDSet *exceptfds, TimeValue *timeout) {
  int ret = select(nfds, readfds, writefds, exceptfds, timeout);
  if (ret == -1) { ERROR("Select() failed with: %s", strerror(errno)); }
  return ret;
}

void System::FDSetAdd(int fd, FDSet *set) { FD_SET(fd, set); }

int System::FDSetContains(int fd, FDSet *set) { return FD_ISSET(fd, set); }

void System::FDSetEmpty(FDSet *set) { FD_ZERO(set); }

void System::FDSetRemove(int fd, FDSet *set) { FD_CLR(fd, set); }

int System::Open(const char *path, int oflag, ...) {
  int ret;
  if ((oflag & O_CREAT) == O_CREAT) {
    va_list ap;
    va_start(ap, oflag);
    mode_t mode = va_arg(ap, mode_t);
    va_end(ap);
    ret = open(path, oflag, mode);
  } else {
    ret = open(path, oflag);
  }
  if (ret == -1) { ERROR("Open(%s) failed with: %s", path, strerror(errno)); }
  return ret;
}

int System::DuplicateFD(int old_fd, int new_fd) {
  int ret = dup2(old_fd, new_fd);
  if (ret == -1) { ERROR("DuplicateFD() failed with: %s", strerror(errno)); }
  return ret;
}

int System::FileControl(int fd, int cmd, ...) {
  int ret = -1;
  va_list ap;
  va_start(ap, cmd);
  switch (cmd) {
    case F_DUPFD:
    case F_DUPFD_CLOEXEC:
    case F_NOTIFY:
    case F_SETFD:
    case F_SETFL:
    case F_SETLEASE:
    case F_SETPIPE_SZ:
    case F_SETOWN:
    case F_SETSIG: {
      long arg = va_arg(ap, long);  // NOLINT
      ret = fcntl(fd, cmd, arg);
      break;
    }
    case F_GETFD:
    case F_GETFL:
    case F_GETLEASE:
    case F_GETPIPE_SZ:
    case F_GETOWN:
    case F_GETSIG: {
      ret = fcntl(fd, cmd);
      break;
    }
    case F_GETLK:
    case F_SETLK:
    case F_SETLKW: {
      struct flock *arg = va_arg(ap, struct flock *);
      ret = fcntl(fd, cmd, arg);
      break;
    }
    case F_GETOWN_EX:
    case F_SETOWN_EX: {
      struct f_owner_ex *arg = va_arg(ap, struct f_owner_ex *);
      ret = fcntl(fd, cmd, arg);
      break;
    }
  }
  va_end(ap);
  if (ret == -1) {
    ERROR("FileControl(%d, %d) failed with: %s", fd, cmd, strerror(errno));
  }
  return ret;
}

std::string System::FileControlCommandToString(int cmd) {
  switch (cmd) {
    case F_DUPFD:
      return "F_DUPFD";
    case F_DUPFD_CLOEXEC:
      return "F_DUPFD_CLOEXEC";
    case F_NOTIFY:
      return "F_NOTIFY";
    case F_SETFD:
      return "F_SETFD";
    case F_SETFL:
      return "F_SETFL";
    case F_SETLEASE:
      return "F_SETLEASE";
    case F_SETPIPE_SZ:
      return "F_SETPIPE_SZ";
    case F_SETOWN:
      return "F_SETOWN";
    case F_SETSIG:
      return "F_SETSIG";
    case F_GETFD:
      return "F_GETFD";
    case F_GETFL:
      return "F_GETFL";
    case F_GETLEASE:
      return "F_GETLEASE";
    case F_GETPIPE_SZ:
      return "F_GETPIPE_SZ";
    case F_GETOWN:
      return "F_GETOWN";
    case F_GETSIG:
      return "F_GETSIG";
    case F_GETLK:
      return "F_GETLK";
    case F_SETLK:
      return "F_SETLK";
    case F_SETLKW:
      return "F_SETLKW";
    case F_GETOWN_EX:
      return "F_GETOWN_EX";
    case F_SETOWN_EX:
      return "F_SETOWN_EX";
  }
  return "UNKNOWN";
}

bool System::FileExists(const char *path) {
  return (access(path, F_OK) != -1);
}

System::File *System::FileOpen(const char *path, const char *mode) {
  File *file = fopen(path, mode);
  if (!file) {
    ERROR("FileOpen(%s, %s) failed with: %s", path, mode, strerror(errno));
  }
  return file;
}

int System::FileClose(File *file) {
  int ret = fclose(file);
  if (ret == EOF) { ERROR("FileClose() failed with: %s", strerror(errno)); }
  return ret;
}

int System::FileFlush(File *file) {
  int ret = fflush(file);
  if (ret == EOF) { ERROR("FileFlush() failed with: %s", strerror(errno)); }
  return ret;
}

int System::FilePrint(File *file, const char *format, ...) {
  int ret;
  va_list ap;
  va_start(ap, format);
  ret = vfprintf(file, format, ap);
  va_end(ap);

  return ret;
}

int System::FileScan(File *file, const char *format, ...) {
  int ret;
  va_list ap;
  va_start(ap, format);
  ret = vfscanf(file, format, ap);
  va_end(ap);
  if (ret == EOF) { ERROR("FileScan() failed with: %s", strerror(errno)); }
  return ret;
}

void System::FileGetString(char *buffer, int size, File *file) {
  char *ret = fgets(buffer, size, file);
  if (ret == NULL) {
    ERROR("FileGetString() failed with: %s", strerror(errno));
  }
}

int System::FileEOF(File *file) {
  int ret = feof(file);
  if (ret == -1) { ERROR("FileEOF() failed with: %s", strerror(errno)); }
  return ret;
}

int System::ErrorOutput(const char *format, ...) {
  int ret;
  va_list ap;
  va_start(ap, format);
  ret = vfprintf(stderr, format, ap);
  va_end(ap);
  return ret;
}

int System::Output(const char *format, ...) {
  int ret;
  va_list ap;
  va_start(ap, format);
  ret = vfprintf(stdout, format, ap);
  va_end(ap);
  return ret;
}

int System::Input(const char *format, ...) {
  int ret;
  va_list ap;
  va_start(ap, format);
  ret = vfscanf(stdin, format, ap);
  va_end(ap);
  if (ret == EOF) {
    ERROR("Input(%s) failed with: %s", format, strerror(errno));
  }
  return ret;
}

int System::CreateDirectory(const char *path) {
  int ret = mkdir(path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  if (ret == -1) {
    ERROR("CreateDirectory(%s) failed with: %s", path, strerror(errno));
  }
  return ret;
}

int System::Rename(const char *old_path, const char *new_path) {
  int ret = rename(old_path, new_path);
  if (ret == -1) {
    ERROR("Rename(%s, %s) failed with: %s",
          old_path, new_path, strerror(errno));
  }
  return ret;
}

int System::Remove(const char *path) {
  FileStat stat;
  Stat(path, &stat);
  if (stat.st_mode & S_IFREG) { return RemoveFile(path); }
  if (stat.st_mode & S_IFDIR) { return RemoveDirectory(path, true); }
  ERROR("Cannot remove %s", path);
}

int System::RemoveFile(const char *path) {
  int ret = unlink(path);
  if (ret == -1) {
    ERROR("RemoveFile(%s) failed with: %s", path, strerror(errno));
  }
  return ret;
}

int System::RemoveDirectory(const char *path, bool recursive) {
  const int kMaxOpenFD = 10;
  if (recursive) {
    int ret = nftw(path, &System::RemoveHelper,
                   kMaxOpenFD, FTW_PHYS | FTW_DEPTH);
    if (ret == -1) { ERROR("Failed to remove directory %s", path); }
    return ret;
  } else {
    int ret = rmdir(path);
    if (ret == -1) {
      ERROR("RemoveDirectory(%s) failed with: %s", path, strerror(errno));
    }
    return ret;
  }
}

int System::RemoveHelper(const char *path, const FileStat * /* buffer */,
                         int type, struct FTW * /* ftw_buffer */) {
  switch (type) {
    case FTW_F:
      System::RemoveFile(path);
      break;
    case FTW_DP:
      System::RemoveDirectory(path, false);
      break;
    default:
      ERROR("Cannot remove %s", path);
  }
  return 0;
}

int System::Stat(const char *path, FileStat *buffer) {
  int ret = stat(path, buffer);
  if (ret == -1) { ERROR("Stat(%s) failed with: %s", path, strerror(errno)); }
  return ret;
}

int System::IOControl(int fd, int request, void *data) {
  int ret = ioctl(fd, request, data);
  if (ret == -1) { ERROR("IOControl() failed with: %s", strerror(errno)); }
  return ret;
}

int System::GetSocketName(int sockfd,
                          SocketAddress *addr,
                          socklen_t *addrlen) {
  int ret = getsockname(sockfd, addr, addrlen);
  if (ret == -1) {
    ERROR("GetSocketName() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::GetSocketOption(int sockfd, int level, int optname,
                            void *optval, socklen_t *optlen) {
  int ret = getsockopt(sockfd, level, optname, optval, optlen);
  if (ret == -1) {
    ERROR("GetSocketOption() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SetSocketOption(int sockfd, int level, int optname,
                            const void *optval, socklen_t optlen) {
  int ret = setsockopt(sockfd, level, optname, optval, optlen);
  if (ret == -1) {
    ERROR("SetSocketOption() failed with: %s", strerror(errno));
  }
  return ret;
}

System::UInt32 System::HostToNetworkLong(UInt32 host) { return htonl(host); }

System::UInt16 System::HostToNetworkShort(UInt16 host) { return htons(host); }

System::UInt32 System::NetworkToHostLong(UInt32 net) { return ntohl(net); }

System::UInt16 System::NetworkToHostShort(UInt16 net) { return ntohs(net); }

int System::IPv4StringToNetwork(const char *cp, IPv4Address *inp) {
  int ret = inet_aton(cp, inp);
  if (ret == 0) { ERROR("Invalid address %s.", cp); }
  return ret;
}

int System::IPv4GetNetwork(const char * /* cp */) {
  ERROR("Use IPv4StringToNetwork() instead of IPv4GetNetwork");
}

int System::IPv4GetHost(const char *cp) {
  int ret = inet_network(cp);
  if (ret == -1) { ERROR("Invalid address %s.", cp); }
  return ret;
}

std::string System::IPv4NetworkToString(IPv4Address in) {
  char buffer[INET_ADDRSTRLEN];
  return inet_ntop(AF_INET, &in, buffer, INET_ADDRSTRLEN);
}

std::string System::IPv6NetworkToString(IPv6Address in) {
  char buffer[INET6_ADDRSTRLEN];
  return inet_ntop(AF_INET6, &in, buffer, INET6_ADDRSTRLEN);
}

System::IPv4Address System::IPv4MakeAddress(int net, int host) {
  return inet_makeaddr(net, host);
}

int System::IPv4GetLocalNetworkOf(IPv4Address in) {
  return inet_lnaof(in);
}

int System::IPv4GetNetworkOf(IPv4Address in) {
  return inet_netof(in);
}

// Thread management

int System::ThreadCreate(Thread *thread, const ThreadAttr *attr,
                         void *(*function)(void *), void *arg) {  // NOLINT
  int ret = pthread_create(thread, attr, function, arg);
  if (ret != 0) { ERROR("ThreadCreate() failed with: %s", strerror(ret)); }
  return ret;
}

int System::ThreadCancel(Thread thread) {
  int ret = pthread_cancel(thread);
  if (ret != 0) { ERROR("ThreadCancel() failed with: %s", strerror(ret)); }
  return ret;
}

int System::ThreadDetach(Thread thread) {
  int ret = pthread_detach(thread);
  if (ret != 0) { ERROR("ThreadDetach() failed with: %s", strerror(ret)); }
  return ret;
}

int System::ThreadJoin(Thread thread, void **retval) {
  int ret = pthread_join(thread, retval);
  if (ret != 0) { ERROR("ThreadJoin() failed with: %s", strerror(ret)); }
  return ret;
}

void System::ThreadExit(void *retval) { pthread_exit(retval); }

System::Thread System::ThreadSelf() { return pthread_self(); }

int System::ThreadSetCancelState(int state, int *oldstate) {
  int ret = pthread_setcancelstate(state, oldstate);
  if (ret != 0) {
    ERROR("ThreadSetCancelState() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadSetCancelType(int type, int *oldtype) {
  int ret = pthread_setcanceltype(type, oldtype);
  if (ret != 0) {
    ERROR("ThreadSetCancelType() failed with: %s", strerror(ret));
  }
  return ret;
}

// Thread attribute

int System::ThreadAttrInit(ThreadAttr *attr) {
  int ret = pthread_attr_init(attr);
  if (ret != 0) { ERROR("ThreadAttrInit() failed with: %s", strerror(ret)); }
  return ret;
}

int System::ThreadAttrDestroy(ThreadAttr *attr) {
  int ret = pthread_attr_destroy(attr);
  if (ret != 0) { ERROR("ThreadAttrDestroy() failed with: %s", strerror(ret)); }
  return ret;
}

int System::ThreadAttrSetInheritSched(ThreadAttr *attr, int inheritsched) {
  int ret = pthread_attr_setinheritsched(attr, inheritsched);
  if (ret != 0) {
    ERROR("ThreadAttrSetInheritSched() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetInheritSched(ThreadAttr *attr, int *inheritsched) {
  int ret = pthread_attr_getinheritsched(attr, inheritsched);
  if (ret != 0) {
    ERROR("ThreadAttrGetInheritSched() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrSetSchedParam(ThreadAttr *attr,
                                    const SchedParam *param) {
  int ret = pthread_attr_setschedparam(attr, param);
  if (ret != 0) {
    ERROR("ThreadAttrSetSchedParam() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetSchedParam(ThreadAttr *attr, SchedParam *param) {
  int ret = pthread_attr_getschedparam(attr, param);
  if (ret != 0) {
    ERROR("ThreadAttrGetSchedParam() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrSetSchedPolicy(ThreadAttr *attr, int policy) {
  int ret = pthread_attr_setschedpolicy(attr, policy);
  if (ret != 0) {
    ERROR("ThreadAttrSetSchedPolicy() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetSchedPolicy(ThreadAttr *attr, int *policy) {
  int ret = pthread_attr_getschedpolicy(attr, policy);
  if (ret != 0) {
    ERROR("ThreadAttrGetSchedPolicy() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrSetScope(ThreadAttr *attr, int scope) {
  int ret = pthread_attr_setscope(attr, scope);
  if (ret != 0) {
    ERROR("ThreadAttrSetScope() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetScope(ThreadAttr *attr, int *scope) {
  int ret = pthread_attr_getscope(attr, scope);
  if (ret != 0) {
    ERROR("ThreadAttrGetScope() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrSetStackSize(ThreadAttr *attr, SizeType stacksize) {
  int ret = pthread_attr_setstacksize(attr, stacksize);
  if (ret != 0) {
    ERROR("ThreadAttrSetStackSize() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetStackSize(ThreadAttr *attr, SizeType *stacksize) {
  int ret = pthread_attr_getstacksize(attr, stacksize);
  if (ret != 0) {
    ERROR("ThreadAttrGetStackSize() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrSetStackAddr(ThreadAttr * /* attr */,
                                   void * /* stackaddr */) {
  ERROR("ThreadAttrSetStackAddr() is deprecated.");
}

int System::ThreadAttrGetStackAddr(ThreadAttr * /* attr */,
                                   void ** /* stackaddr */) {
  ERROR("ThreadAttrGetStackAddr() is deprecated.");
}

int System::ThreadAttrSetStack(ThreadAttr *attr,
                               void *stackaddr,
                               SizeType stacksize) {
  int ret = pthread_attr_setstack(attr, stackaddr, stacksize);
  if (ret != 0) {
    ERROR("ThreadAttrSetStack() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetStack(ThreadAttr *attr,
                               void **stackaddr,
                               SizeType *stacksize) {
  int ret = pthread_attr_getstack(attr, stackaddr, stacksize);
  if (ret != 0) {
    ERROR("ThreadAttrGetStack() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrSetGuardSize(ThreadAttr *attr, SizeType guardsize) {
  int ret = pthread_attr_setguardsize(attr, guardsize);
  if (ret != 0) {
    ERROR("ThreadAttrSetGuardSize() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetGuardSize(ThreadAttr *attr, SizeType *guardsize) {
  int ret = pthread_attr_getguardsize(attr, guardsize);
  if (ret != 0) {
    ERROR("ThreadAttrGetGuardSize() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrSetDetachState(ThreadAttr *attr, int detachstate) {
  int ret = pthread_attr_setdetachstate(attr, detachstate);
  if (ret != 0) {
    ERROR("ThreadAttrSetDetachState() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::ThreadAttrGetSetachState(ThreadAttr *attr, int *detachstate) {
  int ret = pthread_attr_getdetachstate(attr, detachstate);
  if (ret != 0) {
    ERROR("ThreadAttrGetDetachState() failed with: %s", strerror(ret));
  }
  return ret;
}

// Barrier

int System::BarrierDestroy(Barrier *barrier) {
  int ret = pthread_barrier_destroy(barrier);
  if (ret != 0) { ERROR("BarrierDestroy() failed with: %s", strerror(ret)); }
  return ret;
}
int System::BarrierInit(Barrier *DBUG_RESTRICT barrier,
                        const BarrierAttr *DBUG_RESTRICT attr,
                        unsigned count) {
  int ret = pthread_barrier_init(barrier, attr, count);
  if (ret != 0) { ERROR("BarrierInit() failed with: %s", strerror(ret)); }
  return ret;
}

int System::BarrierWait(Barrier *barrier) {
  int ret = pthread_barrier_wait(barrier);
  if (ret != 0 && ret != PTHREAD_BARRIER_SERIAL_THREAD) {
    ERROR("BarrierWait() failed with: %s", strerror(ret));
  }
  return ret;
}

// Condition variable

int System::CVarDestroy(CVar *cond) {
  int ret = pthread_cond_destroy(cond);
  if (ret != 0) { ERROR("CVarDestroy() failed with: %s", strerror(ret)); }
  return ret;
}

int System::CVarInit(CVar *DBUG_RESTRICT cond,
                     const CVarAttr *DBUG_RESTRICT attr) {
  int ret = pthread_cond_init(cond, attr);
  if (ret != 0) { ERROR("CVarInit() failed with: %s", strerror(ret)); }
  return ret;
}

int System::CVarTimedWait(CVar *DBUG_RESTRICT cond,
                          Mutex *DBUG_RESTRICT mutex,
                          const TimeSpec *DBUG_RESTRICT abstime) {
  int ret = pthread_cond_timedwait(cond, mutex, abstime);
  if (ret != 0 && ret != ETIMEDOUT) {
    ERROR("CVarTimedWait() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::CVarWait(CVar *DBUG_RESTRICT cond,
                     Mutex *DBUG_RESTRICT mutex) {
  int ret = pthread_cond_wait(cond, mutex);
  if (ret != 0) { ERROR("CVarWait() failed with: %s", strerror(ret)); }
  return ret;
}

int System::CVarBroadcast(CVar *cond) {
  int ret = pthread_cond_broadcast(cond);
  if (ret != 0) { ERROR("CVarBroadcast() failed with: %s", strerror(ret)); }
  return ret;
}

int System::CVarSignal(CVar *cond) {
  int ret = pthread_cond_signal(cond);
  if (ret != 0) { ERROR("CVarSignal() failed with: %s", strerror(ret)); }
  return ret;
}

// Thread-specific data

int System::KeyCreate(ThreadKey *key, void (*destructor)(void*)) {  // NOLINT
  int ret = pthread_key_create(key, destructor);
  if (ret != 0) { ERROR("KeyCreate() failed with: %s", strerror(ret)); }
  return ret;
}

int System::KeyDelete(ThreadKey key) {
  int ret = pthread_key_delete(key);
  if (ret != 0) { ERROR("KeyDelete() failed with: %s", strerror(ret)); }
  return ret;
}

void *System::KeyGetSpecific(ThreadKey key) { return pthread_getspecific(key); }

int System::KeySetSpecific(ThreadKey key, const void *value) {
  int ret = pthread_setspecific(key, value);
  if (ret != 0) { ERROR("KeySetSpecific() failed with: %s", strerror(ret)); }
  return ret;
}

// Mutex

int System::MutexDestroy(Mutex *mutex) {
  int ret = pthread_mutex_destroy(mutex);
  if (ret != 0) { ERROR("MutexDestroy() failed with: %s", strerror(ret)); }
  return ret;
}

int System::MutexInit(Mutex *DBUG_RESTRICT mutex,
                      const MutexAttr *DBUG_RESTRICT attr) {
  int ret = pthread_mutex_init(mutex, attr);
  if (ret != 0) { ERROR("MutexInit() failed with: %s", strerror(ret)); }
  return ret;
}

int System::MutexLock(Mutex *mutex) {
  int ret = pthread_mutex_lock(mutex);
  if (ret != 0) { ERROR("MutexLock() failed with: %s", strerror(ret)); }
  return ret;
}

int System::MutexTryLock(Mutex *mutex) {
  int ret = pthread_mutex_trylock(mutex);
  if (ret != 0 && ret != EBUSY) {
    ERROR("MutexTryLock() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::MutexUnlock(Mutex *mutex) {
  int ret = pthread_mutex_unlock(mutex);
  if (ret != 0) { ERROR("MutexUnlock() failed with: %s", strerror(ret)); }
  return ret;
}

// Read-write lock

int System::RWLockDestroy(RWLock *rwlock) {
  int ret = pthread_rwlock_destroy(rwlock);
  if (ret != 0) { ERROR("RWLockDestroy() failed with: %s", strerror(ret)); }
  return ret;
}

int System::RWLockInit(RWLock *DBUG_RESTRICT rwlock,
                       const RWLockAttr *DBUG_RESTRICT attr) {
  int ret = pthread_rwlock_init(rwlock, attr);
  if (ret != 0) { ERROR("RWLockInit() failed with: %s", strerror(ret)); }
  return ret;
}

int System::RWLockReadLock(RWLock *rwlock) {
  int ret = pthread_rwlock_rdlock(rwlock);
  if (ret != 0) {
    ERROR("RWLockReadLock() failed with: %s (error %d)", strerror(ret), ret); }
  return ret;
}

int System::RWLockTryReadLock(RWLock *rwlock) {
  int ret = pthread_rwlock_tryrdlock(rwlock);
  if (ret != 0 && ret != EBUSY) {
    ERROR("RWLockTryReadLock() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::RWLockWriteLock(RWLock *rwlock) {
  int ret = pthread_rwlock_wrlock(rwlock);
  if (ret != 0) { ERROR("RWLockWriteLock() failed with: %s", strerror(ret)); }
  return ret;
}

int System::RWLockTryWriteLock(RWLock *rwlock) {
  int ret = pthread_rwlock_trywrlock(rwlock);
  if (ret != 0 && ret != EBUSY) {
    ERROR("RWLockTryWriteLock() failed with: %s", strerror(ret));
  }
  return ret;
}

int System::RWLockUnlock(RWLock *rwlock) {
  int ret = pthread_rwlock_unlock(rwlock);
  if (ret != 0) { ERROR("RWLockUnlock() failed with: %s", strerror(ret)); }
  return ret;
}

// Semaphore

int System::SemaphoreInit(Semaphore *sem, int pshared, unsigned int value) {
  int ret = sem_init(sem, pshared, value);
  if (ret == -1) { ERROR("SemaphoreInit() failed with: %s", strerror(errno)); }
  return ret;
}

int System::SemaphoreDestroy(Semaphore *sem) {
  int ret = sem_destroy(sem);
  if (ret == -1) {
    ERROR("SemaphoreDestroy() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SemaphoreWait(Semaphore *sem) {
  int ret = sem_wait(sem);
  if (ret == -1) { ERROR("SemaphoreWait() failed with: %s", strerror(errno)); }
  return ret;
}

int System::SemaphoreTryWait(Semaphore *sem) {
  int ret = sem_trywait(sem);
  if (ret == -1 && errno != EAGAIN) {
    ERROR("SemaphoreTryWait() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SemaphoreTimedWait(Semaphore *sem, const TimeSpec *abs_timeout) {
  int ret = sem_timedwait(sem, abs_timeout);
  if (ret == -1 && errno != ETIMEDOUT) {
    ERROR("SemaphoreTimedWait() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::SemaphorePost(Semaphore *sem) {
  int ret = sem_post(sem);
  if (ret == -1) { ERROR("SemaphorePost() failed with: %s", strerror(errno)); }
  return ret;
}

int System::SemaphoreGetValue(Semaphore *sem, int *sval) {
  int ret = sem_getvalue(sem, sval);
  if (ret == -1) {
    ERROR("SemaphoreGetValue() failed with: %s", strerror(errno));
  }
  return ret;
}

System::Semaphore *System::SemaphoreOpen(const char *path, int oflag, ...) {
  System::Semaphore *ret;
  if ((oflag & O_CREAT) == O_CREAT) {
    va_list ap;
    va_start(ap, oflag);
    mode_t mode = va_arg(ap, mode_t);
    int value = va_arg(ap, unsigned);
    va_end(ap);
    Output("%d\n", value);
    ret = sem_open(path, oflag, mode, value);
  } else {
    Output("0\n");
    ret = sem_open(path, oflag);
  }
  if (ret == SEM_FAILED) {
    ERROR("SemaphoreOpen(%s) failed with: %s", path, strerror(errno));
  }
  return ret;
}

int System::SemaphoreClose(Semaphore *sem) {
  int ret = sem_close(sem);
  if (ret == -1) { ERROR("SemaphoreClose() failed with: %s", strerror(errno)); }
  return ret;
}

int System::SemaphoreUnlink(const char *path) {
  int ret = sem_unlink(path);
  if (ret == -1) {
    ERROR("SemaphoreUnlink(%s) failed with: %s", path, strerror(errno));
  }
  return ret;
}

// Dynamic memory

void *System::Malloc(SizeType num) {
  if (num == 0) { return NULL; }
  void *p = malloc(num);
  if (!p) { ERROR("Dynamic memory exhausted."); }
  return p;
}

void *System::Realloc(void *p, SizeType num) {
  if (!p) { return Malloc(num); }
  void *new_p = realloc(p, num);
  if (!new_p) { ERROR("Dynamic memory exhausted."); }
  return new_p;
}

void *System::Calloc(SizeType num, SizeType size) {
  void *p = Malloc(num * size);
  bzero(p, num * size);
  return p;
}

void System::Free(void *p) { free(p); }

// Time

System::ClockType System::Clock() {
  ClockType ret = clock();
  if (ret == -1) { ERROR("Clock() failed with: %s", strerror(errno)); }
  return ret;
}

int System::GetTimeOfDay(TimeValue *time_value, TimeZone *time_zone) {
  int ret = gettimeofday(time_value, time_zone);
  if (ret == -1) { ERROR("GetTimeOfDay() failed with: %s", strerror(errno)); }
  return ret;
}

System::TimeType System::Time(TimeType *tloc) {
  TimeType ret = time(tloc);
  if (ret == -1) { ERROR("Time() failed with: %s", strerror(errno)); }
  return ret;
}

// Randomness

char *System::InitState(unsigned int seed, char *state, SizeType statelen) {
  char *ret = initstate(seed, state, statelen);
  if (ret == NULL) { ERROR("InitState() failed with: %s", strerror(errno)); }
  return ret;
}

int System::Random() { return random(); }

int System::Random(int n) { return random() % n; }

void System::SeedRandom(unsigned int seed) { srandom(seed); }

char *System::SetState(char *state) {
  char *ret = setstate(state);
  if (ret == NULL) { ERROR("SetState() failed with: %s", strerror(errno)); }
  return ret;
}

int System::ReentrantInitState(unsigned int seed,
                               char *state,
                               SizeType statelen,
                               RandomData *buffer) {
  int ret = initstate_r(seed, state, statelen, buffer);
  if (ret == -1) {
    ERROR("ReentrantInitState() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::ReentrantRandom(RandomData *buffer, int *result) {
  int ret = random_r(buffer, result);
  if (ret == -1) {
    ERROR("ReentrantRandom() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::ReentrantRandom(RandomData *buffer, int *result, int n) {
  int ret = random_r(buffer, result);
  if (ret == -1) {
    ERROR("ReentrantRandom() failed with: %s", strerror(errno));
  }
  *result = *result % n;
  return ret;
}

int System::ReentrantSeedRandom(unsigned int seed, RandomData *buffer) {
  int ret = srandom_r(seed, buffer);
  if (ret == -1) {
    ERROR("ReentrantSeedRandom() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::ReentrantSetState(char *state, RandomData *buffer) {
  int ret = setstate_r(state, buffer);
  if (ret == -1) {
    ERROR("ReentrantSetState() failed with: %s", strerror(errno));
  }
  return ret;
}

// Miscellaneous

int System::Backtrace(void **buffer, int size) {
  int ret = backtrace(buffer, size);
  if (ret == 0) { ERROR("Backtrace() failed."); }
  return ret;
}

char **System::BacktraceSymbols(void **buffer, int size) {
  char **ret = backtrace_symbols(buffer, size);
  if (ret == NULL) { ERROR("BacktraceSymbols() failed."); }
  return ret;
}

void System::CopyBytes(const void *src, void *dest, SizeType n) {
  return bcopy(src, dest, n);
}

int System::DLAddress(void *addr, DLInfo *info) {
  int ret = dladdr(addr, info);
  if (ret == 0) { ERROR("DLAddr() failed."); }
  return ret;
}

char **System::GetEnvironment() { return environ; }

void System::SetEnvironment(char **environment) { environ = environment; }

char *System::GetEnvironmentVariable(const char *name) {
  return getenv(name);
}

int System::SetEnvironmentVariable(const char *name,
                                   const char *value,
                                   int overwrite) {
  int ret = setenv(name, value, overwrite);
  if (ret == -1) {
    ERROR("SetEnvironmentVariable() failed with: %s", strerror(errno));
  }
  return ret;
}

int System::UnsetEnvironmentVariable(const char *name) {
  int ret = unsetenv(name);
  if (ret == -1) {
    ERROR("UnsetEnvironmentVariable() failed with: %s", strerror(errno));
  }
  return ret;
}

}  // namespace dbug
