/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/common/util.h"

#include <bfd.h>
// The following efinition avoids redefinition of basename() in demangle.h.
#define HAVE_DECL_BASENAME 1
#include <demangle.h>

#include <map>
#include <sstream>
#include <string>

#include "dbug/common/command.h"
#include "dbug/common/configuration.h"
#include "dbug/common/lock.h"
#include "dbug/common/system.h"

#define DEBUG(args...) DEBUG_ON_CHANNEL(Logging::UTIL, ##args)
#define ERROR(args...) ERROR_ON_CHANNEL(Logging::UTIL, ##args)
#define MESSAGE(args...) MESSAGE_ON_CHANNEL(Logging::UTIL, ##args)
#define WARNING(args...) WARNING_ON_CHANNEL(Logging::UTIL, ##args)

#define LOG(verbosity, channel, args...) {      \
    switch (verbosity) {                        \
      case Logging::DEBUG:                      \
        DEBUG_ON_CHANNEL(channel, ##args);      \
        break;                                  \
      case Logging::ERROR:                      \
        ERROR_ON_CHANNEL(channel, ##args);      \
        break;                                  \
      case Logging::MESSAGE:                    \
        MESSAGE_ON_CHANNEL(channel, ##args);    \
        break;                                  \
      case Logging::WARNING:                    \
        WARNING_ON_CHANNEL(channel, ##args);    \
        break;                                  \
    }                                           \
  }

using dbug::thrift::Location;

namespace dbug {

// public methods

void Util::Initialize() {
  Logging::add_channel(Logging::UTIL, NULL, Configuration::verbosity());
  assert(!context_map_);
  context_map_ = new std::map<std::string, Util::Context *>();
  System::MutexInit(&context_map_mutex_, NULL);
  next_id_ = 0;
  System::MutexInit(&id_mutex_, NULL);
}

void Util::Finalize() {
  System::MutexDestroy(&id_mutex_);
  System::MutexDestroy(&context_map_mutex_);
  for (std::pair<std::string, Context *> pair : (*context_map_)) {
    Context *context = pair.second;
    if (context == NULL) { continue; }
    DeleteContext(context);
  }
  delete context_map_;
}

std::string Util::CreateResourceID(std::string type,
                                   int pid,
                                   System::UInt64 rid) {
  std::stringstream resource_id;
  resource_id << type << "-" << pid << "-" << rid;
  return resource_id.str();
}

std::vector<Location> Util::GetBacktrace() {
  std::vector<Location> result;
  if (Configuration::max_backtrace_length() == 0) { return result; }
  void **callstack = reinterpret_cast<void **>(
      System::Calloc(Configuration::max_backtrace_length(), sizeof(callstack)));
  int n = System::Backtrace(callstack, Configuration::max_backtrace_length());
  char **strings = System::BacktraceSymbols(callstack, n);
  for (int i = 0; i < n ; i++) {
    DEBUG("%d %s", i+1, strings[i]);
    System::DLInfo info;
    System::DLAddress(callstack[i], &info);
    DEBUG("Filename: %s [%p]", info.dli_fname, info.dli_fbase);
    DEBUG("Symbol: %s [%p]", info.dli_sname, info.dli_saddr);
    // Do not add dbug code to the backtrace.
    if (System::FindString(info.dli_fname, "libdbug.so")) { continue; }
    // Do not add valgrind code to the backtrace.
    if (System::FindString(info.dli_fname, "valgrind")) { continue; }

    System::MutexLock(&context_map_mutex_);
    Context *context = NULL;
    std::string filename(info.dli_fname);
    // Check if the context for this file has been already created.
    if (context_map_->find(filename) != context_map_->end()) {
      context = context_map_->at(filename);
    } else {
      // If not, create it and add it to the context map.
      context = CreateContext(info.dli_fname);
      context_map_->insert(std::pair<std::string, Context *>
                           (filename, context));
    }
    FindAddressArguments args;
    args.address = reinterpret_cast<bfd_vma>(callstack[i]);
    args.file_name = NULL;
    args.found = false;
    args.function_name = NULL;
    args.line = 0;
    char *demangled = NULL;
    if (context != NULL) {
      args.symbol_table = context->symbol_table;
      bfd_map_over_sections(context->descriptor,
                            Util::FindAddressInSection,
                            &args);
      if (args.found) {
        demangled = cplus_demangle(args.function_name,
                                   DMGL_ANSI | DMGL_PARAMS);
        if (demangled) { demangled[strchr(demangled, '(') - demangled] = 0; }
      }
    }
    System::MutexUnlock(&context_map_mutex_);
    Location location;
    location.file_name = (args.file_name ? args.file_name :
                          (info.dli_fname ? info.dli_fname : "??"));
    location.func_name = (demangled ? demangled :
                          (args.function_name ? args.function_name :
                           (info.dli_sname ? info.dli_sname : "??")));
    location.line = static_cast<int>(args.line);
    if (demangled) { System::Free(demangled); }
    result.push_back(location);
    DEBUG("%s at %s:%d",
          location.file_name.c_str(),
          location.func_name.c_str(),
          location.line);
  }
  System::Free(strings);
  System::Free(callstack);
  return result;
}

std::string Util::GetMACAddress(int socket) {
  std::string result = "";
  System::IFRequest interfaces[Configuration::max_interfaces()];
  System::IFConfig config;
  config.ifc_len = sizeof(interfaces);
  config.ifc_req = interfaces;
  System::IOControl(socket, SIOCGIFCONF, &config);
  for (int i = 0; i < Configuration::max_interfaces(); ++i) {
    System::IFRequest request = config.ifc_req[i];
    if (request.ifr_addr.sa_family == AF_INET &&
        System::FindString(request.ifr_name, "eth") != NULL ) {
      System::IOControl(socket, SIOCGIFHWADDR, &request);
      std::stringstream stream;
      unsigned char *address =
          reinterpret_cast<unsigned char *>(request.ifr_hwaddr.sa_data);
      for (int j = 0; j < 6; ++j) {
        if (j != 0) stream << ':';
        stream.width(2);
        stream.fill('0');
        stream << std::hex << static_cast<int>(address[j]);
      }
      result = stream.str();
      break;
    }
  }
  return result;
}

int Util::NewIdentifier() {
  MutexLock lock(&id_mutex_);
  return ++next_id_;
}

System::File *Util::OpenLogFile(const char *name) {
  char *prefix = System::GetEnvironmentVariable("DBUG_LOG_DIR");
  std::string path;
  if (prefix) {
    path.append(prefix).append("/").append(name);
  } else {
    path.append(name);
  }
  return System::FileOpen(path.c_str(), "a");
}

void Util::ParseArgumentList(char ***argv,
                             char ***envp,
                             char *arg,
                             va_list arglist) {
  // va_start has already been called by the caller
  char *argp;
  const int kInitialSize = 32;
  int argv_size = kInitialSize;
  int argc = 1;
  *argv = reinterpret_cast<char **>
      (System::Malloc((argv_size + 1) * sizeof(**argv)));
  (*argv)[0] = arg;
  while ((argp = va_arg(arglist, char *)) != NULL) {
    if (argc > argv_size) {
      argv_size *= 2;
      *argv = reinterpret_cast<char **>
          (System::Realloc(*argv, (argv_size + 1) * sizeof(**argv)));
    }
    (*argv)[argc] = argp;
    argc++;
  }
  (*argv)[argc] = NULL;
  if (envp != NULL) {
    *envp = reinterpret_cast<char **>(va_arg(arglist, char *));
  }
  // the caller calls va_end
}

void Util::PrintBacktrace(Logging::Verbosity verbosity,
                          Logging::Channel channel) {
  std::vector<Location> backtrace = GetBacktrace();
  for (unsigned i = 0; i < backtrace.size(); i++) {
    LOG(verbosity, channel, "%s at %s():%d",
        backtrace[i].func_name.c_str(),
        backtrace[i].file_name.c_str(),
        backtrace[i].line);
  }
}

void Util::SetupEnvironment(char ***envp, int pid, int tid) {
  std::stringstream stream;
  int count = 0;
  while ((*envp)[count] != NULL) {
    printf("%d: %s\n", count, (*envp)[count]);
    count++;
  }
  char *hostname = System::GetEnvironmentVariable("DBUG_HOSTNAME");
  if (hostname) {
    *envp = reinterpret_cast<char **>
        (System::Realloc(*envp, (count + 1) * sizeof(char *)));
    stream << "DBUG_HOSTNAME=" << hostname;
    (*envp)[count] = strdup(stream.str().c_str());
    stream.str("");
    count++;
  }
  char *port = System::GetEnvironmentVariable("DBUG_PORT");
  if (port) {
    *envp = reinterpret_cast<char **>
        (System::Realloc(*envp, (count + 1) * sizeof(char *)));
    stream << "DBUG_PORT=" << port;
    (*envp)[count] = strdup(stream.str().c_str());
    count++;
    stream.str("");
  }
  *envp = reinterpret_cast<char **>
      (System::Realloc(*envp, (count + 1) * sizeof(char *)));
  stream << "DBUG_PID=" << pid;
  (*envp)[count] = strdup(stream.str().c_str());
  count++;
  stream.str("");
  *envp = reinterpret_cast<char **>
      (System::Realloc(*envp, (count + 1) * sizeof(char *)));
  stream << "DBUG_TID=" << tid;
  (*envp)[count] = strdup(stream.str().c_str());
  count++;
  for (int i = 0; i < count; i++) { printf("%d: %s\n", i, (*envp)[i]); }
  *envp = reinterpret_cast<char **>
      (System::Realloc(*envp, (count + 1) * sizeof(char *)));
  (*envp)[count] = NULL;
}

std::string Util::SignalToString(int signal) {
  switch (signal) {
    case SIGHUP:
      return "SIGHUP";
    case SIGINT:
      return "SIGINT";
    case SIGQUIT:
      return "SIGQUIT";
    case SIGILL:
      return "SIGILL";
    case SIGTRAP:
      return "SIGTRAP";
    case SIGABRT:
      return "SIGABRT";
    case SIGBUS:
      return "SIGBUS";
    case SIGFPE:
      return "SIGFPE";
    case SIGKILL:
      return "SIGKILL";
    case SIGUSR1:
      return "SIGUSR1";
    case SIGSEGV:
      return "SIGSEGV";
    case SIGUSR2:
      return "SIGUSR2";
    case SIGPIPE:
      return "SIGPIPE";
    case SIGALRM:
      return "SIGALRM";
    case SIGTERM:
      return "SIGTERM";
    case SIGSTKFLT:
      return "SIGSTKFLT";
    case SIGCHLD:
      return "SIGCHLD";
    case SIGCONT:
      return "SIGCONT";
    case SIGSTOP:
      return "SIGSTOP";
    case SIGTSTP:
      return "SIGTSTP";
    case SIGTTIN:
      return "SIGTTIN";
    case SIGTTOU:
      return "SIGTTOU";
    case SIGURG:
      return "SIGURG";
    case SIGXCPU:
      return "SIGXCPU";
    case SIGXFSZ:
      return "SIGXFSZ";
    case SIGVTALRM:
      return "SIGVTALRM";
    case SIGPROF:
      return "SIGPROF";
    case SIGWINCH:
      return "SIGWINCH";
    case SIGIO:
      return "SIGIO";
    case SIGPWR:
      return "SIGPWR";
    case SIGSYS:
      return "SIGSYS";
    default:
      return "UNKNOWN SIGNAL";
  }
}

void Util::StartCommand(const Command *command,
                        std::map<std::string, std::string> environment) {
  std::string program = command->program();
  std::string input = command->input();
  std::string output = command->output();
  std::string error_output = command->error_output();
  std::vector<std::string> arguments = command->args();
  if (System::FileExists(program.c_str())) {
    char ** argv = reinterpret_cast<char **>(
        System::Calloc(arguments.size() + 2, sizeof(argv)));
    // Copy the name of the program.
    argv[0] = reinterpret_cast<char *>(System::Malloc(program.size() + 1));
    System::CopyBytes(program.c_str(), argv[0], program.size());
    argv[0][program.size()] = 0;
    // Copy the program arguments.
    for (unsigned i = 0; i < arguments.size(); i++) {
      int length = arguments[i].size();
      argv[i + 1] = reinterpret_cast<char *>(System::Malloc(length + 1));
      System::CopyBytes(arguments[i].c_str(), argv[i + 1], length);
      argv[i + 1][length] = 0;
    }
    // Terminate argv with NULL.
    argv[arguments.size() + 1] = NULL;

    for (std::pair<std::string, std::string> p : environment) {
      System::SetEnvironmentVariable(p.first.c_str(), p.second.c_str(), 1);
    }

    if (input != "") {
      // Redirect standard input.
      if (System::FileExists(input.c_str())) {
        int fd = System::Open(input.c_str(), O_RDONLY);
        System::DuplicateFD(fd, STDIN_FILENO);
      } else {
        ERROR("File %s for redirecting input does not exist.", input.c_str());
      }
    }

    if (output != "") {
      // Redirect standard output.
      int fd = System::Open(output.c_str(), O_CREAT | O_WRONLY, 0666);
      System::DuplicateFD(fd, STDOUT_FILENO);
    }

    if (error_output != "") {
      // Redirect error output.
      int fd = System::Open(error_output.c_str(), O_CREAT | O_WRONLY, 0666);
      System::DuplicateFD(fd, STDERR_FILENO);
    }

    System::Execute(program.c_str(), argv, NULL);
    ERROR("Failed to start binary %s: %s", program.c_str(), strerror(errno));
  } else {
    ERROR("Binary %s does not exist.", program.c_str());
  }
}

// private methods

void Util::FindAddressInSection(bfd *descriptor,
                                asection *section,
                                void *data) {
  FindAddressArguments *args = reinterpret_cast<FindAddressArguments *>(data);

  if (args->found) { return; }

  if ((bfd_get_section_flags(descriptor, section) & SEC_ALLOC) == 0) { return; }

  bfd_vma vma = bfd_get_section_vma(descriptor, section);
  if (args->address < vma) { return; }

  bfd_size_type size = bfd_get_section_size(section);
  if (args->address >= vma + size) { return; }

  args->found = bfd_find_nearest_line(descriptor,
                                      section,
                                      args->symbol_table,
                                      args->address - vma,
                                      &args->file_name,
                                      &args->function_name,
                                      &args->line);
}

asymbol **Util::GetSymbolTable(bfd *descriptor) {
  asymbol **symbol_table = NULL;
  System::Int64 storage;
  System::Int64 symcount;

  if ((bfd_get_file_flags(descriptor) & HAS_SYMS) == 0) { return NULL; }

  if ((storage = bfd_get_symtab_upper_bound(descriptor)) > 0) {
    symbol_table = reinterpret_cast<asymbol **>(System::Malloc(storage));
    symcount = bfd_canonicalize_symtab(descriptor, symbol_table);
    if (symcount == 0) {
      System::Free(symbol_table);
      symbol_table = NULL;
    }
  }

  if (symbol_table == NULL &&
      (storage = bfd_get_dynamic_symtab_upper_bound(descriptor)) > 0) {
    symbol_table = reinterpret_cast<asymbol **>(System::Malloc(storage));
    symcount = bfd_canonicalize_dynamic_symtab(descriptor, symbol_table);
    if (symcount == 0) {
      System::Free(symbol_table);
      symbol_table = NULL;
    }
  }

  return symbol_table;
}

Util::Context *Util::CreateContext(const char *name) {
  Context *context = new Context();
  context->descriptor = bfd_openr(name, NULL);
  if (context->descriptor == NULL) {
    WARNING("%s: File to create binary file descriptor.", name);
    DeleteContext(context);
    return NULL;
  }
  if (!bfd_check_format(context->descriptor, bfd_object)) {
    WARNING("%s: Format check failed.",
            bfd_get_filename(context->descriptor));
    DeleteContext(context);
    return NULL;
  }
  context->symbol_table = GetSymbolTable(context->descriptor);
  if (context->symbol_table == NULL) {
    DeleteContext(context);
    return NULL;
  }
  return context;
}

void Util::DeleteContext(Context *context) {
  if (context->symbol_table) {
    System::Free(context->symbol_table);
  }
  if (context->descriptor) {
    if (!bfd_close(context->descriptor)) {
      WARNING("%s: Failed to delete binary file descriptor.",
              bfd_get_filename(context->descriptor));
    }
  }
  delete context;
}

// private static fields

std::map<std::string, Util::Context *> *Util::context_map_;
System::Mutex Util::context_map_mutex_;
int Util::next_id_;
System::Mutex Util::id_mutex_;

}  // namespace dbug
