/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/explorer/frontier.h"

#include <algorithm>
#include <queue>
#include <vector>

#include "dbug/common/system.h"
#include "dbug/explorer/edge.h"
#include "dbug/explorer/node.h"
#include "dbug/explorer/logging.h"
#include "src/thrift/explorer/explorer_types.h"

using dbug::thrift::Move;

namespace dbug {
namespace explorer {

// Fragment::IncreasingOrder class

bool Fragment::IncreasingOrder::operator() (const Fragment *f1,
                                            const Fragment *f2) const {
  ASSERT(f1, "Fragment does not exist.");
  ASSERT(f2, "Fragment does not exist.");
  ASSERT(!f1->empty(), "Fragment is empty.");
  ASSERT(!f2->empty(), "Fragment is empty.");
  return f1->front()->depth() > f2->front()->depth();
}

// Frontier class

Frontier::Frontier(const Node *root)
    : max_fragments_(1),
      root_(root),
      strategy_(DFS) {
  Fragment *new_fragment = new Fragment();
  new_fragment->push_back(root);
  fragments_.push_back(new_fragment);
  active_fragment_ = 0;
}

Frontier::~Frontier() {
  for (const Fragment *fragment : fragments_) { delete fragment; }
  fragments_.clear();
}

// accessors

Frontier::Strategy Frontier::strategy() const { return strategy_; }

// mutators

void Frontier::add_node(const Node *node) {
  if (strategy_ == RANDOM) { return; }
  ASSERT(active_fragment_ >= 0, "Invalid active fragment index.");
  ASSERT((unsigned) active_fragment_ < fragments_.size(),
         "Vector of fragments is too short.");
  Fragment *fragment = fragments_[active_fragment_];
  fragment->push_back(node);
  std::push_heap(fragment->begin(), fragment->end(), Node::DecreasingOrder);
}

void Frontier::set_max_fragments(int max_fragments) {
  max_fragments_ = max_fragments;
}

void Frontier::set_strategy(Strategy strategy) { strategy_ = strategy; }

// public methods

std::vector<Move> Frontier::GenerateSchedule() {
  // Find a seed to generate a schedule from.
  const Node *seed = NULL;
  switch (strategy_) {
    case DFS: {
      active_fragment_ = System::Random(fragments_.size());
      Fragment *fragment = fragments_[active_fragment_];
      ASSERT(!fragment->empty(), "Fragment is empty.");
      seed = fragment->front();
      std::pop_heap(fragment->begin(), fragment->end(), Node::DecreasingOrder);
      fragment->pop_back();
      break;
    }
    case RANDOM: {
      seed = root_;
      while (seed->visited()) {
        // Count the number of edges that could be explored.
        int noptions = 0;
        std::map<int, const Node *> lookup;
        for (std::pair<int, const Edge *> pair : seed->edges()) {
          const Edge *edge = pair.second;
          const Node *node = edge->to();
          if (!node->explored() && (node->visited() || node->seeded())) {
            lookup[noptions++] = node;
          }
        }
        ASSERT(noptions > 0, "No edges to explore.");
        seed = lookup[System::Random(noptions)];
      }
      ASSERT(seed->seeded(), "Node has not been seeded.");
      break;
    }
    default:
      WARNING("Unknown explorer strategy: %u", strategy_);
      break;
  }
  // Generate a schedule from <seed>.
  std::list<dbug::thrift::Move> schedule;
  while (seed->parent()) {
    const Node *parent = seed->parent();
    int tid = -1;
    for (std::pair<int, const Edge *> edge : parent->edges()) {
      if (edge.second->to() == seed) {
        tid = edge.first;
        break;
      }
    }
    ASSERT(tid != -1, "Edge of thread %d is mising.", seed->id());
    dbug::thrift::Move move;
    move.choice_id = tid;
    move.nondet_events = parent->nondet_events();
    schedule.push_front(move);
    seed = parent;
  }
  std::vector<dbug::thrift::Move> result;
  result.insert(result.begin(), schedule.begin(), schedule.end());
  return result;
}

void Frontier::Partition() {
  if (strategy_ == RANDOM) { return; }
  // If the active fragment is empty, remove it from the frontier.
  ASSERT(active_fragment_ >= 0, "Invalid active fragment index.");
  ASSERT((unsigned) active_fragment_ < fragments_.size(),
         "Vector of fragments is too short.");
  const Fragment *fragment = fragments_[active_fragment_];
  if (fragment->empty()) {
    fragments_[active_fragment_] = fragments_[fragments_.size() - 1];
    fragments_.pop_back();
    delete fragment;
  }

  // Gather all fragments with at least two element, while reverting
  // the order of all fragments so that nodes closer to the root of the
  // execution tree are closer to the top of the fragment heap.
  std::priority_queue
      <Fragment *, std::vector<Fragment *>, Fragment::IncreasingOrder> frags;
  for (Fragment *fragment : fragments_) {
    std::make_heap(fragment->begin(), fragment->end(), Node::IncreasingOrder);
    if (fragment->size() >= 2) { frags.push(fragment); }
  }

  // While the number of existing fragments is lower than the maximum
  // number of fragments and there exist a fragment F with at least
  // two elements, create a new fragment using the top element of F.
  while (frags.size() > 0 && fragments_.size() < (unsigned) max_fragments_) {
    Fragment *fragment = frags.top();
    frags.pop();
    const Node *node = fragment->front();
    std::pop_heap(fragment->begin(), fragment->end(), Node::IncreasingOrder);
    fragment->pop_back();
    DEBUG("Creating a new fragment using a node with depth %d", node->depth());
    Fragment *new_fragment = new Fragment();
    new_fragment->push_back(node);
    fragments_.push_back(new_fragment);
    if (fragment->size() >= 2) { frags.push(fragment); }
  }

  // Revert the order of all fragments so that nodes farther from the
  // root of the execution tree are closer to the top of the fragment heap.
  for (Fragment *fragment : fragments_) {
    std::make_heap(fragment->begin(), fragment->end(), Node::DecreasingOrder);
  }
  DEBUG("Number of fragments after partition: %d", fragments_.size());
}

}  // namespace explorer
}  // namespace dbug
