/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/explorer/node.h"

#include <list>
#include <map>
#include <sstream>
#include <stack>
#include <string>

#include "dbug/common/util.h"
#include "dbug/explorer/edge.h"
#include "dbug/explorer/logging.h"

namespace dbug {
namespace explorer {

bool Node::DecreasingOrder(const Node *n1, const Node *n2) {
  ASSERT(n1, "Node does not exist.");
  ASSERT(n2, "Node does not exist.");
  return n1->depth() < n2->depth();
}

bool Node::IncreasingOrder(const Node *n1, const Node *n2) {
  ASSERT(n1, "Node does not exist.");
  ASSERT(n2, "Node does not exist.");
  return n1->depth() > n2->depth();
}

Node::Node(int id, Node *parent)
    : depth_(parent ? parent->depth() + 1 : 0),
      estimate_(0),
      explored_(false),
      id_(id),
      num_explored_branches_(0),
      parent_(parent),
      probability_(0),
      seeded_(false),
      seeds_computed_(false),
      visited_(false) {}

Node::~Node() { if (!parent_) { DeallocateSubtree(); } }

// accessors

std::map<int, Node *> Node::back_pointers() const { return back_pointers_; }

int Node::depth() const { return depth_; }

BigFloat Node::estimate() const { return estimate_; }

std::map<int, Edge *> Node::edges() const { return edges_; }

bool Node::explored() const { return explored_; }

int Node::id() const { return id_; }

std::vector<int> Node::nondet_events() const {
  return nondet_events_;
}

int Node::num_explored_branches() const { return num_explored_branches_; }

Node * Node::parent() const { return parent_; }

BigFloat Node::probability() const { return probability_; }

bool Node::seeded() const { return seeded_; }

bool Node::seeds_computed() const { return seeds_computed_; }

bool Node::visited() const { return visited_; }

// mutators

void Node::add_edge(int id, Edge *edge) { edges_[id] = edge; }

void Node::clear_edges() { edges_.clear(); }

void Node::set_back_pointers(const std::map<int, Node *> &back_pointers) {
  back_pointers_ = back_pointers;
}

void Node::set_estimate(BigFloat estimate) {
  estimate_ = estimate;
}

void Node::set_explored(bool explored) { explored_ = explored; }

void Node::set_nondet_events(std::vector<int> nondet_events) {
  nondet_events_ = nondet_events;
}

void Node::set_num_explored_branches(int num_explored_branches) {
  num_explored_branches_ = num_explored_branches;
}

void Node::set_parent(Node *parent) { parent_ = parent; }

void Node::set_probability(BigFloat probability) {
  probability_ = probability;
}

void Node::set_seeded(bool seeded) { seeded_ = seeded; }

void Node::set_seeds_computed(bool seeds_computed) {
  seeds_computed_ = seeds_computed;
}

void Node::set_visited(bool visited) { visited_ = visited; }

// public methods

void Node::DeallocateSubtree() {
  std::stack<Edge *> dfs_stack;
  std::map<int, Edge *>::iterator eit;
  for (std::pair<int, Edge *> pair : edges_) {
    Edge *edge = pair.second;
    dfs_stack.push(edge);
  }
  edges_.clear();
  while (!dfs_stack.empty()) {
    Edge *edge = dfs_stack.top();
    dfs_stack.pop();
    Node *node = edge->to();
    for (std::pair<int, Edge *> pair : node->edges()) {
      Edge *edge = pair.second;
      dfs_stack.push(edge);
    }
    delete node;
    delete edge;
  }
}

Edge *Node::GetEdge(int id) const {
  std::map<int, Edge *>::const_iterator it = edges_.find(id);
  if (it != edges_.end()) { return it->second; }
  return NULL;
}

Edge *Node::GetThreadEdge(int tid) const {
  for (std::pair<int, Edge *> pair : edges_) {
    Edge *edge = pair.second;
    for (const dbug::thrift::Thread &thread : edge->threads()) {
      if (thread.id == tid) { return edge; }
    }
  }
  return NULL;
}

Node *Node::FollowEdge(int id) const {
  Edge *edge = GetEdge(id);
  if (edge) { return edge->to(); }
  return NULL;
}

void Node::Print(Logging::Verbosity verbosity) const {
  const Logging::Config *config = Logging::get_channel(Logging::EXPLORER);
  if (config->verbosity() < verbosity) { return; }
  for (const std::string &line : ToString()) {
    LOG(verbosity, "%s", line.c_str());
  }
}

std::list<std::string> Node::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  stream << " Node:";
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  depth: " << depth_;
  result.push_back(stream.str());
  stream.str(std::string());
  for (std::pair<int, const Edge *> pair : edges_) {
    result.splice(result.end(), pair.second->ToString());
  }
  stream << "  estimate: " << estimate_.ToString();
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  explored: " << (explored_ ? "yes" : "no");
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  id: " << id_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  nondet events: "
         << Util::VectorToString<int>(nondet_events_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  number of explored branches: " << num_explored_branches_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  parent: " << (parent_ ? parent_->id() : 0);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  probability: " << probability_.ToString();
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  seeded: " << (seeded_ ? "yes" : "no");
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  seeds computed: " << (seeds_computed_ ? "yes" : "no");
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  visited: " << (visited_ ? "yes" : "no");
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

}  // namespace explorer
}  // namespace dbug
