/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/interposition/arbiter.h"

#include <map>
#include <utility>

#include "dbug/common/system.h"
#include "dbug/common/lock.h"

namespace dbug {
namespace interposition {

Arbiter::Arbiter() : nrunning_threads_(0) {
  System::MutexInit(&processes_mutex_, NULL);
  System::MutexInit(&threads_mutex_, NULL);
}

Arbiter::~Arbiter() {
  System::MutexLock(&threads_mutex_);
  for (std::pair<System::Thread, ThreadInfo *> p : threads_) {
    ThreadInfo *thread = p.second;
    delete thread;
  }
  threads_.clear();
  System::MutexUnlock(&threads_mutex_);
  System::MutexDestroy(&threads_mutex_);
  System::MutexDestroy(&processes_mutex_);
}

// accessors

int Arbiter::nrunning_threads() const {
  dbug::MutexLock lock(&threads_mutex_);
  return nrunning_threads_;
}

// public methods

int Arbiter::ProcessID() const {
  dbug::MutexLock lock(&processes_mutex_);
  std::map<System::Thread, int>::const_iterator it =
      processes_.find(System::ThreadSelf());
  if (it == processes_.end()) { return 0; }
  return processes_.at(System::ThreadSelf());
}

int Arbiter::ProcessID(const System::Thread thread) const {
  dbug::MutexLock lock(&processes_mutex_);
  std::map<System::Thread, int>::const_iterator it =
      processes_.find(thread);
  if (it == processes_.end()) { return 0; }
  return processes_.at(thread);
}

void Arbiter::Reset() {
  System::MutexInit(&threads_mutex_, NULL);
  nrunning_threads_ = 0;
  std::map<System::Thread, ThreadInfo *>::iterator mit;
  for (mit = threads_.begin(); mit != threads_.end(); ++mit) {
    delete mit->second;
  }
  threads_.clear();
}

int Arbiter::ThreadID() const {
  dbug::MutexLock lock(&threads_mutex_);
  std::map<System::Thread, ThreadInfo *>::const_iterator it =
      threads_.find(System::ThreadSelf());
  if (it == threads_.end()) { return 0; }
  return threads_.at(System::ThreadSelf())->id;
}

int Arbiter::ThreadID(const System::Thread thread) const {
  dbug::MutexLock lock(&threads_mutex_);
  std::map<System::Thread, ThreadInfo *>::const_iterator it =
      threads_.find(thread);
  if (it == threads_.end()) { return 0; }
  return threads_.at(thread)->id;
}

}  // namespace interposition
}  // namespace dbug
