/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/interposition/interface.h"

#include "dbug/common/configuration.h"
#include "dbug/common/init.h"
#include "dbug/common/logging.h"
#include "dbug/common/macros.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/original.h"

using dbug::Configuration;
using dbug::Logging;
using dbug::System;
using dbug::Util;
using dbug::interposition::Handler;
using dbug::interposition::Original;

int dbug_choose(int n) { return Handler::Choose(n); }

void dbug_detach() { Handler::Detach(); }

int dbug_nondet(int event) { return Handler::RecordNonDeterminism(event); }

void dbug_off() { Handler::Off(); }

void dbug_on() { Handler::On(); }

void dbug_ignore_mutex(pthread_mutex_t *mutex) {
  return Handler::IgnoreMutex(mutex);
}

void dbug_thread_active(pthread_t tid) { Handler::ThreadActive(tid); }

void dbug_thread_waiting() { Handler::ThreadWaiting(); }

// Upon exit, unregister all threads of the process and clean up.
void DBUG_DESTRUCTOR finalize(void) {
  Handler::Disable(NULL);
  Handler::Finalize();
  dbug::Finalize();
}

// The constructor can be invoked in several different scenarios:
// 1) A process is started by dBug. In this case, the state of the process
//    needs to be set up from scratch.
// 2) A process is started by another process via exec() or spawn().
//    In this case the environment variable DBUG_TAG is set.
//    The initial thread of the process has been already registered
//    at the arbiter at its ID can be retrieved from DBUG_TAG.
//    All other state needs to be set up from scratch.
void DBUG_CONSTRUCTOR initialize(void) {
  // Initialize common dbug components.
  dbug::Initialize();

  // Create a logging channel for interposition.
  System::File *log_file =
      Util::OpenLogFile("dbug-interposition.log");
  Logging::add_channel(Logging::INTERPOSITION,
                       log_file,
                       Configuration::interposition_verbosity());

  // Initialize pointers to original implementations of the interposed
  // functions.
  Original::Initialize();

  // Initialize interposition handler.
  Handler::Initialize();
}
