/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Add support for passing file descriptors to child
// processes through fork() and exec() functions.

// TODO(jsimsa): Model dup(), dup2(), pread(), pwrite(), recvmsg(),
// and sendmsg().

#include <sys/epoll.h>

#include <map>
#include <set>
#include <sstream>
#include <vector>

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/io.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_ACCEPT
int Handler::accept(int socket,
                    System::SocketAddress * DBUG_RESTRICT address,
                    socklen_t * DBUG_RESTRICT address_len) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    WARNING("Thread %d accepts a connection from an unknown socket %d.",
            arbiter_->ThreadID(), socket);
    System::MutexUnlock(&socket_map_mutex_);
    return Original::accept(socket, address, address_len);
  }
  int id = Util::NewIdentifier();
  std::string resource_id = socket_map_[socket];
  std::string new_resource_id =
      Util::CreateResourceID("socket", arbiter_->ProcessID(), id);
  System::MutexUnlock(&socket_map_mutex_);
  arbiter_->SocketAccept(arbiter_->ThreadID(),
                         "accept",
                         Util::GetBacktrace(),
                         resource_id,
                         new_resource_id);
  int ret = Original::accept(socket, address, address_len);
  if (ret != -1) {
    System::MutexLock(&socket_map_mutex_);
    socket_map_[ret] = new_resource_id;
    System::MutexUnlock(&socket_map_mutex_);
    // Report socket buffer size
    int opt;
    socklen_t opt_size = sizeof(opt_size);
    System::GetSocketOption(ret, SOL_SOCKET, SO_RCVBUF,
                            reinterpret_cast<void *>(&opt), &opt_size);
    arbiter_->SetSocketSize(arbiter_->ThreadID(), new_resource_id, opt);
  }
  return ret;
}
#endif

#if DBUG_HAVE_BIND
int Handler::bind(int socket,
                  const System::SocketAddress * addr,
                  socklen_t addrlen) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    WARNING("Thread %d binds an unknown socket %d.",
            arbiter_->ThreadID(), socket);
    System::MutexUnlock(&socket_map_mutex_);
    return Original::bind(socket, addr, addrlen);
  }
  std::string resource_id = socket_map_[socket];
  System::MutexUnlock(&socket_map_mutex_);
  switch (addr->sa_family) {
    case AF_INET: {
      int ret = Original::bind(socket, addr, addrlen);
      if (ret != -1) {
        System::SocketAddress addr;
        System::GetSocketName(socket, &addr, &addrlen);
        const System::SocketIPv4Address *ip_addr =
            reinterpret_cast<System::SocketIPv4Address *>(&addr);
        std::string address = System::IPv4NetworkToString(ip_addr->sin_addr);
        unsigned short port = System::NetworkToHostShort(ip_addr->sin_port);
        arbiter_->SocketBind(arbiter_->ThreadID(),
                             "bind",
                             Util::GetBacktrace(),
                             resource_id,
                             address,
                             port);
      }
      return ret;
    }
    case AF_INET6: {
      int ret = Original::bind(socket, addr, addrlen);
      if (ret != -1) {
        System::SocketAddress addr;
        System::GetSocketName(socket, &addr, &addrlen);
        const System::SocketIPv6Address *ip_addr =
            reinterpret_cast<System::SocketIPv6Address *>(&addr);
        std::string address = System::IPv6NetworkToString(ip_addr->sin6_addr);
        unsigned short port = System::NetworkToHostShort(ip_addr->sin6_port);
        arbiter_->SocketBind(arbiter_->ThreadID(),
                             "bind",
                             Util::GetBacktrace(),
                             resource_id,
                             address,
                             port);
      }
      return ret;
    }
    default:
      WARNING("Thread %d: dbug does not model socket address family %d.",
              arbiter_->ThreadID(), addr->sa_family);
      return Original::bind(socket, addr, addrlen);
  }
}
#endif

#if DBUG_HAVE_CLOSE
int Handler::close(int fd) {
  // Epoll
  System::MutexLock(&epoll_map_mutex_);
  if (epoll_map_.find(fd) != epoll_map_.end()) {
    std::string resource_id = epoll_map_[fd];
    System::MutexUnlock(&epoll_map_mutex_);
    arbiter_->EPollDelete(arbiter_->ThreadID(),
                          "close",
                          Util::GetBacktrace(),
                          resource_id);
    // Currently dbug does not support file descriptor reuse. Instead
    // of calling Original::close(fd), we return 0 to prevent the OS
    // from reusing the file descriptor.
    return 0;
  }
  System::MutexUnlock(&epoll_map_mutex_);
  // File
  System::MutexLock(&file_map_mutex_);
  if (file_map_.find(fd) != file_map_.end()) {
    std::string resource_id = file_map_[fd];
    System::MutexUnlock(&file_map_mutex_);
    arbiter_->FileClose(arbiter_->ThreadID(),
                        "close",
                        Util::GetBacktrace(),
                        resource_id);
    // Currently dbug does not support file descriptor reuse. Instead
    // of calling Original::close(fd), we return 0 to prevent the OS
    // from reusing the file descriptor.
    return 0;
  }
  System::MutexUnlock(&file_map_mutex_);
  // Pipe
  System::MutexLock(&pipe_map_mutex_);
  if (pipe_map_.find(fd) != pipe_map_.end()) {
    std::string resource_id = pipe_map_[fd];
    System::MutexUnlock(&pipe_map_mutex_);
    arbiter_->PipeDelete(arbiter_->ThreadID(),
                         "close",
                         Util::GetBacktrace(),
                         resource_id,
                         fd);
    // Currently dbug does not support file descriptor reuse. Instead
    // of calling Original::close(fd), we return 0 to prevent the OS
    // from reusing the file descriptor.
    return 0;
  }
  System::MutexUnlock(&pipe_map_mutex_);
  // Socket
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(fd) != socket_map_.end()) {
    std::string resource_id = socket_map_[fd];
    System::MutexUnlock(&socket_map_mutex_);
    arbiter_->SocketClose(arbiter_->ThreadID(),
                          "close",
                          Util::GetBacktrace(),
                          resource_id);
    // Currently dbug does not support file descriptor reuse. Instead
    // of calling Original::close(fd), we return 0 to prevent the OS
    // from reusing the file descriptor.
    return 0;
  }
  System::MutexUnlock(&socket_map_mutex_);
  WARNING("Thread %d closes an unknown file descriptor %d.",
          arbiter_->ThreadID(), fd);
  return Original::close(fd);
}
#endif

#if DBUG_HAVE_CONNECT
int Handler::connect(int socket,
                     const System::SocketAddress *addr,
                     socklen_t addrlen) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    WARNING("Thread %d connects to an unknown socket %d.",
            arbiter_->ThreadID(), socket);
    System::MutexUnlock(&socket_map_mutex_);
    return Original::connect(socket, addr, addrlen);
  }
  std::string resource_id = socket_map_[socket];
  System::MutexUnlock(&socket_map_mutex_);
  switch (addr->sa_family) {
    case AF_INET: {
      const System::SocketIPv4Address *ip_addr =
          reinterpret_cast<const System::SocketIPv4Address *>(addr);
      std::string address = System::IPv4NetworkToString(ip_addr->sin_addr);
      unsigned short port = System::NetworkToHostShort(ip_addr->sin_port);
      arbiter_->SocketConnect(arbiter_->ThreadID(),
                              "connect",
                              Util::GetBacktrace(),
                              resource_id,
                              address,
                              static_cast<int>(port));
      break;
    }
    case AF_INET6: {
      const System::SocketIPv6Address *ip_addr =
          reinterpret_cast<const System::SocketIPv6Address *>(addr);
      std::string address = System::IPv6NetworkToString(ip_addr->sin6_addr);
      unsigned short port = System::NetworkToHostShort(ip_addr->sin6_port);
      arbiter_->SocketConnect(arbiter_->ThreadID(),
                              "connect",
                              Util::GetBacktrace(),
                              resource_id,
                              address,
                              static_cast<int>(port));
      break;
    }
    default:
      WARNING("Thread %d: dbug does not model socket address family %d.",
              arbiter_->ThreadID(), addr->sa_family);
      break;
  }
  return Original::connect(socket, addr, addrlen);
}
#endif

#if DBUG_HAVE_DUP
int Handler::dup(int oldfd) {
  WARNING("Thread %d: dbug does not model dup().", arbiter_->ThreadID());
  return Original::dup(oldfd);
}
#endif

#if DBUG_HAVE_DUP2
int Handler::dup2(int oldfd, int newfd) {
  WARNING("Thread %d: dbug does not model dup2().", arbiter_->ThreadID());
  return Original::dup2(oldfd, newfd);
}
#endif

#if DBUG_HAVE_EPOLL_CREATE
int Handler::epoll_create(int size) {
  int id = Util::NewIdentifier();
  std::string resource_id =
      Util::CreateResourceID("epoll", arbiter_->ProcessID(), id);
  arbiter_->EPollCreate(arbiter_->ThreadID(),
                        "epoll_create",
                        Util::GetBacktrace(),
                        resource_id);
  int ret = Original::epoll_create(size);
  if (ret != -1) {
    System::MutexLock(&epoll_map_mutex_);
    epoll_map_[ret] = resource_id;
    System::MutexUnlock(&epoll_map_mutex_);
  } else {
    arbiter_->EPollDelete(arbiter_->ThreadID(),
                          "epoll_create",
                          Util::GetBacktrace(),
                          resource_id);
  }
  return ret;
}
#endif

#if DBUG_HAVE_EPOLL_CREATE1
int Handler::epoll_create1(int flags) {
  int id = Util::NewIdentifier();
  std::string resource_id =
      Util::CreateResourceID("epoll", arbiter_->ProcessID(), id);
  arbiter_->EPollCreate(arbiter_->ThreadID(),
                        "epoll_create1",
                        Util::GetBacktrace(),
                        resource_id);
  int ret = Original::epoll_create1(flags);
  if (ret != -1) {
    System::MutexLock(&epoll_map_mutex_);
    epoll_map_[ret] = resource_id;
    System::MutexUnlock(&epoll_map_mutex_);
  } else {
    arbiter_->EPollDelete(arbiter_->ThreadID(),
                          "epoll_create",
                          Util::GetBacktrace(),
                          resource_id);
  }
  return ret;
}
#endif

#if DBUG_HAVE_EPOLL_CTL
int Handler::epoll_ctl(int epfd, int op, int fd, struct epoll_event *event) {
  System::MutexLock(&epoll_map_mutex_);
  if (epoll_map_.find(epfd) != epoll_map_.end()) {
    std::string resource_id = epoll_map_[epfd];
    System::MutexUnlock(&epoll_map_mutex_);
    std::string fd_rid = "";
    System::MutexLock(&file_map_mutex_);
    if (file_map_.find(fd) != file_map_.end()) { fd_rid = file_map_[fd]; }
    System::MutexUnlock(&file_map_mutex_);
    System::MutexLock(&file_map_mutex_);
    if (pipe_map_.find(fd) != pipe_map_.end()) { fd_rid = pipe_map_[fd]; }
    System::MutexUnlock(&file_map_mutex_);
    System::MutexLock(&socket_map_mutex_);
    if (socket_map_.find(fd) != socket_map_.end()) { fd_rid = socket_map_[fd]; }
    System::MutexUnlock(&socket_map_mutex_);
    if (fd_rid == "") {
      WARNING("Thread %d controls an epoll with an unknown file descriptor %d.",
              arbiter_->ThreadID(), fd);
    } else {
      dbug::thrift::EPollEvent event_thrift;
      event_thrift.fd_rid = fd_rid;
      event_thrift.events = event->events;
      event_thrift.data = event->data.u64;
      arbiter_->EPollControl(arbiter_->ThreadID(),
                             "epoll_ctl",
                             Util::GetBacktrace(),
                             resource_id,
                             op,
                             event_thrift);
    }
    return Original::epoll_ctl(epfd, op, fd, event);
  }
  System::MutexUnlock(&epoll_map_mutex_);
  WARNING("Thread %d controls an unknown epoll %d.",
          arbiter_->ThreadID(), epfd);
  return Original::epoll_ctl(epfd, op, fd, event);
}
#endif

#if DBUG_HAVE_EPOLL_WAIT
int Handler::epoll_wait(int epfd, struct epoll_event *events,
                        int maxevents, int timeout) {
  System::MutexLock(&epoll_map_mutex_);
  if (epoll_map_.find(epfd) != epoll_map_.end()) {
    std::string resource_id = epoll_map_[epfd];
    System::MutexUnlock(&epoll_map_mutex_);
    std::vector<dbug::thrift::EPollEvent> return_value =
        arbiter_->EPollWait(arbiter_->ThreadID(),
                            "epoll_wait",
                            Util::GetBacktrace(),
                            resource_id,
                            maxevents);
    // Process the returned event set.
    int i = 0;
    for (dbug::thrift::EPollEvent event : return_value) {
      events[i].events = event.events;
      events[i].data.u64 = event.data;
      i++;
    }
    return i;
  }
  System::MutexUnlock(&epoll_map_mutex_);
  WARNING("Thread %d waits on an unknown epoll %d.",
          arbiter_->ThreadID(), epfd);
  return Original::epoll_wait(epfd, events, maxevents, timeout);
}
#endif

#if DBUG_HAVE_FCNTL
int Handler::fcntl(int fd, int cmd, va_list ap) {
  int ret;
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    switch (cmd) {
      case F_DUPFD:
      case F_DUPFD_CLOEXEC:
      case F_NOTIFY:
      case F_SETLEASE:
      case F_SETPIPE_SZ:
      case F_SETOWN:
      case F_SETSIG: {
        WARNING("Thread %d: dbug does not model fcntl() command %s.",
                arbiter_->ThreadID(),
                System::FileControlCommandToString(cmd).c_str());
        long arg = va_arg(ap, long);  // NOLINT
        ret = Original::fcntl(fd, cmd, arg);
        DEBUG_EXTERNAL("%d,%d,%ld", "%d", fd, cmd, arg, ret);
        break;
      }
      case F_GETFD:
      case F_GETFL:
      case F_GETLEASE:
      case F_GETPIPE_SZ:
      case F_GETOWN:
      case F_GETSIG: {
        WARNING("Thread %d: dbug does not model fcntl() command %s.",
                arbiter_->ThreadID(),
                System::FileControlCommandToString(cmd).c_str());
        ret = Original::fcntl(fd, cmd);
        DEBUG_EXTERNAL("%d,%d", "%d", fd, cmd, ret);
        break;
      }
      case F_GETLK:
      case F_SETLK:
      case F_SETLKW: {
        WARNING("Thread %d: dbug does not model fcntl() command %s.",
                arbiter_->ThreadID(),
                System::FileControlCommandToString(cmd).c_str());
        struct flock *arg = va_arg(ap, struct flock *);
        ret = Original::fcntl(fd, cmd, arg);
        DEBUG_EXTERNAL("%d,%d,%p", "%d", fd, cmd, arg, ret);
        break;
      }
      case F_GETOWN_EX:
      case F_SETOWN_EX: {
        WARNING("Thread %d: dbug does not model fcntl() command %s.",
                arbiter_->ThreadID(),
                System::FileControlCommandToString(cmd).c_str());
        struct f_owner_ex *arg = va_arg(ap, struct f_owner_ex *);
        ret = Original::fcntl(fd, cmd, arg);
        DEBUG_EXTERNAL("%d,%d,%p", "%d", fd, cmd, arg, ret);
        break;
      }
      case F_SETFD:  {
        long arg = va_arg(ap, long);  // NOLINT
        // Clear out the FD_CLOEXEC bit to prevent the OS from reusing
        // this file descriptor upon exec().
        ret = Original::fcntl(fd, cmd, arg & ~FD_CLOEXEC);
        DEBUG_EXTERNAL("%d,%d,%ld", "%d", fd, cmd, arg, ret);
        std::string rid;
        System::MutexLock(&file_map_mutex_);
        if (file_map_.find(fd) != file_map_.end()) {
          rid = file_map_[fd];
        }
        System::MutexUnlock(&file_map_mutex_);
        if (rid != "") {
          arbiter_->SetFileFlags(arbiter_->ThreadID(), rid, arg);
          break;
        }
        System::MutexLock(&pipe_map_mutex_);
        if (pipe_map_.find(fd) != pipe_map_.end()) {
          rid = pipe_map_[fd];
        }
        System::MutexUnlock(&pipe_map_mutex_);
        if (rid != "") {
          arbiter_->SetPipeFlags(arbiter_->ThreadID(), rid, arg);
          break;
        }
        System::MutexLock(&socket_map_mutex_);
        if (socket_map_.find(fd) != socket_map_.end()) {
          rid = socket_map_[fd];
        }
        System::MutexUnlock(&socket_map_mutex_);
        if (rid != "") {
          arbiter_->SetSocketFlags(arbiter_->ThreadID(), rid, arg);
          break;
        }
        WARNING("Thread %d sets descriptor flags "
                "of an unknown file descriptor %d.",
                arbiter_->ThreadID(), fd);
      }
      case F_SETFL: {
        long arg = va_arg(ap, long);  // NOLINT
        ret = Original::fcntl(fd, cmd, arg);
        DEBUG_EXTERNAL("%d,%d,%ld", "%d", fd, cmd, arg, ret);
        std::string rid;
        System::MutexLock(&file_map_mutex_);
        if (file_map_.find(fd) != file_map_.end()) {
          rid = file_map_[fd];
        }
        System::MutexUnlock(&file_map_mutex_);
        if (rid != "") {
          arbiter_->SetFileFlags(arbiter_->ThreadID(), rid, arg);
          break;
        }
        System::MutexLock(&pipe_map_mutex_);
        if (pipe_map_.find(fd) != pipe_map_.end()) {
          rid = pipe_map_[fd];
        }
        System::MutexUnlock(&pipe_map_mutex_);
        if (rid != "") {
          arbiter_->SetPipeFlags(arbiter_->ThreadID(), rid, arg);
          break;
        }
        System::MutexLock(&socket_map_mutex_);
        if (socket_map_.find(fd) != socket_map_.end()) {
          rid = socket_map_[fd];
        }
        System::MutexUnlock(&socket_map_mutex_);
        if (rid != "") {
          arbiter_->SetSocketFlags(arbiter_->ThreadID(), rid, arg);
          break;
        }
        WARNING("Thread %d sets status flags "
                "of an unknown file descriptor %d.",
                arbiter_->ThreadID(), fd);
      }
      default: {
        WARNING("Thread %d issues an unknown fcntl() command %s.",
                arbiter_->ThreadID(),
                System::FileControlCommandToString(cmd).c_str());
        ret = Original::fcntl(fd, cmd);
        DEBUG_EXTERNAL("%d,%d", "%d", fd, cmd, ret);
      }
    }
    return ret;
  } else {
    DBUG_REQUIRE_SYMBOL(Original::fcntl, FCNTL_TYPE, "fcntl");
    switch (cmd) {
      case F_DUPFD:
      case F_DUPFD_CLOEXEC:
      case F_NOTIFY:
      case F_SETFD:
      case F_SETFL:
      case F_SETLEASE:
      case F_SETPIPE_SZ:
      case F_SETOWN:
      case F_SETSIG: {
        long arg = va_arg(ap, long);  // NOLINT
        ret = Original::fcntl(fd, cmd, arg);
        DEBUG_INTERNAL("%d,%d,%ld", "%d", fd, cmd, arg, ret);
        break;
      }
      case F_GETLK:
      case F_SETLK:
      case F_SETLKW: {
        struct flock *arg = va_arg(ap, struct flock *);
        ret = Original::fcntl(fd, cmd, arg);
        DEBUG_INTERNAL("%d,%d,%p", "%d", fd, cmd, arg, ret);
        break;
      }
      case F_GETOWN_EX:
      case F_SETOWN_EX: {
        struct f_owner_ex *arg = va_arg(ap, struct f_owner_ex *);
        ret = Original::fcntl(fd, cmd, arg);
        DEBUG_INTERNAL("%d,%d,%p", "%d", fd, cmd, arg, ret);
        break;
      }
      case F_GETFD:
      case F_GETFL:
      case F_GETLEASE:
      case F_GETPIPE_SZ:
      case F_GETOWN:
      case F_GETSIG:
      default: {
        ret = Original::fcntl(fd, cmd);
        DEBUG_INTERNAL("%d,%d", "%d", fd, cmd, ret);
      }
    }
    return ret;
  }
}
#endif

#if DBUG_HAVE_LISTEN
int Handler::listen(int socket, int backlog) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    WARNING("Thread %d listens on an unknown %d.",
            arbiter_->ThreadID(), socket);
    System::MutexUnlock(&socket_map_mutex_);
    return Original::listen(socket, backlog);
  }
  std::string resource_id = socket_map_[socket];
  System::MutexUnlock(&socket_map_mutex_);
  arbiter_->SocketListen(arbiter_->ThreadID(),
                         "listen",
                         Util::GetBacktrace(),
                         resource_id);
  return Original::listen(socket, backlog);
}
#endif

#if DBUG_HAVE_OPEN
int Handler::open(const char *pathname, int flags, va_list ap) {
  int ret;
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    int id = Util::NewIdentifier();
    std::string resource_id =
        Util::CreateResourceID("file", arbiter_->ProcessID(), id);
    arbiter_->FileOpen(arbiter_->ThreadID(),
                       "open",
                       Util::GetBacktrace(),
                       resource_id);
    mode_t mode = va_arg(ap, mode_t);
    // Clear out the O_CLOEXEC bit to prevent the OS from reusing
    // this file descriptor upon exec().
    if (mode != 0) {
      ret = Original::open(pathname, flags & ~O_CLOEXEC, mode);
      DEBUG_EXTERNAL("%s,%d,%lu", "%d", pathname, flags, mode, ret);
    } else {
      ret = Original::open(pathname, flags & ~O_CLOEXEC);
      DEBUG_EXTERNAL("%s,%d", "%d", pathname, flags, ret);
    }
    if (ret != -1) {
      System::MutexLock(&file_map_mutex_);
      file_map_[ret] = resource_id;
      System::MutexUnlock(&file_map_mutex_);
    } else {
      arbiter_->FileClose(arbiter_->ThreadID(),
                          "open",
                          Util::GetBacktrace(),
                          resource_id);
    }
    return ret;
  } else {
    DBUG_REQUIRE_SYMBOL(Original::open, OPEN_TYPE, "open");
    mode_t mode = va_arg(ap, mode_t);
    if (mode != 0) {
      ret = Original::open(pathname, flags, mode);
      DEBUG_INTERNAL("%s,%d,%lu", "%d", pathname, flags, mode, ret);
    } else {
      ret = Original::open(pathname, flags);
      DEBUG_INTERNAL("%s,%d", "%d", pathname, flags, ret);
    }
    return ret;
  }
}
#endif

#if DBUG_HAVE_PIPE
int Handler::pipe(int pipefd[2]) {
  std::string rid =
      arbiter_->PipeCreate(arbiter_->ThreadID(), "pipe", Util::GetBacktrace());
  int ret = Original::pipe(pipefd);
  if (ret != -1) {
    System::MutexLock(&pipe_map_mutex_);
    pipe_map_[pipefd[0]] = rid;
    pipe_map_[pipefd[1]] = rid;
    System::MutexUnlock(&pipe_map_mutex_);
    int size = System::FileControl(pipefd[0], F_GETPIPE_SZ);
    arbiter_->SetPipeSize(arbiter_->ThreadID(), rid, size);
    arbiter_->SetPipeInput(arbiter_->ThreadID(), rid, pipefd[0]);
    arbiter_->SetPipeOutput(arbiter_->ThreadID(), rid, pipefd[1]);
  } else {
    arbiter_->PipeDelete(arbiter_->ThreadID(),
                         "pipe",
                         Util::GetBacktrace(),
                         rid,
                         pipefd[0]);
    arbiter_->PipeDelete(arbiter_->ThreadID(),
                         "pipe",
                         Util::GetBacktrace(),
                         rid,
                         pipefd[1]);
  }
  return ret;
}
#endif

#if DBUG_HAVE_POLL
int Handler::poll(struct pollfd *fds, nfds_t nfds, int timeout) {
  std::set<std::string> dbug_read_fds, dbug_write_fds, empty;
  // Encode poll file descriptors
  System::MutexLock(&file_map_mutex_);
  System::MutexLock(&pipe_map_mutex_);
  System::MutexLock(&socket_map_mutex_);
  for (unsigned i = 0; i < nfds; i++) {
    if (file_map_.find(fds[i].fd) != file_map_.end()) {
      if ((fds[i].events & ~(POLLIN | POLLOUT)) != 0) {
        WARNING("Thread %d: dbug does not model poll events %d.",
                arbiter_->ThreadID(), fds[i].events);
      }
      std::string resource_id = file_map_[fds[i].fd];
      if (fds[i].events & POLLIN) { dbug_read_fds.insert(resource_id); }
      if (fds[i].events & POLLOUT) { dbug_write_fds.insert(resource_id); }
      continue;
    }
    if (pipe_map_.find(fds[i].fd) != pipe_map_.end()) {
      if ((fds[i].events & ~(POLLIN | POLLOUT)) != 0) {
        WARNING("Thread %d: dbug does not model poll events %d.",
                arbiter_->ThreadID(), fds[i].events);
      }
      std::string resource_id = pipe_map_[fds[i].fd];
      if (fds[i].events & POLLIN) { dbug_read_fds.insert(resource_id); }
      if (fds[i].events & POLLOUT) { dbug_write_fds.insert(resource_id); }
      continue;
    }
    if (socket_map_.find(fds[i].fd) != socket_map_.end()) {
      if ((fds[i].events & ~(POLLIN | POLLOUT)) != 0) {
        WARNING("Thread %d: dbug does not model poll events %d.",
                arbiter_->ThreadID(), fds[i].events);
      }
      std::string resource_id = socket_map_[fds[i].fd];
      if (fds[i].events & POLLIN) { dbug_read_fds.insert(resource_id); }
      if (fds[i].events & POLLOUT) { dbug_write_fds.insert(resource_id); }
      continue;
    }
    WARNING("Thread %d polls an unknown file descriptor %d.",
            arbiter_->ThreadID(), fds[i].fd);
  }
  System::MutexUnlock(&socket_map_mutex_);
  System::MutexUnlock(&pipe_map_mutex_);
  System::MutexUnlock(&file_map_mutex_);

  // If there were no file descriptors, return 0.
  if (dbug_read_fds.empty() && dbug_write_fds.empty()) { return 0; }

  // Have arbiter resolve the call to socket.
  dbug::thrift::SelectReturn select_return =
      arbiter_->Select(arbiter_->ThreadID(),
                       "poll",
                       Util::GetBacktrace(),
                       dbug_read_fds,
                       dbug_write_fds,
                       timeout >= 0);

  // Process returned arguments.
  //
  // TODO(jsimsa): This could be made faster if mapping of resource
  // IDs to file descriptors was maintained.
  for (unsigned i = 0; i < nfds; i++) { fds[i].revents = 0; }

  int ret = 0;
  // File
  System::MutexLock(&file_map_mutex_);
  for (std::pair<int, std::string> file : file_map_) {
    std::string resource_id = file.second;
    for (unsigned i = 0; i < nfds; i++) {
      if (fds[i].fd == file.first) {
        if (select_return.rfds.find(resource_id) != select_return.rfds.end()) {
          fds[i].revents |= POLLIN;
        }
        if (select_return.wfds.find(resource_id) != select_return.wfds.end()) {
          fds[i].revents |= POLLOUT;
        }
        if (fds[i].revents != 0) { ret++; }
        break;
      }
    }
  }
  System::MutexUnlock(&file_map_mutex_);
  // Pipe
  System::MutexLock(&pipe_map_mutex_);
  for (std::pair<int, std::string> pipe : pipe_map_) {
    std::string resource_id = pipe.second;
    for (unsigned i = 0; i < nfds; i++) {
      if (fds[i].fd == pipe.first) {
        if (select_return.rfds.find(resource_id) != select_return.rfds.end()) {
          fds[i].revents |= POLLIN;
        }
        if (select_return.wfds.find(resource_id) != select_return.wfds.end()) {
          fds[i].revents |= POLLOUT;
        }
        if (fds[i].revents != 0) { ret++; }
        break;
      }
    }
  }
  System::MutexUnlock(&pipe_map_mutex_);
  // Socket
  System::MutexLock(&socket_map_mutex_);
  for (std::pair<int, std::string> socket : socket_map_) {
    std::string resource_id = socket.second;
    for (unsigned i = 0; i < nfds; i++) {
      if (fds[i].fd == socket.first) {
        if (select_return.rfds.find(resource_id) != select_return.rfds.end()) {
          fds[i].revents |= POLLIN;
        }
        if (select_return.wfds.find(resource_id) != select_return.wfds.end()) {
          fds[i].revents |= POLLOUT;
        }
        if (fds[i].revents != 0) { ret++; }
        break;
      }
    }
  }
  System::MutexUnlock(&socket_map_mutex_);
  return ret;
}
#endif

#if DBUG_HAVE_PREAD
ssize_t Handler::pread(int fd, void *buf, size_t count, off_t offset) {
  WARNING("Thread %d: dbug does not model pread().", arbiter_->ThreadID());
  return Original::pread(fd, buf, count, offset);
}
#endif

#if DBUG_HAVE_PWRITE
ssize_t Handler::pwrite(int fd, const void *buf, size_t count, off_t offset) {
  WARNING("Thread %d: dbug does not model pwrite().", arbiter_->ThreadID());
  return Original::pwrite(fd, buf, count, offset);
}
#endif

#if DBUG_HAVE_READ
ssize_t Handler::read(int fd, void *buf, size_t count) {
  // File
  System::MutexLock(&file_map_mutex_);
  if (file_map_.find(fd) != file_map_.end()) {
    std::string resource_id = file_map_[fd];
    System::MutexUnlock(&file_map_mutex_);
    arbiter_->FileRead(arbiter_->ThreadID(),
                       "read",
                       Util::GetBacktrace(),
                       resource_id,
                       count);
    ssize_t ret = Original::read(fd, buf, count);
    return ret;
  }
  System::MutexUnlock(&file_map_mutex_);
  // Pipe
  System::MutexLock(&pipe_map_mutex_);
  if (pipe_map_.find(fd) != pipe_map_.end()) {
    std::string resource_id = pipe_map_[fd];
    System::MutexUnlock(&pipe_map_mutex_);
    arbiter_->PipeRead(arbiter_->ThreadID(),
                       "read",
                       Util::GetBacktrace(),
                       resource_id,
                       count);
    ssize_t ret = Original::read(fd, buf, count);
    if (ret > 0) {
      arbiter_->UpdatePipeBytes(arbiter_->ThreadID(), resource_id, -ret);
    }
    return ret;
  }
  System::MutexUnlock(&pipe_map_mutex_);
  // Socket
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(fd) != socket_map_.end()) {
    std::string resource_id = socket_map_[fd];
    System::MutexUnlock(&socket_map_mutex_);
    arbiter_->SocketReceive(arbiter_->ThreadID(),
                            "read",
                            Util::GetBacktrace(),
                            resource_id,
                            count,
                            0);
    ssize_t ret = Original::read(fd, buf, count);
    if (ret > 0) {
      arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, -ret);
    }
    return ret;
  }
  System::MutexUnlock(&socket_map_mutex_);
  WARNING("Thread %d reads from an unknown file descriptor %d.",
          arbiter_->ThreadID(), fd);
  return Original::read(fd, buf, count);
}
#endif

#if DBUG_HAVE_READV
ssize_t Handler::readv(int fd, const struct iovec *iov, int iovcnt) {
  int count = 0;
  for (int i = 0; i < iovcnt; i++) { count += iov[i].iov_len; }
  // File
  System::MutexLock(&file_map_mutex_);
  if (file_map_.find(fd) != file_map_.end()) {
    std::string resource_id = file_map_[fd];
    System::MutexUnlock(&file_map_mutex_);
    arbiter_->FileRead(arbiter_->ThreadID(),
                       "readv",
                       Util::GetBacktrace(),
                       resource_id,
                       count);
    ssize_t ret = Original::readv(fd, iov, iovcnt);
    return ret;
  }
  System::MutexUnlock(&file_map_mutex_);
  // Pipe
  System::MutexLock(&pipe_map_mutex_);
  if (pipe_map_.find(fd) != pipe_map_.end()) {
    std::string resource_id = pipe_map_[fd];
    System::MutexUnlock(&pipe_map_mutex_);
    arbiter_->PipeRead(arbiter_->ThreadID(),
                       "readv",
                       Util::GetBacktrace(),
                       resource_id,
                       count);
    ssize_t ret = Original::readv(fd, iov, iovcnt);
    if (ret > 0) {
      arbiter_->UpdatePipeBytes(arbiter_->ThreadID(), resource_id, -ret);
    }
    return ret;
  }
  System::MutexUnlock(&pipe_map_mutex_);
  // Socket
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(fd) != socket_map_.end()) {
    std::string resource_id = socket_map_[fd];
    System::MutexUnlock(&socket_map_mutex_);
    arbiter_->SocketReceive(arbiter_->ThreadID(),
                            "readv",
                            Util::GetBacktrace(),
                            resource_id,
                            count,
                            0);
    ssize_t ret = Original::readv(fd, iov, iovcnt);
    if (ret > 0) {
      arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, -ret);
    }
    return ret;
  }
  System::MutexUnlock(&socket_map_mutex_);
  WARNING("Thread %d reads from an unknown file descriptor %d.",
          arbiter_->ThreadID(), fd);
  return Original::readv(fd, iov, iovcnt);
}
#endif

#if DBUG_HAVE_RECV
ssize_t Handler::recv(int socket, void * buffer, size_t length, int flags) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    System::MutexUnlock(&socket_map_mutex_);
    WARNING("Thread %d receives from an unknown socket %d.",
            arbiter_->ThreadID(), socket);
    return Original::recv(socket, buffer, length, flags);
  }
  std::string resource_id = socket_map_[socket];
  System::MutexUnlock(&socket_map_mutex_);
  arbiter_->SocketReceive(arbiter_->ThreadID(),
                          "recv",
                          Util::GetBacktrace(),
                          resource_id,
                          length,
                          flags);
  ssize_t ret = Original::recv(socket, buffer, length, flags);
  if (ret > 0) {
    arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, -ret);
  }
  return ret;
}
#endif

#if DBUG_HAVE_RECVFROM
ssize_t Handler::recvfrom(int socket,
                          void * DBUG_RESTRICT buffer,
                          size_t length,
                          int flags,
                          System::SocketAddress * DBUG_RESTRICT src_addr,
                          socklen_t * DBUG_RESTRICT addrlen) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    System::MutexUnlock(&socket_map_mutex_);
    WARNING("Thread %d receives from an unknown socket %d.",
            arbiter_->ThreadID(), socket);
    return Original::recvfrom(socket, buffer, length,
                              flags, src_addr, addrlen);
  }
  std::string resource_id = socket_map_[socket];
  System::MutexUnlock(&socket_map_mutex_);
  if (!src_addr) {
    arbiter_->SocketReceive(arbiter_->ThreadID(),
                            "recvfrom",
                            Util::GetBacktrace(),
                            resource_id,
                            length,
                            flags);
  } else {
    arbiter_->SocketReceiveFrom(arbiter_->ThreadID(),
                                "recvfrom",
                                Util::GetBacktrace(),
                                resource_id,
                                length,
                                flags);
  }
  ssize_t ret = Original::recvfrom(socket, buffer, length,
                                   flags, src_addr, addrlen);
  if (ret > 0) {
    arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, -ret);
  }
  return ret;
}
#endif

#if DBUG_HAVE_RECVMSG
ssize_t Handler::recvmsg(int socket, struct msghdr * message, int flags) {
  WARNING("Thread %d: dbug does not model recvmsg().", arbiter_->ThreadID());
  return Original::recvmsg(socket, message, flags);
}
#endif

#if DBUG_HAVE_SELECT
int Handler::select(int nfds,
                    fd_set *readfds,
                    fd_set *writefds,
                    fd_set *exceptfds,
                    System::TimeValue *timeout) {
  // Encode file descriptors as sets of resource identifiers
  std::set<std::string> rfds, wfds;
  if (exceptfds) {
    WARNING("Thread %d: dbug does not model exception file descriptors.",
            arbiter_->ThreadID());
  }
  System::MutexLock(&file_map_mutex_);
  System::MutexLock(&pipe_map_mutex_);
  System::MutexLock(&socket_map_mutex_);
  for (int i = 0; i < nfds; i++) {
    // Read file descriptors
    if ((readfds && System::FDSetContains(i, readfds))) {
      bool known = false;
      if (file_map_.find(i) != file_map_.end()) {
        std::string resource_id = file_map_[i];
        rfds.insert(resource_id);
        known = true;
      }
      if (pipe_map_.find(i) != pipe_map_.end()) {
        std::string resource_id = pipe_map_[i];
        rfds.insert(resource_id);
        known = true;
      }
      if (socket_map_.find(i) != socket_map_.end()) {
        std::string resource_id = socket_map_[i];
        rfds.insert(resource_id);
        known = true;
      }
      if (!known) {
        WARNING("Thread %d selects using an unkownn read file descriptor %d.",
                arbiter_->ThreadID(), i);
      }
    }
    // Write file descriptors
    if (writefds && System::FDSetContains(i, writefds)) {
      bool known = false;
      if (file_map_.find(i) != file_map_.end()) {
        std::string resource_id = file_map_[i];
        wfds.insert(resource_id);
        known = true;
      }
      if (pipe_map_.find(i) != pipe_map_.end()) {
        std::string resource_id = pipe_map_[i];
        wfds.insert(resource_id);
        known = true;
      }
      if (socket_map_.find(i) != socket_map_.end()) {
        std::string resource_id = socket_map_[i];
        wfds.insert(resource_id);
        known = true;
      }
      if (!known) {
        WARNING("Thread %d selects using an unknown write file descriptor %d.",
                arbiter_->ThreadID(), i);
      }
    }
  }
  System::MutexUnlock(&socket_map_mutex_);
  System::MutexUnlock(&pipe_map_mutex_);
  System::MutexUnlock(&file_map_mutex_);

  // If there were no file descriptors, return 0
  if (rfds.empty() && wfds.empty()) { return 0; }

  // Have arbiter resolve the call to socket
  dbug::thrift::SelectReturn select_return =
      arbiter_->Select(arbiter_->ThreadID(),
                       "select",
                       Util::GetBacktrace(),
                       rfds,
                       wfds,
                       timeout != NULL);

  // Decode resource identifiers into file descriptor sets
  if (readfds) { System::FDSetEmpty(readfds); }
  if (writefds) { System::FDSetEmpty(writefds); }

  int ret = 0;
  // Files
  System::MutexLock(&file_map_mutex_);
  for (std::pair<int, std::string> file : file_map_) {
    std::string resource_id = file.second;
    if (readfds &&
        select_return.rfds.find(resource_id) != select_return.rfds.end()) {
      System::FDSetAdd(file.first, readfds);
      ret++;
    }
    if (writefds &&
        select_return.wfds.find(resource_id) != select_return.wfds.end()) {
      System::FDSetAdd(file.first, writefds);
      ret++;
    }
  }
  System::MutexUnlock(&file_map_mutex_);
  // Pipes
  System::MutexLock(&pipe_map_mutex_);
  for (std::pair<int, std::string> pipe : pipe_map_) {
    std::string resource_id = pipe.second;
    if (readfds &&
        select_return.rfds.find(resource_id) != select_return.rfds.end()) {
      System::FDSetAdd(pipe.first, readfds);
      ret++;
    }
    if (writefds &&
        select_return.wfds.find(resource_id) != select_return.wfds.end()) {
      System::FDSetAdd(pipe.first, writefds);
      ret++;
    }
  }
  System::MutexUnlock(&pipe_map_mutex_);
  // Sockets
  System::MutexLock(&socket_map_mutex_);
  for (std::pair<int, std::string> socket : socket_map_) {
    std::string resource_id = socket.second;
    if (readfds &&
        select_return.rfds.find(resource_id) != select_return.rfds.end()) {
      System::FDSetAdd(socket.first, readfds);
      ret++;
    }
    if (writefds &&
        select_return.wfds.find(resource_id) != select_return.wfds.end()) {
      System::FDSetAdd(socket.first, writefds);
      ret++;
    }
  }
  System::MutexUnlock(&socket_map_mutex_);
  return ret;
}
#endif

#if DBUG_HAVE_SEND
ssize_t Handler::send(int socket,
                      const void * buffer,
                      size_t length,
                      int flags) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    System::MutexUnlock(&socket_map_mutex_);
    WARNING("Thread %d sends to an unknown socket %d.",
            arbiter_->ThreadID(), socket);
    return Original::send(socket, buffer, length, flags);
  }
  std::string resource_id = socket_map_[socket];
  System::MutexUnlock(&socket_map_mutex_);
  arbiter_->SocketSend(arbiter_->ThreadID(),
                       "send",
                       Util::GetBacktrace(),
                       resource_id,
                       length,
                       flags);
  ssize_t ret = Original::send(socket, buffer, length, flags);
  if (ret > 0) {
    arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, ret);
  }
  return ret;
}
#endif

#if DBUG_HAVE_SENDMSG
ssize_t Handler::sendmsg(int socket, const struct msghdr *message, int flags) {
  WARNING("Thread %d: dbug does not model sendmsg().", arbiter_->ThreadID());
  return Original::sendmsg(socket, message, flags);
}
#endif

#if DBUG_HAVE_SENDTO
ssize_t Handler::sendto(int socket,
                        const void * buffer,
                        size_t length,
                        int flags,
                        const System::SocketAddress *dest_addr,
                        socklen_t dest_len) {
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(socket) == socket_map_.end()) {
    System::MutexUnlock(&socket_map_mutex_);
    WARNING("Thread %d sends to an unknown socket %d.",
            arbiter_->ThreadID(), socket);
    return Original::sendto(socket, buffer, length, flags, dest_addr, dest_len);
  }
  std::string resource_id = socket_map_[socket];
  System::MutexUnlock(&socket_map_mutex_);
  if (!dest_addr) {
    arbiter_->SocketSend(arbiter_->ThreadID(),
                         "sendto",
                         Util::GetBacktrace(),
                         resource_id,
                         length,
                         flags);
  } else {
    const System::SocketIPv4Address *ip_addr =
        reinterpret_cast<const System::SocketIPv4Address *>(dest_addr);
    std::string address = System::IPv4NetworkToString(ip_addr->sin_addr);
    unsigned short port = System::NetworkToHostShort(ip_addr->sin_port);
    arbiter_->SocketSendTo(arbiter_->ThreadID(),
                           "sendto",
                           Util::GetBacktrace(),
                           resource_id,
                           length,
                           flags,
                           address,
                           static_cast<int>(port));
  }
  ssize_t ret =
      Original::sendto(socket, buffer, length, flags, dest_addr, dest_len);
  if (ret > 0) {
    arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, ret);
  }
  return ret;
}
#endif

#if DBUG_HAVE_SOCKET
int Handler::socket(int domain, int type, int protocol) {
  switch (domain) {
    case AF_INET:
    case AF_INET6: {
      int id = Util::NewIdentifier();
      std::string resource_id =
          Util::CreateResourceID("socket", arbiter_->ProcessID(), id);
      arbiter_->SocketOpen(arbiter_->ThreadID(),
                           "socket",
                           Util::GetBacktrace(),
                           resource_id,
                           domain,
                           type,
                           protocol);
      // Clear out the SOCK_CLOEXEC bit to prevent the OS from reusing
      // this file descriptor upon exec().
      int ret = Original::socket(domain, type & ~SOCK_CLOEXEC, protocol);
      if (ret != -1) {
        System::MutexLock(&socket_map_mutex_);
        socket_map_[ret] = resource_id;
        System::MutexUnlock(&socket_map_mutex_);
        // Send your MAC address to the arbiter
        std::string mac_address = Util::GetMACAddress(ret);
        if (mac_address == "") { ERROR("Could not get MAC address."); }
        arbiter_->SetSocketNodeID(arbiter_->ThreadID(),
                                  resource_id,
                                  mac_address);
        // Send the buffer size to the arbiter
        int opt;
        socklen_t opt_size = sizeof(opt_size);
        System::GetSocketOption(ret, SOL_SOCKET, SO_RCVBUF,
                                reinterpret_cast<void *>(&opt), &opt_size);
        arbiter_->SetSocketSize(arbiter_->ThreadID(), resource_id, opt);
      } else {
        arbiter_->SocketClose(arbiter_->ThreadID(),
                              "socket",
                              Util::GetBacktrace(),
                              resource_id);
      }
      return ret;
    }
    default:
      WARNING("Thread %d: dbug does not model socket domain %d.",
              domain, arbiter_->ThreadID());
      return Original::socket(domain, type, protocol);
  }
}
#endif

#if DBUG_HAVE_SOCKETPAIR
int Handler::socketpair(int domain, int type, int protocol, int *sv) {
  switch (domain) {
    case AF_UNIX: {
      int id1 = Util::NewIdentifier();
      std::string resource1_id =
          Util::CreateResourceID("socket", arbiter_->ProcessID(), id1);
      int id2 = Util::NewIdentifier();
      std::string resource2_id =
          Util::CreateResourceID("socket", arbiter_->ProcessID(), id2);
      arbiter_->SocketPair(arbiter_->ThreadID(),
                           "socket",
                           Util::GetBacktrace(),
                           resource1_id,
                           resource2_id,
                           domain,
                           type,
                           protocol);
      int ret = Original::socketpair(domain, type, protocol, sv);
      if (ret != -1) {
        System::MutexLock(&socket_map_mutex_);
        socket_map_[sv[0]] = resource1_id;
        socket_map_[sv[1]] = resource2_id;
        System::MutexUnlock(&socket_map_mutex_);
        // Send the buffer size to the arbiter
        int opt;
        socklen_t opt_size = sizeof(opt_size);
        System::GetSocketOption(sv[0], SOL_SOCKET, SO_RCVBUF,
                                reinterpret_cast<void *>(&opt), &opt_size);
        arbiter_->SetSocketSize(arbiter_->ThreadID(), resource1_id, opt);
        System::GetSocketOption(sv[1], SOL_SOCKET, SO_RCVBUF,
                                reinterpret_cast<void *>(&opt), &opt_size);
        arbiter_->SetSocketSize(arbiter_->ThreadID(), resource2_id, opt);
      } else {
        arbiter_->SocketClose(arbiter_->ThreadID(),
                              "socket",
                              Util::GetBacktrace(),
                              resource1_id);
        arbiter_->SocketClose(arbiter_->ThreadID(),
                              "socket",
                              Util::GetBacktrace(),
                              resource2_id);
      }
      return ret;
    }
    default:
      WARNING("Thread %d: dbug does not model socket domain %d.",
              arbiter_->ThreadID(), domain);
      return Original::socketpair(domain, type, protocol, sv);
  }
}
#endif

#if DBUG_HAVE_UNLINK
int Handler::unlink(const char *pathname) {
  WARNING("Thread %d: dbug does not model unlink().", arbiter_->ThreadID());
  return Original::unlink(pathname);
}
#endif

#if DBUG_HAVE_WRITE
ssize_t Handler::write(int fd, const void *buf, size_t count) {
  // File
  System::MutexLock(&file_map_mutex_);
  if (file_map_.find(fd) != file_map_.end()) {
    std::string resource_id = file_map_[fd];
    System::MutexUnlock(&file_map_mutex_);
    arbiter_->FileWrite(arbiter_->ThreadID(),
                        "write",
                        Util::GetBacktrace(),
                        resource_id,
                        count);
    ssize_t ret = Original::write(fd, buf, count);
    return ret;
  }
  System::MutexUnlock(&file_map_mutex_);
  // Pipe
  System::MutexLock(&pipe_map_mutex_);
  if (pipe_map_.find(fd) != pipe_map_.end()) {
    std::string resource_id = pipe_map_[fd];
    System::MutexUnlock(&pipe_map_mutex_);
    arbiter_->PipeWrite(arbiter_->ThreadID(),
                        "write",
                        Util::GetBacktrace(),
                        resource_id,
                        count);
    ssize_t ret = Original::write(fd, buf, count);
    if (ret > 0) {
      arbiter_->UpdatePipeBytes(arbiter_->ThreadID(), resource_id, ret);
    }
    return ret;
  }
  System::MutexUnlock(&pipe_map_mutex_);
  // Socket
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(fd) != socket_map_.end()) {
    std::string resource_id = socket_map_[fd];
    System::MutexUnlock(&socket_map_mutex_);
    arbiter_->SocketSend(arbiter_->ThreadID(),
                         "write",
                         Util::GetBacktrace(),
                         resource_id,
                         count,
                         0);
    ssize_t ret = Original::write(fd, buf, count);
    if (ret > 0) {
      arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, ret);
    }
    return ret;
  }
  System::MutexUnlock(&socket_map_mutex_);
  WARNING("Thread %d writes to an unknown file descriptor %d.",
          arbiter_->ThreadID(), fd);
  return Original::write(fd, buf, count);
}
#endif

#if DBUG_HAVE_WRITEV
ssize_t Handler::writev(int fd, const struct iovec *iov, int iovcnt) {
  int count = 0;
  for (int i = 0; i < iovcnt; i++) { count += iov[i].iov_len; }
  // File
  System::MutexLock(&file_map_mutex_);
  if (file_map_.find(fd) != file_map_.end()) {
    std::string resource_id = file_map_[fd];
    System::MutexUnlock(&file_map_mutex_);
    arbiter_->FileWrite(arbiter_->ThreadID(),
                        "writev",
                        Util::GetBacktrace(),
                        resource_id,
                        count);
    ssize_t ret = Original::writev(fd, iov, iovcnt);
    return ret;
  }
  System::MutexUnlock(&file_map_mutex_);
  // Pipe
  System::MutexLock(&pipe_map_mutex_);
  if (pipe_map_.find(fd) != pipe_map_.end()) {
    std::string resource_id = pipe_map_[fd];
    System::MutexUnlock(&pipe_map_mutex_);
    arbiter_->PipeWrite(arbiter_->ThreadID(),
                        "writev",
                        Util::GetBacktrace(),
                        resource_id,
                        count);
    ssize_t ret = Original::writev(fd, iov, iovcnt);
    if (ret > 0) {
      arbiter_->UpdatePipeBytes(arbiter_->ThreadID(), resource_id, ret);
    }
    return ret;
  }
  System::MutexUnlock(&pipe_map_mutex_);
  // Socket
  System::MutexLock(&socket_map_mutex_);
  if (socket_map_.find(fd) != socket_map_.end()) {
    std::string resource_id = socket_map_[fd];
    System::MutexUnlock(&socket_map_mutex_);
    arbiter_->SocketSend(arbiter_->ThreadID(),
                         "writev",
                         Util::GetBacktrace(),
                         resource_id,
                         count,
                         0);
    ssize_t ret = Original::writev(fd, iov, iovcnt);
    if (ret > 0) {
      arbiter_->UpdateSocketBytes(arbiter_->ThreadID(), resource_id, ret);
    }
    return ret;
  }
  System::MutexUnlock(&socket_map_mutex_);
  WARNING("Thread %d writes to an unknown file descriptor %d.",
          arbiter_->ThreadID(), fd);
  return Original::writev(fd, iov, iovcnt);
}
#endif

}  // namespace interposition
}  // namespace dbug
