/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Cancel threads when their request fails to execute.

#include "dbug/interposition/local-arbiter.h"

#include <list>
#include <map>
#include <sstream>
#include <string>
#include <vector>

#include "dbug/arbiter/handler.h"
#include "dbug/arbiter/process.h"
#include "dbug/arbiter/server.h"
#include "dbug/arbiter/thread.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/logging.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::arbiter::Process;
using dbug::arbiter::Thread;

using dbug::thrift::Location;
using dbug::thrift::ThreadCreateReturn;

namespace dbug {
namespace interposition {

LocalArbiter::LocalArbiter() : DefaultLocalArbiter() {}

LocalArbiter::~LocalArbiter() {}

void LocalArbiter::CancelThreadRegistration(const int tid) {
  DefaultLocalArbiter::UnregisterThread(tid);
}

int LocalArbiter::Choose(const int tid,
                         const std::string &func,
                         const std::vector<Location> &backtrace,
                         const int n) {
  int ret = 0;
  if (System::CompareStrings(func.c_str(), "dbug_choose") == 0) {
    ret = DefaultLocalArbiter::Choose(tid, func, backtrace, n);
  }
  return ret;
}

void LocalArbiter::CompleteThreadRegistration(const int pid,
                                              const int tid) {
  DefaultLocalArbiter::SetProcessInfo(tid,
                                      System::GetProcessID(),
                                      System::GetParentProcessID(),
                                      System::GetProcessGroupID(),
                                      Process::RUNNING,
                                      0,
                                      0);
  System::MutexLock(&processes_mutex_);
  System::MutexLock(&threads_mutex_);
  processes_[System::ThreadSelf()] = pid;
  ThreadInfo *new_thread = new ThreadInfo();
  new_thread->id = tid;
  new_thread->running = true;
  threads_[System::ThreadSelf()] = new_thread;
  nrunning_threads_++;
  System::MutexUnlock(&threads_mutex_);
  System::MutexUnlock(&processes_mutex_);
}

void LocalArbiter::PostThreadRegistration(
    const int tid,
    const std::string &func,
    const std::vector<Location> &backtrace,
    const bool create_process,
    int *new_pid,
    int *new_tid) {
  ThreadCreateReturn ret =
      DefaultLocalArbiter::ThreadCreate(tid, func, backtrace, create_process);
  *new_pid = ret.pid;
  *new_tid = ret.tid;
}

void LocalArbiter::ProcessSignal(const int tid,
                                 const std::string &func,
                                 const std::vector<Location> &backtrace,
                                 const int pid,
                                 const int signal) {
  System::MutexLock(&threads_mutex_);
  for (std::pair<pthread_t, ThreadInfo *> pair : threads_) {
    ThreadInfo *thread = pair.second;
    if (thread->running) {
      thread->running = false;
      nrunning_threads_--;
    }
  }
  assert(nrunning_threads_ == 0);
  System::MutexUnlock(&threads_mutex_);
  DefaultLocalArbiter::ProcessSignal(tid, func, backtrace, pid, signal);
}

void LocalArbiter::ProcessTerminate(const int tid,
                                    const std::string &func,
                                    const std::vector<Location> &backtrace,
                                    const int pid) {
  System::MutexLock(&threads_mutex_);
  for (std::pair<pthread_t, ThreadInfo *> pair : threads_) {
    ThreadInfo *thread = pair.second;
    if (thread->running) {
      thread->running = false;
      nrunning_threads_--;
    }
  }
  assert(nrunning_threads_ == 0);
  System::MutexUnlock(&threads_mutex_);
  DefaultLocalArbiter::ProcessTerminate(tid, func, backtrace, pid);
}

void LocalArbiter::RegisterThread() {
  System::MutexLock(&threads_mutex_);
  if (threads_.find(System::ThreadSelf()) == threads_.end()) {
    System::MutexUnlock(&threads_mutex_);
    RegisterThreadReturn ret =
        DefaultLocalArbiter::RegisterThread(0, true);
    DefaultLocalArbiter::SetProcessInfo(
        ret.tid,
        System::GetProcessID(),
        System::GetParentProcessID(),
        System::GetProcessGroupID(),
        Process::RUNNING,
        0,
        0);
    System::MutexLock(&processes_mutex_);
    System::MutexLock(&threads_mutex_);
    processes_[System::ThreadSelf()] = ret.pid;
    ThreadInfo *new_thread = new ThreadInfo();
    new_thread->id = ret.tid;
    new_thread->running = true;
    threads_[System::ThreadSelf()] = new_thread;
    nrunning_threads_++;
    System::MutexUnlock(&threads_mutex_);
    System::MutexUnlock(&processes_mutex_);
  } else {
    System::MutexUnlock(&threads_mutex_);
    WARNING("Trying to register an existing thread.");
  }
}

void LocalArbiter::UnregisterAllThreads() {
  std::list<int> unregister_queue;
  System::MutexLock(&threads_mutex_);
  for (std::pair<System::Thread, ThreadInfo *> pair : threads_) {
    ThreadInfo *thread = pair.second;
    if (thread->running) {
      thread->running = false;
      nrunning_threads_--;
      unregister_queue.push_back(thread->id);
    }
  }
  assert(nrunning_threads_ == 0);
  System::MutexUnlock(&threads_mutex_);
  for (int tid : unregister_queue) {
    DefaultLocalArbiter::UnregisterThread(tid);
  }
}

void LocalArbiter::UnregisterThread() {
  System::MutexLock(&threads_mutex_);
  threads_[System::ThreadSelf()]->running = false;
  nrunning_threads_--;
  System::MutexUnlock(&threads_mutex_);
  DefaultLocalArbiter::UnregisterThread(ThreadID());
}

}  // namespace interposition
}  // namespace dbug
