/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Use virtual time to model sleep() and usleep().

// TODO(jsimsa): Implement coordination for sigwait().

// TODO(jsimsa): Implement time travel version of record-replay for
// gettimeofday() and time().

#include "dbug/common/system.h"
#include "dbug/common/time.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/misc.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_CLOCK
clock_t Handler::clock() {
  clock_t ret = Original::clock();
  if (ret != -1) { ret = arbiter_->RecordNonDeterminism(ret); }
  return ret;
}
#endif

#if DBUG_HAVE_GETTIMEOFDAY
int Handler::gettimeofday(struct timeval *tv, struct timezone *tz) {
  int ret = Original::gettimeofday(tv, tz);
  if (ret != -1) {
    if (tv) {
      tv->tv_sec = arbiter_->RecordNonDeterminism(tv->tv_sec);
      tv->tv_usec = arbiter_->RecordNonDeterminism(tv->tv_usec);
    }
    if (tz) {
      tz->tz_minuteswest = arbiter_->RecordNonDeterminism(tz->tz_minuteswest);
      tz->tz_dsttime = arbiter_->RecordNonDeterminism(tz->tz_dsttime);
    }
  }
  return ret;
  /* UNUSED relative record-replay mechanism
  // Compute the delta from start of the process.
  System::Int64 sec_delta = tv->tv_sec - gettimeofday_.tv_sec;
  System::Int64 usec_delta = tv->tv_usec - gettimeofday_.tv_usec;
  if (usec_delta < 0) {
  usec_delta += Time::kS2U;
  sec_delta -= 1;
  }
  // Record the non-deterministic deltas.
  sec_delta = arbiter_->RecordNonDeterminism(sec_delta);
  usec_delta = arbiter_->RecordNonDeterminism(usec_delta);
  // Compute the final value.
  tv->tv_sec = gettimeofday_.tv_sec + sec_delta;
  tv->tv_usec = gettimeofday_.tv_usec + usec_delta;
  if (tv->tv_usec > Time::kS2U) {
  tv->tv_usec -= Time::kS2U;
  tv->tv_sec += 1;
  }
  */
}
#endif

#if DBUG_HAVE_NANOSLEEP
int Handler::nanosleep(const System::TimeSpec * /* rqtp */,
                       System::TimeSpec * /* rmtp */) {
  arbiter_->Sleep(arbiter_->ThreadID(),
                  "nanosleep",
                  Util::GetBacktrace());
  return 0;
}
#endif

#if DBUG_HAVE_SCHED_YIELD
int Handler::sched_yield() {
  arbiter_->Yield(arbiter_->ThreadID(),
                  "sched_yield",
                  Util::GetBacktrace());
  return 0;
}
#endif

#if DBUG_HAVE_SIGWAIT
int Handler::sigwait(const sigset_t * DBUG_RESTRICT set,
                     int * DBUG_RESTRICT sig) {
  int ret;
  if (Interpose()) {
    // We unregister the waiting thread. This hack is used to make dbug
    // work with programs (such as aget) that use a deamon thread that
    // waits for external signals.
    Disable(NULL);
    arbiter_->UnregisterThread();
    EnableFork();
    EnableCancellation();
    ret = Original::sigwait(set, sig);
    DEBUG_EXTERNAL("%p, %p", "%d", set, sig, ret);
    return ret;
  } else {
    ret = Original::sigwait(set, sig);
    DEBUG_INTERNAL("%p, %p", "%d", set, sig, ret);
    return ret;
    assert(0);
  }
}
#endif

#if DBUG_HAVE_SLEEP
unsigned Handler::sleep(unsigned /* seconds */) {
  arbiter_->Sleep(arbiter_->ThreadID(),
                  "sleep",
                  Util::GetBacktrace());
  return 0;
}
#endif

#if DBUG_HAVE_TIME
time_t Handler::time(time_t *t) {
  time_t ret = Original::time(t);
  if (ret != -1) {
    ret = arbiter_->RecordNonDeterminism(ret);
    if (t) { *t = ret; }
  }
  return ret;
  /* UNUSED relative record-replay mechanism
  // Compute the delta from start of the process.
  time_t delta = ret - time_;
  // Record the non-deterministic delta.
  delta = arbiter_->RecordNonDeterminism(delta);
  // Compute the final value.
  ret = time_ + delta;
  if (t) { *t = ret; }
  }
  */
}
#endif

#if DBUG_HAVE_USLEEP
int Handler::usleep(useconds_t /* usecs */) {
  arbiter_->Sleep(arbiter_->ThreadID(),
                  "usleep",
                  Util::GetBacktrace());
  return 0;
}
#endif

}  // namespace interposition
}  // namespace dbug
