/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Model virtual time of mutex_timedlock().

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/pthread-mutex.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_PTHREAD_MUTEX_DESTROY
int Handler::pthread_mutex_destroy(pthread_mutex_t *mutex) {
  // Ignore mutexes that are to be ignored.
  System::MutexLock(&ignore_mutex_set_mutex_);
  if (ignore_mutex_set_.find(mutex) != ignore_mutex_set_.end()) {
    System::MutexUnlock(&ignore_mutex_set_mutex_);
    return Original::pthread_mutex_destroy(mutex);
  }
  System::MutexUnlock(&ignore_mutex_set_mutex_);
  std::string resource_id;
  System::MutexLock(&mutex_map_mutex_);
  if (mutex_map_.find(mutex) != mutex_map_.end()) {
    resource_id = mutex_map_[mutex];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-mutex", arbiter_->ProcessID(), id);
    mutex_map_[mutex] = resource_id;
  }
  System::MutexUnlock(&mutex_map_mutex_);
  return arbiter_->MutexDelete(arbiter_->ThreadID(),
                               "pthread_mutex_destroy",
                               Util::GetBacktrace(),
                               resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_MUTEX_INIT
int Handler::pthread_mutex_init(pthread_mutex_t *mutex,
                                const pthread_mutexattr_t * attr) {
  // Ignore mutexes that are to be ignored.
  System::MutexLock(&ignore_mutex_set_mutex_);
  if (ignore_mutex_set_.find(reinterpret_cast<pthread_mutex_t *>(mutex)) !=
      ignore_mutex_set_.end()) {
    System::MutexUnlock(&ignore_mutex_set_mutex_);
    return Original::pthread_mutex_init(mutex, attr);
  }
  System::MutexUnlock(&ignore_mutex_set_mutex_);

  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_mutex_init",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of non-memory resources.
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return ENOMEM;
    default:
      std::string resource_id;
      System::MutexLock(&mutex_map_mutex_);
      if (mutex_map_.find(mutex) != mutex_map_.end()) {
        resource_id = mutex_map_[mutex];
      } else {
        int id = Util::NewIdentifier();
        resource_id =
            Util::CreateResourceID("mutex", arbiter_->ProcessID(), id);
        mutex_map_[mutex] = resource_id;
      }
      System::MutexUnlock(&mutex_map_mutex_);
      int ret;
      if (attr == NULL) {
        ret = arbiter_->MutexCreate(arbiter_->ThreadID(),
                                    "pthread_mutex_init",
                                    Util::GetBacktrace(),
                                    resource_id,
                                    false);
      } else {
        int type;
        pthread_mutexattr_gettype(attr, &type);
        switch (type) {
          case PTHREAD_MUTEX_RECURSIVE:
            ret = arbiter_->MutexCreate(arbiter_->ThreadID(),
                                        "pthread_mutex_init",
                                        Util::GetBacktrace(),
                                        resource_id,
                                        true);
            break;
          default:
            ret = arbiter_->MutexCreate(arbiter_->ThreadID(),
                                        "pthread_mutex_init",
                                        Util::GetBacktrace(),
                                        resource_id,
                                        false);
            break;
        }
      }
      return ret;
  }
}
#endif

#if DBUG_HAVE_PTHREAD_MUTEX_LOCK
int Handler::pthread_mutex_lock(pthread_mutex_t *mutex) {
  // Ignore mutexes that are to be ignored.
  System::MutexLock(&ignore_mutex_set_mutex_);
  if (ignore_mutex_set_.find(mutex) != ignore_mutex_set_.end()) {
    System::MutexUnlock(&ignore_mutex_set_mutex_);
    return Original::pthread_mutex_lock(mutex);
  }
  System::MutexUnlock(&ignore_mutex_set_mutex_);
  std::string resource_id;
  System::MutexLock(&mutex_map_mutex_);
  if (mutex_map_.find(mutex) != mutex_map_.end()) {
    resource_id = mutex_map_[mutex];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-mutex", arbiter_->ProcessID(), id);
    mutex_map_[mutex] = resource_id;
  }
  System::MutexUnlock(&mutex_map_mutex_);
  return arbiter_->MutexLock(arbiter_->ThreadID(),
                             "pthread_mutex_lock",
                             Util::GetBacktrace(),
                             resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_MUTEX_TIMEDLOCK
int Handler::pthread_mutex_timedlock(pthread_mutex_t *mutex,
                                     const struct timespec */* abs_time */) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_mutex_timedlock",
                           Util::GetBacktrace(),
                           2)) {
    case 1:
      // Simulate time out.
      errno = ETIMEDOUT;
      return ETIMEDOUT;
    default:
      return Handler::pthread_mutex_lock(mutex);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_MUTEX_TRYLOCK
int Handler::pthread_mutex_trylock(pthread_mutex_t *mutex) {
  // Ignore mutexes that are to be ignored.
  System::MutexLock(&ignore_mutex_set_mutex_);
  if (ignore_mutex_set_.find(mutex) != ignore_mutex_set_.end()) {
    System::MutexUnlock(&ignore_mutex_set_mutex_);
    return Original::pthread_mutex_trylock(mutex);
  }
  System::MutexUnlock(&ignore_mutex_set_mutex_);
  std::string resource_id;
  System::MutexLock(&mutex_map_mutex_);
  if (mutex_map_.find(mutex) != mutex_map_.end()) {
    resource_id = mutex_map_[mutex];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-mutex", arbiter_->ProcessID(), id);
    mutex_map_[mutex] = resource_id;
  }
  System::MutexUnlock(&mutex_map_mutex_);
  return arbiter_->MutexTryLock(arbiter_->ThreadID(),
                                "pthread_mutex_trylock",
                                Util::GetBacktrace(),
                                resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_MUTEX_UNLOCK
int Handler::pthread_mutex_unlock(pthread_mutex_t *mutex) {
  // Ignore mutexes that are to be ignored.
  System::MutexLock(&ignore_mutex_set_mutex_);
  if (ignore_mutex_set_.find(mutex) != ignore_mutex_set_.end()) {
    System::MutexUnlock(&ignore_mutex_set_mutex_);
    return Original::pthread_mutex_unlock(mutex);
  }
  System::MutexUnlock(&ignore_mutex_set_mutex_);
  std::string resource_id;
  System::MutexLock(&mutex_map_mutex_);
  if (mutex_map_.find(mutex) != mutex_map_.end()) {
    resource_id = mutex_map_[mutex];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-mutex", arbiter_->ProcessID(), id);
    mutex_map_[mutex] = resource_id;
  }
  System::MutexUnlock(&mutex_map_mutex_);
  return arbiter_->MutexUnlock(arbiter_->ThreadID(),
                               "pthread_mutex_unlock",
                               Util::GetBacktrace(),
                               resource_id);
}
#endif

}  // namespace interposition
}  // namespace dbug
