/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Model virtual time of rwlock_timedrdlock() and
// rwlock_timedwrlock().

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/pthread-rwlock.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_PTHREAD_RWLOCK_DESTROY
int Handler::pthread_rwlock_destroy(pthread_rwlock_t *rwlock) {
  std::string resource_id;
  System::MutexLock(&rwlock_map_mutex_);
  if (rwlock_map_.find(rwlock) != rwlock_map_.end()) {
    resource_id = rwlock_map_[rwlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-rwlock", arbiter_->ProcessID(), id);
    rwlock_map_[rwlock] = resource_id;
  }
  System::MutexUnlock(&rwlock_map_mutex_);
  return arbiter_->RWLockDelete(arbiter_->ThreadID(),
                                "pthread_rwlock_destroy",
                                Util::GetBacktrace(),
                                resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_INIT
int Handler::pthread_rwlock_init(pthread_rwlock_t *rwlock,
                                 const pthread_rwlockattr_t * /* attr */) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_rwlock_init",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate the lack of non-memory resources
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate the lack of memory resources
      errno = ENOMEM;
      return ENOMEM;
    default:
      std::string resource_id;
      System::MutexLock(&rwlock_map_mutex_);
      if (rwlock_map_.find(rwlock) != rwlock_map_.end()) {
        resource_id = rwlock_map_[rwlock];
      } else {
        int id = Util::NewIdentifier();
        resource_id =
            Util::CreateResourceID("rwlock", arbiter_->ProcessID(), id);
        rwlock_map_[rwlock] = resource_id;
      }
      System::MutexUnlock(&rwlock_map_mutex_);
      return arbiter_->RWLockCreate(arbiter_->ThreadID(),
                                    "pthread_rwlock_init",
                                    Util::GetBacktrace(),
                                    resource_id);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_RDLOCK
int Handler::pthread_rwlock_rdlock(pthread_rwlock_t *rwlock) {
  std::string resource_id;
  System::MutexLock(&rwlock_map_mutex_);
  if (rwlock_map_.find(rwlock) != rwlock_map_.end()) {
    resource_id = rwlock_map_[rwlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-rwlock", arbiter_->ProcessID(), id);
    rwlock_map_[rwlock] = resource_id;
  }
  System::MutexUnlock(&rwlock_map_mutex_);
  return arbiter_->RWLockReadLock(arbiter_->ThreadID(),
                                  "pthread_rwlock_rdlock",
                                  Util::GetBacktrace(),
                                  resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_TIMEDRDLOCK
int Handler::pthread_rwlock_timedrdlock(
    pthread_rwlock_t *rwlock,
    const struct timespec * /* abs_time */) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_rwlock_timedrdlock",
                           Util::GetBacktrace(),
                           2)) {
    case 1:
      // Simulate time out.
      errno = ETIMEDOUT;
      return ETIMEDOUT;
    default:
      return Handler::pthread_rwlock_rdlock(rwlock);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_TIMEDWRLOCK
int Handler::pthread_rwlock_timedwrlock(
    pthread_rwlock_t *rwlock,
    const struct timespec * /* abs_time */) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_rwlock_timedwrlock",
                           Util::GetBacktrace(),
                           2)) {
    case 1:
      // Simulate time out.
      errno = ETIMEDOUT;
      return ETIMEDOUT;
    default:
      return Handler::pthread_rwlock_wrlock(rwlock);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_TRYRDLOCK
int Handler::pthread_rwlock_tryrdlock(pthread_rwlock_t *rwlock) {
  std::string resource_id;
  System::MutexLock(&rwlock_map_mutex_);
  if (rwlock_map_.find(rwlock) != rwlock_map_.end()) {
    resource_id = rwlock_map_[rwlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-rwlock", arbiter_->ProcessID(), id);
    rwlock_map_[rwlock] = resource_id;
  }
  System::MutexUnlock(&rwlock_map_mutex_);
  return arbiter_->RWLockTryReadLock(arbiter_->ThreadID(),
                                     "pthread_rwlock_tryrdlock",
                                     Util::GetBacktrace(),
                                     resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_TRYWRLOCK
int Handler::pthread_rwlock_trywrlock(pthread_rwlock_t *rwlock) {
  std::string resource_id;
  System::MutexLock(&rwlock_map_mutex_);
  if (rwlock_map_.find(rwlock) != rwlock_map_.end()) {
    resource_id = rwlock_map_[rwlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-rwlock", arbiter_->ProcessID(), id);
    rwlock_map_[rwlock] = resource_id;
  }
  System::MutexUnlock(&rwlock_map_mutex_);
  return arbiter_->RWLockTryWriteLock(arbiter_->ThreadID(),
                                      "pthread_rwlock_trywrlock",
                                      Util::GetBacktrace(),
                                      resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_UNLOCK
int Handler::pthread_rwlock_unlock(pthread_rwlock_t *rwlock) {
  std::string resource_id;
  System::MutexLock(&rwlock_map_mutex_);
  if (rwlock_map_.find(rwlock) != rwlock_map_.end()) {
    resource_id = rwlock_map_[rwlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-rwlock", arbiter_->ProcessID(), id);
    rwlock_map_[rwlock] = resource_id;
  }
  System::MutexUnlock(&rwlock_map_mutex_);
  return arbiter_->RWLockUnlock(arbiter_->ThreadID(),
                                "pthread_rwlock_unlock",
                                Util::GetBacktrace(),
                                resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_RWLOCK_WRLOCK
int Handler::pthread_rwlock_wrlock(pthread_rwlock_t *rwlock) {
  std::string resource_id;
  System::MutexLock(&rwlock_map_mutex_);
  if (rwlock_map_.find(rwlock) != rwlock_map_.end()) {
    resource_id = rwlock_map_[rwlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-rwlock", arbiter_->ProcessID(), id);
    rwlock_map_[rwlock] = resource_id;
  }
  System::MutexUnlock(&rwlock_map_mutex_);
  return arbiter_->RWLockWriteLock(arbiter_->ThreadID(),
                                   "pthread_rwlock_wrlock",
                                   Util::GetBacktrace(),
                                   resource_id);
}
#endif

}  // namespace interposition
}  // namespace dbug
