/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sstream>
#include <string>

#include "dbug/common/command.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "gtest/gtest.h"

namespace dbug {
namespace {

void CreateXMLConfigFile(int num_processes,
                         int timeout,
                         bool use_dpor,
                         std::string test_filename) {
  System::File *file = System::FileOpen("config.xml", "w");
  std::stringstream stream;
  stream << "<config>\n"
         << "  <arbiter arguments=\" --start_num_threads "
         << num_processes << "\"/>\n"
         << "  <explorer timeout=\"" << timeout << "\""
         << " dpor=\"" << (use_dpor ? "true" : "false") << "\"/>\n"
         << "  <mpi_program mpirun=\"" << MPIEXEC << "\""
         << " np=\"" << num_processes << "\""
         << " command=\"" << BUILDDIR << "/" << test_filename << "\"/>\n"
         << "</config>\n";
  System::FilePrint(file, stream.str().c_str());
  System::FileClose(file);
}

TEST(MPITest, DISABLED_Abort) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-abort");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, AllGather) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-allgather");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, AllGatherVariable) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-allgatherv");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, AllReduce) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-allreduce");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, AllToAll) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-alltoall");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, AllToAllVariable) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-alltoallv");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, AllToAllVariableType) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-alltoallw");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Barrier) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-barrier");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Broadcast) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-broadcast");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, BufferedSend) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-bsend");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, CartesianCommunicator) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(4, 10, false, "mpi-cart-create");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, CommunicatorCreation) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-comm-create");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, CommunicatorCreationGroup) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(4, 10, false, "mpi-comm-create-group");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, CommunicatorDuplication) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-comm-dup");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, CommunicatorSplit) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(4, 10, false, "mpi-comm-split");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Deadlock) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-deadlock");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Exchange) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-exchange");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, ExchangeThreaded) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-exchange-mt");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, NonBlockingExchange) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 60, false, "mpi-iexchange");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, ExclusiveScan) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-exscan");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Gather) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-gather");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, GatherVariable) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-gatherv");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, NonDeterminism) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, true, "mpi-nondet");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, MatchingNonDeterminism) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, true, "mpi-mnondet");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Order) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-order");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, NonBlockingOrder) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 60, false, "mpi-iorder");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Probe) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-probe");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, NonBlockingProbe) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 60, false, "mpi-iprobe");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, MatchingProbe) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-mprobe");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, NonBlockingMatchingProbe) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 60, false, "mpi-improbe");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, ProbeDataType) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(3, 10, false, "mpi-probe-data");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Progress) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-progress");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, NonBlockingProgress) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 60, false, "mpi-iprogress");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, ReadySend) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 10, false, "mpi-rsend");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Reduce) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-reduce");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, ReduceScatter) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-reduce-scatter");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, ReduceScatterBlock) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-reduce-scatter-block");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Scan) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-scan");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Scatter) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-scatter");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, ScatterVariable) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(5, 10, false, "mpi-scatterv");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, SymmetricSendRecv1) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(1, 10, false, "mpi-sendrecv-n");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, SymmetricSendRecv4) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(4, 60, false, "mpi-sendrecv-n");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, AsymmetricSendRecv) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(3, 10, false, "mpi-sendrecv");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, Test) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(2, 20, true, "mpi-test");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, TestAll) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(3, 60, true, "mpi-testall");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, TestAny) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(3, 60, true, "mpi-testany");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(MPITest, TestSome) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(3, 60, true, "mpi-testsome");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

}  // anonymous namespace
}  // namespace dbug

int main(int argc, char **argv) {
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
